/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Encoder
FILE:       BifsEnc.cpp
AUTHOR:     Zvi Lifshitz (VDOnet Corp.)
OVERVIEW:   Implementation of BIFSEncoder, the class that encodes date into
			bits.
-------------------------------------------------------------------------*/
#include "defs.h"
#include "BifsEnc.h"
#include "NodeFld.h"
#include "Token.h"
#include "StrAssoc.h"

/*-----------------------------------------------------------------
BIFSEncoder constructor
-----------------------------------------------------------------*/
BIFSEncoder :: BIFSEncoder (FILE *fBifs, FILE *fList)
{
	m_fBifs = fBifs;
	m_fList = fList;
	m_bLastByte = 0;
	m_nBitCount = 0;
	m_nAULength = 0;
}

/*-----------------------------------------------------------------
Flush the buffer when finished
-----------------------------------------------------------------*/
void BIFSEncoder :: Flush ()
{
	if (m_nBitCount > 0)
	{	m_bLastByte <<= 8 - m_nBitCount;
		m_nBitCount = 0;
		if (m_nAULength >= 255)
			Flush ();
		m_buffer [m_nAULength++] = m_bLastByte;
	}
	if (m_nAULength > 0)
	{	fputc (HIBYTE (HIWORD (m_dwTS)), m_fBifs);
		fputc (LOBYTE (HIWORD (m_dwTS)), m_fBifs);
		fputc (HIBYTE (LOWORD (m_dwTS)), m_fBifs);
		fputc (LOBYTE (LOWORD (m_dwTS)), m_fBifs);
		fputc (m_nAULength, m_fBifs);
		fwrite (m_buffer, m_nAULength, 1, m_fBifs);
		m_nAULength = 0;
	}
}

/*-----------------------------------------------------------------
Encode nValue into nBits bits, and report this action in the list file
-----------------------------------------------------------------*/
void BIFSEncoder :: EncodeInt (int nValue, int nBits, LPCSTR szDesc)
{
	fprintf (m_fList, "bit (%d) %s = %d\n", nBits, szDesc, nValue);
	nValue <<= sizeof (int) * 8 - nBits;
	while (--nBits >= 0)
	{	WriteBit (nValue < 0);
		nValue <<= 1;
	}
}

/*-----------------------------------------------------------------
Encode a float value
-----------------------------------------------------------------*/
void BIFSEncoder :: EncodeFloat (float fValue)
{
#ifdef WIN32
	union
	{	float f;
		char sz [4];
	} value;
	value.f = fValue;

	for (int i = 0; i < 32; i++)
		WriteBit ((value.sz [3 - i / 8] & 1 << (7 - i % 8)) != 0);
#endif		// WIN32
	fprintf (m_fList, "float value = %f\n", fValue);
}

/*-----------------------------------------------------------------
Encode a double value
-----------------------------------------------------------------*/
void BIFSEncoder :: EncodeDouble (double dValue)
{
#ifdef WIN32
	union
	{	double d;
		char sz [8];
	} value;
	value.d = dValue;

	for (int i = 0; i < 64; i++)
		WriteBit ((value.sz [7 - i / 8] & 1 << (7 - i % 8)) != 0);
#endif		// WIN32
	fprintf (m_fList, "double value = %lf\n", dValue);
}

/*-----------------------------------------------------------------
Add one bit to the output stream
-----------------------------------------------------------------*/
void BIFSEncoder :: WriteBit (BOOL bit)
{
	m_bLastByte <<= 1;
	m_bLastByte |= bit;
	if (++m_nBitCount == 8)
	{	m_nBitCount = 0;
		if (m_nAULength >= 255)
			Flush ();
		m_buffer [m_nAULength++] = m_bLastByte;
		m_bLastByte = 0;
	}
}

/*-----------------------------------------------------------------
Read and encode a scene
-----------------------------------------------------------------*/
void BIFSEncoder :: Encode ()
{
	try
	{
		StoreTimeStamp (0);
		EncodeScene ();
		UpdateScene ();
	}
	catch (ExceptionEndOfFile)
	{
		nextToken.Error ("Unexpected end of file");
	}
}

/*-----------------------------------------------------------------
Encode Routes
-----------------------------------------------------------------*/
void BIFSEncoder :: EncodeRoutesDescription ()
{
	BOOL bMoreRoutes;

	EncodeInt (1, 1, "listDescription");
	do
	{
		EncodeRoute ();
		bMoreRoutes = nextToken == TID_ROUTE;
		EncodeInt (bMoreRoutes, 1, "moreRoutes");
	}
	while (bMoreRoutes);
}

/*-----------------------------------------------------------------
Encode one Route
-----------------------------------------------------------------*/
void BIFSEncoder :: EncodeRoute ()
{
	nextToken.Scan ();
	BOOL bIsUpdateable = nextToken == TID_DEF;
	EncodeInt (bIsUpdateable, 1, "isUpdateable");
	if (bIsUpdateable)
	{	int nID;
		nextToken.Scan ();
		if (nextToken == TID_NAME)
		{	nID = nextToken.GetDefinedRoute ();
			if (nID < 0)
				nID = nextToken.AddPredefinedRoute ();
			else
				nextToken.Error ("Route name already defined");
		}
		else
			nextToken.Error ("Unexpected symbol");
		nextToken.Scan ();
		EncodeInt (nID, 10, "RouteID");
	}
		
	EncodeRouteDetails ();
}

/*-----------------------------------------------------------------
Encode update commands
-----------------------------------------------------------------*/
void BIFSEncoder :: UpdateScene ()
{
	/*-----------------------------------------------------------------
	Loop for each update frame
	-----------------------------------------------------------------*/
	while (nextToken != TID_EOF)
	{	if (nextToken != TID_AT)
			nextToken.Error ("AT expected");
		else
			nextToken.Scan ();
		if (nextToken != TID_INTEGER)
			nextToken.Error ("Time (in milliseconds) expected");
		StoreTimeStamp (nextToken.GetInt ());

		if (nextToken != TID_LEFT_BRACE)
			nextToken.Error ("{ expected");
		nextToken.Scan ();
		while (nextToken != TID_RIGHT_BRACE)
		{
			switch (nextToken) {
			case TID_INSERT:
				InsertCommand ();
				break;
			case TID_REPLACE:
				ReplaceCommand ();
				break;
			case TID_DELETE:
				DeleteCommand ();
				break;
			case TID_APPEND:
				AppendCommand ();
				break;
			default:
				nextToken.Error ("Unexpected symbol");
				nextToken.Scan ();
			}
			EncodeInt (nextToken != TID_RIGHT_BRACE, 1, "continue");
		}
		nextToken.Scan ();
	}
}

/*-----------------------------------------------------------------
Start a new frame with the supplied time stamp
-----------------------------------------------------------------*/
void BIFSEncoder :: StoreTimeStamp (DWORD dwTS)
{
	Flush ();
	m_dwTS = dwTS;
}

/*-----------------------------------------------------------------
Encode an INSERT command
-----------------------------------------------------------------*/
void BIFSEncoder :: InsertCommand ()
{
	nextToken.Scan ();
	EncodeInt (0, 2, "InsertionCommand");
	switch (nextToken) {
	case TID_AT:
		InsertToField (FALSE);
		break;
	case TID_ROUTE:
		EncodeInt (2, 2, "RouteInsertion");
		EncodeRoute ();
		break;
	default:
		nextToken.Error ("Unexpected symbol");
		nextToken.Scan ();
	}
}

/*-----------------------------------------------------------------
Encode a replacement command
-----------------------------------------------------------------*/
void BIFSEncoder :: ReplaceCommand ()
{
	nextToken.Scan ();
	if (nextToken == TID_SCENE)
	{	SceneReplaceCommand ();
		return;
	}

	EncodeInt (2, 2, "ReplacementCommand");
	if (nextToken == TID_ROUTE)
		ReplaceRoute ();
	else
	{	BOOL bLast = FALSE;
		if (nextToken == TID_LAST)
		{	bLast = TRUE;
			nextToken.Scan ();
		}
		MediaObject *pNode = ScanNode ();
		if (pNode)
		{	if (!bLast && nextToken == TID_BY)
			{	EncodeInt (0, 2, "NodeReplacement");
				EncodeInt (pNode -> GetID (), 10, "NodeID");
				nextToken.Scan ();
				SFGenericNode ().Encode (this);
			}
			else
				pNode -> FieldReplacement (this, bLast);
		}
		delete pNode;
	}
}

/*-----------------------------------------------------------------
Encode DELETE command
-----------------------------------------------------------------*/
void BIFSEncoder :: DeleteCommand ()
{
	nextToken.Scan ();
	EncodeInt (1, 2, "DeletionCommand");
	if (nextToken == TID_ROUTE)
	{	nextToken.Scan ();
		EncodeInt (2, 2, "RouteDeletion");
		EncodeRouteID ();
	}
	else
	{	BOOL bLast = FALSE;
		if (nextToken == TID_LAST)
		{	bLast = TRUE;
			nextToken.Scan ();
		}
		MediaObject *pNode = ScanNode ();
		if (pNode)
		{	if (nextToken != TID_PERIOD)
			{	EncodeInt (0, 2, "NodeDeletion");
				EncodeInt (pNode -> GetID (), 10, "NodeID");
			}
			else
				pNode -> IndexedValueDeletion (this, bLast);
		}
		delete pNode;
	}
}

/*-----------------------------------------------------------------
Encode insertion of a value to an MF node field. If bAppend is TRUE the
value is appended to the end of the field, otherwise index is given
-----------------------------------------------------------------*/
void BIFSEncoder :: InsertToField (BOOL bAppend)
{
	nextToken.Scan ();
	MediaObject *pNode = ScanNode ();
	if (pNode)
		pNode -> InsertToField (this, bAppend);
	delete pNode;
}

/*-----------------------------------------------------------------
Scan a node name and instantiate the node
-----------------------------------------------------------------*/
MediaObject * BIFSEncoder :: ScanNode ()
{
	MediaObject *pNode = NULL;
	int nNodeID = EncodeNodeID ();
	if (nNodeID >= 0)
		pNode = SFGenericNode ().CreateNode (
			nextToken.GetNodeID (nNodeID)
		);
	if (pNode)
		pNode -> SetID (nNodeID);
	return pNode;
}

/*-----------------------------------------------------------------
Encode and APPEND TO command
-----------------------------------------------------------------*/
void BIFSEncoder :: AppendCommand ()
{
	nextToken.Scan ();
	EncodeInt (0, 2, "InsertionCommand");
	if (nextToken != TID_TO)
		nextToken.Error ("TO expected");
	InsertToField (TRUE);
}

/*-----------------------------------------------------------------
Next token is route name. Translate it to route ID
-----------------------------------------------------------------*/
void BIFSEncoder :: EncodeRouteID ()
{
	int nID;

	if (nextToken == TID_NAME)
	{	nID = nextToken.GetDefinedRoute ();
		if (nID < 0)
			nextToken.Error ("Undefined route name");
		EncodeInt (nID, 10, "RouteID");
	}
	else
		nextToken.Error ("Unexpected symbol");
	nextToken.Scan ();
}

/*-----------------------------------------------------------------
Next token is node name. Translate it to route ID and return the ID
-----------------------------------------------------------------*/
int BIFSEncoder :: EncodeNodeID ()
{
	int nNodeID;
	if (nextToken == TID_NAME)
	{	nNodeID = nextToken.GetDefinedName ();
		if (nNodeID < 0)
			nextToken.Error ("Undefined node name");
	}
	else
		nextToken.Error ("Unexpected symbol");
	nextToken.Scan ();
	return nNodeID;
}

/*-----------------------------------------------------------------
Encode a route (after encoding the DEF part
-----------------------------------------------------------------*/
void BIFSEncoder :: EncodeRouteDetails ()
{
	for (BOOL bIn = FALSE; bIn <= TRUE; ++bIn)
	{
		MediaObject *pNode = ScanNode ();
		if (pNode != NULL)
			pNode -> EncodeRoute (this, bIn);
		delete pNode;
		if (!bIn)
		{	if (nextToken == TID_TO)
				nextToken.Scan ();
			else
				nextToken.Error ("TO expected");
		}
	}
}

void BIFSEncoder :: SceneReplaceCommand ()
{
	nextToken.Scan ();
	if (nextToken != TID_BY)
		nextToken.Error ("BY expected");
	else
		nextToken.Scan ();
	EncodeInt (3, 2, "SceneReplaceCommand");
	EncodeScene ();
}

/*-----------------------------------------------------------------
Encode replace route command
-----------------------------------------------------------------*/
void BIFSEncoder :: ReplaceRoute ()
{	nextToken.Scan ();
	EncodeInt (3, 2, "RouteReplacement");
	EncodeRouteID ();
	if (nextToken == TID_BY)
		nextToken.Scan ();
	else
		nextToken.Error ("BY expected");
	EncodeRouteDetails ();
}

/*-----------------------------------------------------------------
Encode an initial scene (or scene replacement
-----------------------------------------------------------------*/
void BIFSEncoder :: EncodeScene ()
{
	SFRootNode node;
	
	EncodeInt (0, 1, "hasWorldInfo");
	node.Encode (this);
	BOOL bHasRoutes = nextToken == TID_ROUTE;
	EncodeInt (bHasRoutes, 1, "hasRoutes");
	if (bHasRoutes)
		EncodeRoutesDescription ();
	
	BOOL bHasAssociations =
		nextToken == TID_DEF || nextToken == TID_NAME;
	EncodeInt (bHasAssociations, 1, "hasSessionStreamAssociation");
	if (bHasAssociations)
		SFNode<SessionStreamAssociation> ().Encode (this);
}
