/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Encoder
FILE:       EncNode.cpp
AUTHOR:     Zvi Lifshitz (VDOnet Corp.)
OVERVIEW:   Encoding of a MediaObject object.
-------------------------------------------------------------------------*/
#include "defs.h"
#include "MediaObj.h"
#include "Token.h"
#include "BifsEnc.h"
#include "bifs.h"
#include "Quantize.h"
#include "NodeFld.h"

/*-----------------------------------------------------------------
Parse a MediaObject by using the field table.
-----------------------------------------------------------------*/
BOOL MediaObject :: Encode (BIFSEncoder *pBifs)
{
	pBifs -> EncodeInt (0, 1, "maskAccess");

	if (nextToken == TID_LEFT_BRACE)
	{	nextToken.Scan ();
		while (nextToken != TID_RIGHT_BRACE)
		{	
			/*-----------------------------------------------------------------
			Store no-end bit
			-----------------------------------------------------------------*/
			pBifs -> EncodeInt (0, 1, "endFlag");
			if (nextToken != TID_NAME)
			{	nextToken.Error ("Unexpected symbol");
				nextToken.Scan ();
			}
			else
			{	const BifsFieldTable *pEntry =
					GetBifsEntry (nextToken.GetName ());
				if (pEntry)
					pBifs -> EncodeInt (
						pEntry -> m_defID,
						GetBifsBounds () -> m_nBitsPerDef,
						"fieldReference"
					);
				else
					nextToken.Error ("Unrecognized field name");
				nextToken.Scan ();
				EncodeField (pBifs, pEntry);
			}
		}
		nextToken.Scan ();
	}
	pBifs -> EncodeInt (1, 1, "endFlag");
	return TRUE;
}

/*-----------------------------------------------------------------
Encode a node field, when a pointer to the BIFS table entry describing
that field is given
-----------------------------------------------------------------*/
BOOL MediaObject :: EncodeField (
	BIFSEncoder *pBifs,
	const BifsFieldTable *pEntry
)
{	
	if (pEntry == NULL)
		return FALSE;
	Quantizer quant (
		pEntry -> m_fMin,  
		pEntry -> m_fMax,  
		pEntry -> m_nDimension, 
		pEntry -> m_nQuantizationType
	);

	BOOL bResult = GetField (pEntry) -> Encode (pBifs, &quant);
	return bResult;
}

/*-----------------------------------------------------------------
Returns a pointer to the BIFS table entry for the field whose name is
szName.
-----------------------------------------------------------------*/
const BifsFieldTable * MediaObject :: GetBifsEntry (LPCSTR szName)
{
	const BifsFieldTable *pTable = GetFieldTable ();
	for (int i = GetFieldCount (); --i >= 0;)
		if (pTable [i].m_szName && !stricmp (szName, pTable [i].m_szName))
			return pTable + i;
	return NULL;
}

/*-----------------------------------------------------------------
Encode a Route whose departure node is this node
-----------------------------------------------------------------*/
void MediaObject :: EncodeRoute (BIFSEncoder *pBifs, BOOL bIn)
{
	pBifs -> EncodeInt (m_nID, 10, "NodeID");
	if (nextToken != TID_PERIOD)
		nextToken.Error (". expected");
	else
		nextToken.Scan ();

	if (nextToken != TID_NAME)
		nextToken.Error ("Unexpected symbol");
	else
	{	const BifsFieldTable *pEntry =
			GetBifsEntry (nextToken.GetName ());
		if (pEntry)
			pBifs -> EncodeInt (
				bIn ? pEntry -> m_inID : pEntry -> m_outID ,
				bIn ?
					GetBifsBounds () -> m_nBitsPerIn :
					GetBifsBounds () -> m_nBitsPerOut,
				bIn ? "inFieldRefernce" : "outFieldReference"
			);
		else
			nextToken.Error ("Unrecognized field name");
	}
	nextToken.Scan ();
}

/*-----------------------------------------------------------------
Encode an insert command. If bAppend is TRUE, a value is inserted at
the end of an MF field, otherwise an index is given.
-----------------------------------------------------------------*/
void MediaObject :: InsertToField (BIFSEncoder *pBifs, BOOL bAppend)
{
	int nIndex;
	const BifsFieldTable *pEntry = ParseMFElement (bAppend, nIndex);
	if (!bAppend && nIndex < 0)
		nextToken.Error ("[ exptected");

	/*-----------------------------------------------------------------
	Encode the command
	-----------------------------------------------------------------*/
	if (pEntry)
	{	
		if (pEntry -> m_bChildren)
		{	pBifs -> EncodeInt (0, 2, "NodeInsertion");
			pBifs -> EncodeInt (m_nID, 10, "NodeID");
		}
		else
		{	pBifs -> EncodeInt (1, 2, "IndexedValueInsertion");
			pBifs -> EncodeInt (m_nID, 10, "NodeID");
			pBifs -> EncodeInt (
				pEntry -> m_inID,
				GetBifsBounds () -> m_nBitsPerIn,
				"fieldID"
			);
		}
		pBifs -> EncodeInt (
			nIndex > 0 ? 0 : nIndex == 0 ? 1 : 2,
			2,
			"insertionPosition"
		);
		if (nIndex > 0)
			pBifs -> EncodeInt (nIndex, 16, "position");
		EncodeFieldElement (pBifs, pEntry);
	}
}

/*-----------------------------------------------------------------
Encode an element in an MF node field, when a pointer to the BIFS table
entry describing that field is given
-----------------------------------------------------------------*/
void MediaObject :: EncodeFieldElement (
	BIFSEncoder *pBifs,
	const BifsFieldTable *pEntry
)
{	
	Quantizer quant (
		pEntry -> m_fMin,  
		pEntry -> m_fMax,  
		pEntry -> m_nDimension, 
		pEntry -> m_nQuantizationType
	);

	GetField (pEntry) -> EncodeElement (pBifs, &quant);
}

/*-----------------------------------------------------------------
Encode an IndexedValueDeletion command. If bLast is TRUE, this is DELETE
LAST ...
-----------------------------------------------------------------*/
void MediaObject :: IndexedValueDeletion (BIFSEncoder *pBifs, BOOL bLast)
{
	int nIndex;
	const BifsFieldTable *pEntry = ParseMFElement (bLast, nIndex);
	if (!bLast && nIndex < 0)
		nextToken.Error ("[ exptected");

	if (pEntry)
	{	pBifs -> EncodeInt (1, 2, "IndexedValueDeletion");
		pBifs -> EncodeInt (m_nID, 10, "NodeID");
		pBifs -> EncodeInt (
			pEntry -> m_inID,
			GetBifsBounds () -> m_nBitsPerIn,
			"fieldID"
		);
		pBifs -> EncodeInt (
			nIndex > 0 ? 0 : nIndex == 0 ? 1 : 2,
			2,
			"deletionPosition"
		);
		if (nIndex > 0)
			pBifs -> EncodeInt (nIndex, 16, "position");
	}
}

/*-----------------------------------------------------------------
Parse MField Element. Returns the index in rIndex
-----------------------------------------------------------------*/
const BifsFieldTable *MediaObject :: ParseMFElement (BOOL bLast, int &rIndex)
{
	const BifsFieldTable *pEntry = NULL;

	if (nextToken != TID_PERIOD)
		nextToken.Error (". expected");
	else
		nextToken.Scan ();

	if (nextToken != TID_NAME)
	{	nextToken.Error ("Unexpected symbol");
		nextToken.Scan ();
	}	
	else
	{	pEntry = GetBifsEntry (nextToken.GetName ());
		if (!pEntry)
			nextToken.Error ("Unrecognized field name");
		nextToken.Scan ();

		rIndex = -1;
		/*-----------------------------------------------------------------
		If bAppend is FALSE, field name is followed by an index
		-----------------------------------------------------------------*/
		if (!bLast && nextToken == TID_LEFT_BRACKET)
		{	rIndex = 0;
			nextToken.Scan ();
			if (nextToken != TID_INTEGER)
				nextToken.Error ("Integer index value expected");
			rIndex = nextToken.GetInt ();
			if (nextToken != TID_RIGHT_BRACKET)
				nextToken.Error ("] expected");
			nextToken.Scan ();
		}
	}
	return pEntry;
}

/*-----------------------------------------------------------------
Replace a field or an indexed value
-----------------------------------------------------------------*/
void MediaObject :: FieldReplacement (BIFSEncoder *pBifs, BOOL bLast)
{
	int nIndex;
	const BifsFieldTable *pEntry = ParseMFElement (bLast, nIndex);
	
	if (nextToken == TID_BY)
		nextToken.Scan ();
	else
		nextToken.Error ("BY expected");

	if (pEntry)
	{	
		/*-----------------------------------------------------------------
		Encode FieldReplacement
		-----------------------------------------------------------------*/
		if (!bLast && nIndex < 0)
		{	pBifs -> EncodeInt (1, 2, "FieldReplacement");
			pBifs -> EncodeInt (m_nID, 10, "NodeID");
			pBifs -> EncodeInt (
				pEntry -> m_inID,
				GetBifsBounds () -> m_nBitsPerIn,
				"fieldID"
			);
			EncodeField (pBifs, pEntry);
		}
		/*-----------------------------------------------------------------
		Encode IndexedValueReplacement
		-----------------------------------------------------------------*/
		else
		{	pBifs -> EncodeInt (2, 2, "IndexedValueReplacement");
			pBifs -> EncodeInt (m_nID, 10, "NodeID");
			pBifs -> EncodeInt (
				pEntry -> m_inID,
				GetBifsBounds () -> m_nBitsPerIn,
				"fieldID"
			);
			pBifs -> EncodeInt (
				nIndex > 0 ? 0 : nIndex == 0 ? 1 : 2,
				2,
				"replacementPosition"
			);
			if (nIndex > 0)
				pBifs -> EncodeInt (nIndex, 16, "position");
			EncodeFieldElement (pBifs, pEntry);
		}
	}
}