/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Encoder
FILE:       Token.cpp
AUTHOR:     Zvi Lifshitz (VDOnet Corp.)
OVERVIEW:   Implementation of Token, the class that handles text parsing.
-------------------------------------------------------------------------*/
#include "defs.h"
#include "Token.h"
#include "MediaObj.h"

/*-----------------------------------------------------------------
Initiate the Token object
-----------------------------------------------------------------*/
void Token :: Init (FILE *fInput, FILE *fList)
{
	m_fInput = fInput;
	m_fList = fList;
	m_nInputOffset = m_nInputLength = 0;
	m_cNext = GetChar ();
	Scan ();
}

/*-----------------------------------------------------------------
Translate the token to an int
-----------------------------------------------------------------*/
int Token :: GetInt ()
{
	int nResult = 0;

	if (m_code != TID_INTEGER)
		Error ("Integer number expected");
	else
		nResult = atoi (m_string);
	if (m_code == TID_INTEGER || m_code == TID_NUMBER)
		Scan ();
	return nResult;
}

/*-----------------------------------------------------------------
Translate TRUE/FALSE value
-----------------------------------------------------------------*/
BOOL Token :: GetBool ()
{
	BOOL bResult;

	switch (m_code) {
	case TID_TRUE:
		bResult = TRUE;
		break;
	case TID_FALSE:
		bResult = FALSE;
		break;
	case TID_NUMBER:
		Error ("Boolean expected");
		Scan ();
		return FALSE;
	default:
		Error ("Boolean expected");
		return FALSE;
	}
	Scan ();
	return bResult;
}

/*-----------------------------------------------------------------
Translate a real value
-----------------------------------------------------------------*/
double Token :: GetReal ()
{
	double dResult;

	if (m_code != TID_NUMBER && m_code != TID_INTEGER)
	{	Error ("Number expected");
		return 0;
	}

	dResult = atof (m_string);
	Scan ();
	return dResult;
}

/*-----------------------------------------------------------------
If quotes are not used, read all characters till a white space 
-----------------------------------------------------------------*/
LPCSTR Token :: GetString ()
{
	if (m_code != TID_STRING)
		while (
			m_cNext != ' ' && m_cNext != '\t' && m_cNext != '\n' &&
			m_cNext != EOF
		)
		{	m_string += m_cNext;
			m_cNext = GetChar ();
		}
	return m_string;
}
	
/*-----------------------------------------------------------------
Scan a symbol and find its code
-----------------------------------------------------------------*/
void Token :: Scan ()
{
	if (m_code == TID_EOF)
		throw ExceptionEndOfFile ();

	do
	{	/*-----------------------------------------------------------------
		Skip all white spaces and comments
		-----------------------------------------------------------------*/
		while (
			m_cNext == ' ' || m_cNext == '\t' || m_cNext == '\n' ||
			m_cNext == '\r' || m_cNext == '#'
		)
		{
			// Handle a comment, # to the end of the line
			if (m_cNext == '#')
			{
				while ((m_cNext != '\n') && (m_cNext != EOF))
					m_cNext = GetChar ();
			}
			m_cNext = GetChar ();
		}
		if (m_cNext == EOF)
			break;

		m_string = "";
		m_code = TID_UNKNOWN;

		/*-----------------------------------------------------------------
		Parse a name
		-----------------------------------------------------------------*/
		if (isalpha (m_cNext) || m_cNext == '_' || m_cNext == '$')
		{	while (
				isalnum (m_cNext) || m_cNext == '_' || m_cNext == '$' ||
				m_cNext == '-'
			)
			{	m_string += m_cNext;
				m_cNext = GetChar ();
			}
			m_code = TID_NAME;
		}

		/*-----------------------------------------------------------------
		Parse a number
		-----------------------------------------------------------------*/
		else if (isdigit (m_cNext) || m_cNext == '-' || m_cNext == '+')
		{	m_code = TID_INTEGER;
			while (
				isdigit (m_cNext) || m_cNext == '-' || m_cNext == '+' ||
				m_cNext == '.' || tolower (m_cNext) == 'e' ||
				tolower (m_cNext) == 'x'
			)
			{	m_string += m_cNext;
				if (m_cNext == '.' || tolower (m_cNext == 'e'))
					m_code = TID_NUMBER;
				m_cNext = GetChar ();
			}
		}

		/*-----------------------------------------------------------------
		Parse a string
		-----------------------------------------------------------------*/
		else if (m_cNext == '"')
		{	while ((m_cNext = GetChar ()) != '"')
			{	if (m_cNext == EOF)
					throw ExceptionEndOfFile ();
				m_string += m_cNext;
			}
			m_cNext = GetChar ();
			m_code = TID_STRING;
		}

		/*-----------------------------------------------------------------
		Parse one character
		-----------------------------------------------------------------*/
		else
		{	m_string += m_cNext;
			m_cNext = GetChar ();
		}

		/*-----------------------------------------------------------------
		Check if a reserved symbol
		-----------------------------------------------------------------*/
		for (int i = TID_LEFT_BRACE; i < TID_NUMBER_OF_RESERVED; i++)
			if (!stricmp (m_string, m_reserved [i]))
			{	m_code = (TOKEN_CODE) i;
				return;
			}
		
		if (m_code != TID_UNKNOWN)
			return;
		Error ("Invalid symbol");
	}
	while (m_cNext != EOF);
	m_code = TID_EOF;
}

/*-----------------------------------------------------------------
Find a name in a list of defined name
-----------------------------------------------------------------*/
int Token :: GetDefinedName () const
{
	for (int i = m_defined.GetLength (); --i >= 0;)
		if (m_string == m_defined [i].m_name)
			return i;
	return -1;
}

/*-----------------------------------------------------------------
Find a name in a list of defined route
-----------------------------------------------------------------*/
int Token :: GetDefinedRoute () const
{
	for (int i = m_routes.GetLength (); --i >= 0;)
		if (m_string == m_routes [i])
			return i;
	return -1;
}

/*-----------------------------------------------------------------
Display an error message in the list file
-----------------------------------------------------------------*/
void Token :: Error (LPCSTR szDesc)
{
	fprintf (m_fList, "ERROR: %s - %s\n", (LPCSTR) m_string, szDesc);
}

/*-----------------------------------------------------------------
Add a name to the list of predefined name
-----------------------------------------------------------------*/
int Token :: AddPredefinedName ()
{
	int nResult = m_defined.GetLength ();
	(++m_defined).m_name = m_string;
	return nResult;
}

/*-----------------------------------------------------------------
Add a name to the list of predefined routes
-----------------------------------------------------------------*/
int Token :: AddPredefinedRoute ()
{
	int nResult = m_routes.GetLength ();
	m_routes += m_string;
	return nResult;
}

/*-----------------------------------------------------------------
Read one character from the input file. Each input line is copied to the
list file
-----------------------------------------------------------------*/
int Token :: GetChar ()
{
	if (m_nInputOffset >= m_nInputLength)
	{	if (feof (m_fInput))
			return EOF;
		if (!fgets (m_szLine, sizeof (m_szLine), m_fInput))
			return EOF;
		m_nInputLength = strlen (m_szLine);
		m_nInputOffset = 0;
		fputs (m_szLine, m_fList);
	}
	return m_szLine [m_nInputOffset++];
}

/*-----------------------------------------------------------------
Define the list of reserved tokens
-----------------------------------------------------------------*/
LPCSTR Token :: m_reserved [] = 
{
	"",
	"{",
	"}",
	"[",
	"]",
	",",
	".",
	"",
	"USE",
	"DEF",
	"FALSE",
	"TRUE",
	"ROUTE",
	"TO",
	"AT",
	"INSERT",
	"REPLACE",
	"DELETE",
	"APPEND",
	"LAST",
	"SCENE",
	"BY"
};		
