///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File: AudRend-CSELT-DS.cpp
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of a class which provides a DirectSound based
//   audio renderer.
//
// Created: August, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "AudRend-CSELT-DS.h"
#include "AR-DS-StreamInfo.h"
#include "AudioSourceProxy.h"
#include "Caps.h"
#include "Executiv.h"
#include "Presenter-CSELT.h"
#include "Clock.h"
#include "dx-errors.h"
#include "trace.h"
#ifndef MAKE_TEST_APP
#include "MPEG4Player.h"
#include "FatalError.h"
#include "ThreadList.h"
#endif


// =====================================================================

//
// AudioRendererDS::AudioRendererDS
//
// Construct the audio renderer, given a pointer to the current
// Presenter object, and the handle of the window the application
// is running in.
//

AudioRendererDS::AudioRendererDS(Presenter* pPresenter, HWND hWnd)
{
	char errorString[200];
	errorString[0] = 0;

	TRACE_PRINT("Entering AudioRendererDS::AudioRendererDS()\n");

	// Initialise data members
	m_bActive = FALSE;
	m_pPresenter = pPresenter;

	// Create the DirectSound object, the interface to the sound card
	HRESULT hResult = DirectSoundCreate(NULL, &m_pDS, NULL);
	if (hResult != DS_OK)
		sprintf (errorString, "DirectSoundCreate() failed.\n  %s",
				 HRESULTtoText(hResult));

	// Set the cooperative level
	if (errorString[0] == 0)
	{
		hResult = m_pDS->SetCooperativeLevel(hWnd, DSSCL_PRIORITY);
		if (hResult != DS_OK)
			sprintf (errorString, "SetCooperativeLevel() failed.\n  %s",
					 HRESULTtoText(hResult));
	}

	// Initialise the audio format output by the sound card
	memset(&m_wFmt, 0, sizeof(WAVEFORMATEX));
	m_wFmt.wFormatTag	  = WAVE_FORMAT_PCM;
	m_wFmt.nChannels      = 0;
	m_wFmt.nSamplesPerSec = 0;
	m_wFmt.wBitsPerSample = 0;
	m_wFmt.cbSize		  = 0;

	// If an error has occurred, display a message and then exit
	if (errorString[0] != 0)
	{
#ifndef MAKE_TEST_APP
		CFatalError fatalErrorDialog;
		fatalErrorDialog.m_message = errorString;
		fatalErrorDialog.DoModal();
#endif
		m_initOK = FALSE;
	}
	else
		m_initOK = TRUE;

	TRACE_PRINT("Leaving AudioRendererDS::AudioRendererDS()\n");
} /* AudioRendererDS::AudioRendererDS */


// =====================================================================

//
// AudioRendererDS::~AudioRendererDS
//
// Class destructor.
//

AudioRendererDS::~AudioRendererDS()
{
	TRACE_PRINT("Entering AudioRendererDS::~AudioRendererDS():\n");

	// Make sure that the renderer is first terminated
	Terminate();

	// Release the main direct sound object
	if (m_pDS != NULL)
		m_pDS->Release();

	TRACE_PRINT("Leaving AudioRendererDS::~AudioRendererDS():\n");
} /* AudioRendererDS::~AudioRendererDS */


// =====================================================================

//
// AudioRendererDS::GetCapabilities
//
// Get the capabilities of this audio renderer.
//

void AudioRendererDS::GetCapabilities (Capabilities* pCaps)
{
#ifdef MAKE_TEST_APP
	pCaps;
#else
	pCaps->AddFormatCapability(FORMAT_PCM, 2);
#endif
} /* AudioRendererDS::GetCapabilities */


// =====================================================================

//
// AudioRendererDS::Init
//
// Initialise the renderer.  The thread is started in Start(), so there
// is nothing to do here.  This method will be called once for every
// file that is opened.
//

void AudioRendererDS::Init()
{
} /* AudioRendererDS::Init */


// =====================================================================

//
// AudioRendererDS::Terminate
//
// Stops rendering and performs cleanups.  This method will be called
// once for every file that is closed.
//

void AudioRendererDS::Terminate()
{
	TRACE_PRINT("Entering AudioRendererDS::Terminate():\n");

	// Flag that the audio renderer is no longer active
	m_bActive = FALSE;

	// Wait for this thread to terminate
	m_mtxThread.P ();
	m_mtxThread.V ();

	TRACE_PRINT("Leaving AudioRendererDS::Terminate():\n");
} /* AudioRendererDS::Terminate */


// =====================================================================

//
// AudioRendererDS::Start
//
// Starts the audio renderer running in a separate thread.  The clock
// will be started by the thread itself.
//

void AudioRendererDS::Start()
{

	// Flag that the audio renderer is now active
	m_bActive = TRUE;

	// Start the main thread of the audio renderer
	ZThread<AudioRendererDS> thread(Main, this, THREAD_PRIORITY_ABOVE_NORMAL);
	int hThread = thread.Begin();
  	if (hThread == -1)
    	throw;
#ifndef MAKE_TEST_APP
	else
		AddToThreadList("AudioRenderer", hThread);
#endif
} /* AudioRendererDS::Start */


// =====================================================================

//
// AudioRendererDS::UpdateOutputFormat
//
// Given a stream which is being added to the list of streams to play,
// update the output format that the sound card is currently using.
//
// The sound card will be set to the maximum sample rate, bits per sample
// and number of channels of all of the input streams.  This means that
// DirectSound will be doing the sample rate conversion if required.
// This has been known to be a problem with some sounds cards, and also
// under WindowsNT, where this is done in software.
//

void AudioRendererDS::UpdateOutputFormat(StreamInfo* pStream)
{
	// Update the sound board settings
	if (m_wFmt.nSamplesPerSec < pStream->m_wf.nSamplesPerSec)
		m_wFmt.nSamplesPerSec = pStream->m_wf.nSamplesPerSec;
	if (m_wFmt.wBitsPerSample < pStream->m_wf.wBitsPerSample)
		m_wFmt.wBitsPerSample = pStream->m_wf.wBitsPerSample;
	if (m_wFmt.nChannels < pStream->m_wf.nChannels)
		m_wFmt.nChannels = pStream->m_wf.nChannels;

	// Set the output format of the sound card
	DSBUFFERDESC primaryDesc;
	memset(&primaryDesc, 0, sizeof(DSBUFFERDESC));
	primaryDesc.dwSize = sizeof(DSBUFFERDESC);
	primaryDesc.dwFlags = DSBCAPS_PRIMARYBUFFER;

	LPDIRECTSOUNDBUFFER primarySB;
	HRESULT hr = m_pDS->CreateSoundBuffer(&primaryDesc, &primarySB, NULL);
	assert(hr == DS_OK);

	m_wFmt.nBlockAlign	   = m_wFmt.nChannels * m_wFmt.wBitsPerSample / 8;
	m_wFmt.nAvgBytesPerSec = m_wFmt.nSamplesPerSec * m_wFmt.nBlockAlign;
	hr = primarySB->SetFormat(&m_wFmt);
	assert(hr == DS_OK);

	// Release the primary buffer as we have finished with it.
	primarySB->Release();

#ifdef _DEBUG
//	printf("Setting the output format to: %d Hz, %d bits/sample, %d channels.\n",
//		   m_wFmt.nSamplesPerSec, m_wFmt.wBitsPerSample, m_wFmt.nChannels);
#endif
} /* AudioRendererDS::UpdateOutputFormat */


// =====================================================================

//
// AudioRendererDS::AddSourceNode
//
// Add the given AudioSourceProxy node to the list of audio source nodes
// which need to be played.  If a node is already in the list, then
// just leave it there.  The time at which the audio is to start
// playing is also given.
//
// It is intended that this method will be called once the audio node
// has been parsed.
//
// The audio renderer will call AudioSource::GetFormat() to ask the
// node of the audio data format, once Init() has been called.
//

void AudioRendererDS::AddSourceNode(AudioSourceProxy* pSourceNode,	
									DWORD startTime,
									DWORD nSamplesPerSec,
									DWORD nBitsPerSample,
									WORD nNumChannels)
{
	assert(pSourceNode != NULL);
	assert(nSamplesPerSec > 0);
	assert(nBitsPerSample > 0);
	assert(nNumChannels > 0);

	// Need to lock the array before accessing to it
	m_mtxStreamArray.P();

	// If this node is already in the array, then return
	for (int i = 0; i < m_streams.GetLength(); i++)
		if (m_streams[i]->m_pSrc == pSourceNode)
		{
			m_mtxStreamArray.V();
			return;
		}

	// Create a stream object for this audio source node
	StreamInfo* pStream = new StreamInfo(m_pDS, pSourceNode, startTime);
	int result = pStream->Create(nSamplesPerSec, nBitsPerSample,
								 nNumChannels);
	assert(result == DS_OK);

	// Update the audio output format of the sound card
	UpdateOutputFormat(pStream);

	// Add this to the array of streams
	m_streams += pStream;

	// Start the stream playing
	pStream->Play();

	// And now unlock it so that the main audio renderer thread can
	// access it.
	m_mtxStreamArray.V();
} /* AudioRendererDS::AddSourceNode */


// =====================================================================

//
// AudioRendererDS::RemoveSourceNode
//
// Remove the given AudioSourceProxy node from the list of audio source
// nodes which need to be played.  If the given node is not in the
// list, nothing happens.
//

void AudioRendererDS::RemoveSourceNode(AudioSourceProxy* pSourceNode)
{
	// Need to lock the array before accessing to it
	m_mtxStreamArray.P();

	// Find this node in the array
	for (int i = 0; i < m_streams.GetLength(); i++)
		if (m_streams[i]->m_pSrc == pSourceNode)
		{
			// Remove this node from the array
			StreamInfo* pStream = m_streams[i];
			m_streams -= pStream;

			// Close down this stream
			delete pStream;
		}

	// And now unlock it so that the main audio renderer thread can
	// access it.
	m_mtxStreamArray.V();

} /* AudioRendererDS::RemoveSourceNode */


// =====================================================================

//
// AudioRendererDS::GetDelay
//
// Returns the delay, in milliseconds, that the audio renderer has.
// This delay is the average time that it takes samples taken from
// the AudioSourceNode to be played.
//

DWORD AudioRendererDS::GetDelay()
{
	return BUFFER_LENGTH;
} /* AudioRendererDS::GetDelay */


// =====================================================================

//
// AudioRendererDS::Main
//
// This is the main loop of the audio renderer thread.
//

void AudioRendererDS::Main()
{
	// Mark the thread as running
	m_mtxThread.P();

	unsigned long startingTime = GetTickCount();
	unsigned long updatesDone = 0;

	// This loop runs while the audio renderer is active
	// ie. up to the next Terminate() call.  Every BUFFER_UPDATE_TIME ms
	// the buffers will be updated.
	while (m_bActive)
	{
		// Lock the stream information data
		m_mtxStreamArray.P();

		// Read data from each of the active buffers
		TRACE_PRINT1("AR: Using %d streams.\n", m_streams.GetLength());
		for (int i = 0; i < m_streams.GetLength(); i++)
		{
			DWORD currentTime = GetTickCount() - startingTime;

			// Only get data for streams whose start time has arrived
			if (currentTime >= m_streams[i]->m_startTime)
			{
				// Find out how much space is available in this buffer
				DWORD bytesToWrite = m_streams[i]->GetMaxWriteSize();

				// Repeatedly read data for this buffer, until either the
				// buffer is full, or there is no more data, in which case
				// silence needs to be inserted.
				DWORD bytesRead = 1;
				while ((bytesToWrite > 0) && (bytesRead > 0))
				{
					// Read data from the audio node
					LPBYTE pData;
					m_streams[i]->m_pSrc->GetData(bytesToWrite,
												  pData, bytesRead);
					assert(bytesRead <= bytesToWrite);

//					TRACE_PRINT2("AR: Writing %d of %d bytes.\n",
//								 bytesRead, bytesToWrite);

					if (bytesRead > 0)
					{
						// Write this data to the stream
						m_streams[i]->WriteData(pData, bytesRead);

						// Release this data
						m_streams[i]->m_pSrc->DataUsed(pData, bytesRead);

						// Update how much data is left to get
						bytesToWrite -= bytesRead;
					}
				}

				// If the buffer was not completely filled, we need to
				// fill the rest of the buffer with silence.  This is needed
				// because of the circular buffer used by DirectSound.
				if (bytesToWrite > 0)
					m_streams[i]->WriteSilence(bytesToWrite);

				// Start the stream playing if it is not already playing
				if (!m_streams[i]->IsPlaying())
				{
					m_streams[i]->Play();
					TRACE_PRINT("AR: Starting stream playing.\n");
				}
			}
		}

		// Unlock the stream information data
		m_mtxStreamArray.V();

		// Wait until it is time for the next buffer update
		updatesDone++;
		unsigned int elapsedTime = GetTickCount() - startingTime;
		unsigned int startNextTime = updatesDone * BUFFER_UPDATE_TIME;

		// Only sleep if ahead of schedule
		TRACE_PRINT1("AR: Sleeping for %d ms.\n", startNextTime - elapsedTime);
		if (startNextTime > elapsedTime)
			Sleep(startNextTime - elapsedTime);

		// Start the clock running after 3 loops.  This allows time for
		// the streams to attach themselves, and also for the delay
		// that DirectSound appears to have in starting playing.
		// There should possibly be a better way to do this.
		if (updatesDone == 3)
		{
			TRACE_PRINT("AR: Starting the clock running.\n");
#ifndef MAKE_TEST_APP
			m_pPresenter->GetExecutive()->GetClock()->Start();
#endif
		}
	}

	// Flag that the thread is terminating
	m_mtxThread.V();
} /* AudioRendererDS::Main */


// =====================================================================
