///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File: Copy.h
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The definition of the functions which copy RGBA and YUVA images
//   to the offscreen RGB surfaces in the visual renderer.
//
// Created: April, 1997
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#ifndef COPY_H
#define COPY_H


// This structure contains all of the parameters which specify how an
// image is copied into a surface.
typedef struct
{
	// The top-left point in the source image to start copying
	unsigned int srcStartX;
	unsigned int srcStartY;

	// The dimensions of the portion to be copied
	unsigned int copyWidth;
	unsigned int copyHeight;

	// The top-left point in the destination image to copy to
	unsigned int destStartX;
	unsigned int destStartY;
} CopyParameters;


BOOL SetUpCopyParameters(CopyParameters& cpParams,
						 SrcImage const& srcImage,
						 SurfaceDesc const& surfaceDesc);

void CopyRGBAToRGB555(SrcRGBAImage const& srcImage,
					  SurfaceDesc const& surfaceDesc);
void CopyRGBAToRGB565(SrcRGBAImage const& srcImage,
					  SurfaceDesc const& surfaceDesc);
void CopyRGBAToRGB24bpp(SrcRGBAImage const& srcImage,
						SurfaceDesc const& surfaceDesc);
void CopyRGBAToRGB32bpp(SrcRGBAImage const& srcImage,
						SurfaceDesc const& surfaceDesc);
void CopyRGBAToBGR32bpp(SrcRGBAImage const& srcImage,
						SurfaceDesc const& surfaceDesc);

void CopyYUVToRGB555(SrcYUVAImage const& srcImage,
					 SurfaceDesc const& surfaceDesc);
void CopyYUVToRGB565(SrcYUVAImage const& srcImage,
					 SurfaceDesc const& surfaceDesc);
void CopyYUVToRGB24bpp(SrcYUVAImage const& srcImage,
					   SurfaceDesc const& surfaceDesc);
void CopyYUVToRGB32bpp(SrcYUVAImage const& srcImage,
					   SurfaceDesc const& surfaceDesc);
void CopyYUVToBGR32bpp(SrcYUVAImage const& srcImage,
					   SurfaceDesc const& surfaceDesc);

void FillRGB555(BRect const& drawBRect, SurfaceDesc const& surfaceDesc,
				Color3f const& pixelColor, float transparency);
void FillRGB565(BRect const& drawBRect, SurfaceDesc const& surfaceDesc,
				Color3f const& pixelColor, float transparency);
void FillRGB24bpp(BRect const& drawBRect, SurfaceDesc const& surfaceDesc,
				  Color3f const& pixelColor, float transparency);
void FillRGB32bpp(BRect const& drawBRect, SurfaceDesc const& surfaceDesc,
				  Color3f const& pixelColor, float transparency);
void FillBGR32bpp(BRect const& drawBRect, SurfaceDesc const& surfaceDesc,
				  Color3f const& pixelColor, float transparency);


// ============================================================================

//
// ApplyAlphaMaskToRGB
//
// This macro combines the old and new RGB values, using the given alpha
// value.
//

#define ApplyAlphaMaskToRGB(red, green, blue, alpha, oldRed, oldGreen, oldBlue) \
if (alpha == 0) \
{ \
	red = oldRed; \
	green = oldGreen; \
	blue = oldBlue; \
} \
else \
{ \
	red   = (red   * alpha + oldRed   * (256 - alpha)) >> 8; \
	green = (green * alpha + oldGreen * (256 - alpha)) >> 8; \
	blue  = (blue  * alpha + oldBlue  * (256 - alpha)) >> 8; \
}


// ============================================================================

//
// ApplyTransparencyToRGB
//
// This macro combines the old and new RGB values, using the given
// transparency value.
//

#define ApplyTransparencyToRGB(red, green, blue, T, oldRed, oldGreen, oldBlue) \
red   = (int)((float)red   * (1.0 - T) + (float)oldRed   * T + 0.5); \
green = (int)((float)green * (1.0 - T) + (float)oldGreen * T + 0.5); \
blue  = (int)((float)blue  * (1.0 - T) + (float)oldBlue  * T + 0.5)


// ============================================================================

//
// MAKE_RGB555_PIXEL
//
// Combine red, green and blue values (each stored in the least
// significant 8 bits of an integer), into the least significant
// 15 bits of an unsigned short integer, using the most significant
// 5 bits of each colour.
//

#define MAKE_RGB555_PIXEL(red, green, blue) \
	(unsigned short)(((red & 0xf8) << 7) | \
					 ((green & 0xf8) << 2) | \
					 ((blue & 0xf8) >> 3))


// ============================================================================

//
// MAKE_RGB565_PIXEL
//
// Combine red, green and blue values (each stored in the least
// significant 8 bits of an integer), into an unsigned short integer,
// using the most significant 5 bits of red, 6 bits of green, and 5 bits
// of blue.
//

#define MAKE_RGB565_PIXEL(red, green, blue) \
	(unsigned short)(((red & 0xf8) << 8) | \
					 ((green & 0xfc) << 3) | \
					 ((blue & 0xf8) >> 3))


// ============================================================================

//
// MAKE_RGB32bpp_PIXEL
//
// Combine red, green and blue values (each stored in the least
// significant 8 bits of an integer), into the least significant
// 24 bits of an unsigned integer, using the most significant 8 bits
// of each colour.
//

#define MAKE_RGB32bpp_PIXEL(red, green, blue) \
	(red << 16) | (green << 8) | blue


// ============================================================================

//
// MAKE_BGR32bpp_PIXEL
//
// Combine red, green and blue values (each stored in the least
// significant 8 bits of an integer), into the least significant
// 24 bits of an unsigned integer, using the most significant 8 bits
// of each colour.
//

#define MAKE_BGR32bpp_PIXEL(red, green, blue) \
	(blue << 16) | (green << 8) | red


// ============================================================================


#endif // COPY_H


// Emacs uses these...
// Local Variables:
// mode: c++
// tab-width: 4
// End:
