///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File: DaiSession.cpp
//
// Authors: Nuno Delgado, CSELT
//
// Description:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "DaiSession.h"
#include "DmifPeer.h"



/***************************************************************************************

***************************************************************************************/
DaiSession::DaiSession(Dai* pDai)
{
	m_pDai = pDai;

	pHeadListOfPeersElement = NULL;

	m_ThreadCreated = FALSE;

	// Create the control event
	m_hEndThreadEvent = (WSAEVENT)CreateEvent(NULL, FALSE, FALSE, NULL);
	m_hRefreshTableEvent = (WSAEVENT)CreateEvent(NULL, FALSE, FALSE, NULL);


	// Critical section to control the access to the linked list of peers
	InitializeCriticalSection(&mPeersListCriticalSection);
	InitializeCriticalSection(&mAddAndRemoveDataChannelsCriticalSection);


}

/***************************************************************************************

***************************************************************************************/
DaiSession::~DaiSession()
{
	// Lets signal the event that will close the Thread of data exchange
	SetEvent(m_hEndThreadEvent);

	// Send a message to all peers indicating the end of the session
	SignalAllPeersOfSessionDestroy();

	// Delete all Peer objects and the elements in the linked list
	RemoveAndDestroyAllPeersInList();

	// Delete critical section
	DeleteCriticalSection(&mPeersListCriticalSection);
	DeleteCriticalSection(&mAddAndRemoveDataChannelsCriticalSection);
	
	CloseHandle((HANDLE)m_hEndThreadEvent);
	CloseHandle((HANDLE)m_hRefreshTableEvent);
}



/***************************************************************************************

  FindPeerInList

  This method searches for a Peer in the linked lists based on the IpAddress.
  If the element is found, TRUE value is returned and the pointer to thie element
  (pPeer) is filled.

***************************************************************************************/
BOOL DaiSession::FindPeerInList(char* pAddress, CDmifPeer** pPeer)
{
	ListOfPeersElement* pActual;
	BOOL ReturnValue = FALSE;

	EnterCriticalSection(&mPeersListCriticalSection);

	if(pHeadListOfPeersElement == NULL) // List is empty
		ReturnValue = FALSE;
	else
	{
		pActual = pHeadListOfPeersElement;
		do
		{
			if(strcmp((pActual->pDmifPeer)->m_pDmifPeerAddress,pAddress)==0)
			{
				*pPeer = (pActual->pDmifPeer);
				ReturnValue = TRUE;
				break;
			}
			else
			{
				if(pActual->pNextElement != NULL)
					pActual = pActual->pNextElement;
				else
				{
					ReturnValue = FALSE;
					break;
				}
			}

		}while(TRUE);

	}

	LeaveCriticalSection(&mPeersListCriticalSection);

	return ReturnValue;
}



/***************************************************************************************

  	FillTableOfTags

	This method fills the columns TransMuxTag, ChannelAssociationTag and FlexMuxChannel
	number of the table "Table"

***************************************************************************************/
void DaiSession::FillTableOfTags(int (*Table)[6], int NumElements)
{
	int count1, count2;
	BOOL FlagNewTransMuxTagNeeded;
	int temp;

	
	// This first part of the function fills the table with the 
	// TransMuxAssociationTags and with the ChannelAssociationTags
	for(count1=0; count1 < NumElements; count1 ++)
	{
		FlagNewTransMuxTagNeeded = TRUE;

		for(count2=0; count2 < count1; count2 ++)
		{
			if(Table[count1][0] == Table[count2][0])   // No need for new TransMuxTag
			{
				Table[count1][1] = Table[count2][1];   // Give the same TransMux number
				FlagNewTransMuxTagNeeded = FALSE;
				break;
			}
		}

		if(FlagNewTransMuxTagNeeded)
		{
			// Ask for NewTag. I just send the QoS level and I will get the new tag
			Table[count1][1] = GetNewTransMuxTagNumber(Table[count1][0]);
		}

		// Lets give the ChannelAssociationTag
		Table[count1][3] = GetNewChannelAssociationTagNumber();

	}

	
	// This part of the function will fill the column related with  
	// FlexMuxChannel Numbers
	for(count1=0; count1 < NumElements; count1 ++)
	{
		temp = -1;

		for(count2=0; count2 < count1; count2 ++)
		{
			if(Table[count1][0] == Table[count2][0])
			{
				if(Table[count2][2] == -1)
					Table[count2][2] = 0;
		
				temp = Table[count2][2] + 1 ;
			}
		}

		Table[count1][2] = temp;
	}
}




/***************************************************************************************

	GetNewTransMuxTagNumber

	This method is used to attribute the Tag numbers for each TransMux. If the QoS level
	is lower than MAX_QOS_VALUE_FOR_TCP_CHANNEL is given a Tag for TCP channel. If the
	value is higher a Tag for an UDP channel is given.

***************************************************************************************/
int DaiSession::GetNewTransMuxTagNumber(unsigned int QoS)
{
	static TcpTagBaseNumber = 100;
	static UdpTagBaseNumber = 200;

	int NewTag;

	// I assume that all QoS between 0 - MAX_QOS_VALUE_FOR_TCP_CHANNEL are TCP requests
	// and all values above MAX_QOS_VALUE_FOR_TCP_CHANNEL are UDP
	
	if(QoS < MAX_QOS_VALUE_FOR_TCP_CHANNEL)  // TCP request
	{
		NewTag = TcpTagBaseNumber;
		TcpTagBaseNumber++;
	}
	else
	{
		NewTag = UdpTagBaseNumber;
		UdpTagBaseNumber++;
	}

	return(NewTag);
}





/***************************************************************************************
	
	GetNewChannelAssociationTagNumber

	This method returns the channel association tag number for a new channel

***************************************************************************************/
int DaiSession::GetNewChannelAssociationTagNumber()
{
	static ChannelAssociationTag = 0;
	int NewTag;

	NewTag = ChannelAssociationTag;
	ChannelAssociationTag++;

	return(NewTag);
}


/***************************************************************************************

	AddPeerToList

	This method adds a the address of this peer to the linked list of pointers to peers

***************************************************************************************/
void DaiSession::AddPeerToList(CDmifPeer* pNewDmifPeer)
{

	ListOfPeersElement* pAux;
	ListOfPeersElement* pNewElement;


	// First, fill the new element fields
	pNewElement = (ListOfPeersElement*) new ListOfPeersElement;
	pNewElement -> pDmifPeer = pNewDmifPeer;

	EnterCriticalSection(&mPeersListCriticalSection);

	if(pHeadListOfPeersElement == NULL)				// List its empty
		pHeadListOfPeersElement  = pNewElement;
	else											// Add at the tail of the list
	{
		pAux = pHeadListOfPeersElement;
		while(pAux -> pNextElement != NULL)
			pAux = pAux->pNextElement;

		pAux->pNextElement = pNewElement;
	}

	LeaveCriticalSection(&mPeersListCriticalSection);
}




/***************************************************************************************

	RemoveAndDestroyAllPeersInList

	This method removes all the pointers for peers from the linked list and then deletes
	the peer objects

***************************************************************************************/
void DaiSession::RemoveAndDestroyAllPeersInList()
{

	ListOfPeersElement* pActual;
	ListOfPeersElement* pNext;

	EnterCriticalSection(&mPeersListCriticalSection);
	
	if(pHeadListOfPeersElement != NULL)
	{
		pNext = pActual = pHeadListOfPeersElement;
		do
		{
			pActual = pNext;
			pNext = pActual->pNextElement;
			
			//Delete the CDmifPeer object
			delete(pActual-> pDmifPeer);
			// remove the linked list element
			delete pActual;

		}while(pNext != NULL);
	}

	LeaveCriticalSection(&mPeersListCriticalSection);
}



/***************************************************************************************

	RemoveAndDestroyPeerFromList

  	This method removes the pointer for this peer (if it exists) from the linked list 
	and then deletes the peer object

***************************************************************************************/
BOOL DaiSession::RemoveAndDestroyPeerFromList(char* pPeerAddress)
{

	BOOL ReturnValue = FALSE;
	ListOfPeersElement* pActual;
	ListOfPeersElement* pPrevious;


	EnterCriticalSection(&mPeersListCriticalSection);

	if(pHeadListOfPeersElement == NULL)			// The list is empty
		ReturnValue = FALSE;
	else
	{
		// The first element is a special case
		if(strcmp((pHeadListOfPeersElement->pDmifPeer)->m_pDmifPeerAddress,pPeerAddress) == 0)
		{
			pActual = pHeadListOfPeersElement;
			pHeadListOfPeersElement = pActual->pNextElement;
			delete (pActual->pDmifPeer);       // Delete the session element
			delete pActual;						// Delete the linked list element
			ReturnValue = TRUE;
		}
		else 		// Its not the first element
 		{
			pActual = pHeadListOfPeersElement;
			pPrevious = pHeadListOfPeersElement;
			do
			{
				if(strcmp((pActual->pDmifPeer)->m_pDmifPeerAddress,pPeerAddress) == 0)
				{
					pPrevious->pNextElement = pActual->pNextElement;
					delete (pActual->pDmifPeer);       // Delete the session element
					delete pActual;						// Delete the linked list element
					ReturnValue = TRUE;
					break;
				}
				else
				{
					if(pActual->pNextElement != NULL)
					{
						pPrevious = pActual;
						pActual = pActual->pNextElement;
					}
					else
					{
						ReturnValue = FALSE;  // The element doesn't exist
						break;
					}

				}

			}while(TRUE);
		}
	}

	LeaveCriticalSection(&mPeersListCriticalSection);

	return ReturnValue;

}


/***************************************************************************************
	
	EnterDataChannelsCriticalSection

	Method used just to enter the critical section

***************************************************************************************/
void DaiSession::EnterDataChannelsCriticalSection()
{
	EnterCriticalSection(&mAddAndRemoveDataChannelsCriticalSection);
}


/***************************************************************************************

	LeaveDataChannelsCriticalSection

	Method used just to leave the critical section

***************************************************************************************/
void DaiSession::LeaveDataChannelsCriticalSection()
{
	LeaveCriticalSection(&mAddAndRemoveDataChannelsCriticalSection);
}



/***************************************************************************************
	
	FillEventArray

	Adds a new event (which was previously associated with a TransMux) to the event array
	This event array contains all the events for which the thread ExchangeDataThread
	will wait

***************************************************************************************/
void DaiSession::FillEventArray(WSAEVENT NewEvent)
{
	EnterCriticalSection(&mAddAndRemoveDataChannelsCriticalSection);
	m_WsaEventArray+=(WSAEVENT)NewEvent;
	LeaveCriticalSection(&mAddAndRemoveDataChannelsCriticalSection);

}



/***************************************************************************************
	
	SignalAllPeersOfSessionDestroy

	This method sends a message through the DMIF control channel informing the remote
	peers connected to this session that the current session will be destroyed

***************************************************************************************/
void DaiSession::SignalAllPeersOfSessionDestroy()
{

	char BuffAux[CONTROL_MSG_SIZE];
	DWORD SentBytes;
	Conversion_Union conv;

	ListOfPeersElement* pActual;
	ListOfPeersElement* pNext;

	EnterCriticalSection(&mPeersListCriticalSection);
	
	if(pHeadListOfPeersElement != NULL)
	{
		pNext = pActual = pHeadListOfPeersElement;
		do
		{
			pActual = pNext;
			pNext = pActual->pNextElement;
		
			// Inform the opposite side
			conv.int_number = SESSION_DESTROY_REQUEST;
			memcpy(BuffAux,conv.data,4);
			m_pDai->MySend((pActual->pDmifPeer)->m_hDmifCtrlSock,BuffAux,
				                              CONTROL_MSG_SIZE,SentBytes);

		}while(pNext != NULL);
	}

	LeaveCriticalSection(&mPeersListCriticalSection);
}


