///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  DrawableImage.cpp
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of a class which implements functionality common
//   to nodes which need to display images, such as the Image2DRenderer,
//   Movie2DRenderer and Background2DRenderer nodes.
//
// Created: September, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "Stream.h"
#include "Clock.h"
#include "DrawableImage.h"
#include "trace.h"


// =====================================================================

//
// DrawableImage::DrawableImage
//
// Class constructor.
//

DrawableImage::DrawableImage(MediaObject* pMO) : DrawableNode(pMO)
{
	m_updateRequired = FALSE;
	m_isTransparent = FALSE;
	m_isFinished = FALSE;
	m_startUpDelay = 0;
	m_recurrentInDraw = FALSE;
} /* DrawableImage::DrawableImage */


// =====================================================================

//
// DrawableImage::~DrawableImage
//
// Class destructor.
//

DrawableImage::~DrawableImage()
{
//	GetStream()->Terminate();
	GetStreamConsumer()->CloseStream();
} /* DrawableImage::~DrawableImage */


// =====================================================================

//
// DrawableImage::OnEndOfStream
//
// This method is called when the end of a stream is detected in either
// the PreRender() method or the Draw() method.
//

void DrawableImage::OnEndOfStream()
{
	m_updateRequired = FALSE;
	m_isFinished = TRUE;
} /* DrawableImage::OnEndOfStream */


// =====================================================================

//
// DrawableImage::PreRender
//
// Call the DrawableNode class's method to remember the effects and to
// register with the visual renderer.  Open the stream to see if the
// contents has changed, and hence needs to be redrawn.
//
// This function does not call UpdateBoundingRect(), as the derived
// class may modify the bounding rectangle.  The derived class MUST
// call UpdateBoundingRect().
//

void DrawableImage::PreRender(Effects const& effects)
{
	// Only proceed once a stream has been attached
	if (GetStream() == NULL)
		return;

	if (!InitPreRender(effects))
		return;

	// Decide if the stream contains a new frame to display
	int dataSize;
	DWORD rdwTime;
	BOOL recurrent, endOfStream;
	LPVOID dataPtr = GetStream()->Fetch(dataSize, rdwTime, recurrent,
										endOfStream, TRUE, FALSE,
										-m_startUpDelay);

	// If there is no data, no updating is required
	if (dataPtr == NULL)
	{
		TRACE_PRINT("DrawableImage::PreRender(): NULL data pointer returned.\n");
		m_updateRequired = FALSE;
	}

	// If data block returned has 0 size, no updating is required
	else if (dataSize == 0)
	{
		TRACE_PRINT("DrawableImage::PreRender(): 0 data size returned.\n");
		GetStream()->Release();
		m_updateRequired = FALSE;
	}

	// If the recurrent flag is TRUE, no updating is required
	// (also allow for the recurrent flag being set in the Draw method)
	// If the transform has changed, then this image also needs to be drawn.
	else if ((recurrent == TRUE) && (m_recurrentInDraw == TRUE) &&
			 (m_transform == m_oldTransform))
	{
		TRACE_PRINT("DrawableImage::PreRender(): Recurrent image detected.\n");
		GetStream()->Release(FALSE);
		m_updateRequired = FALSE;
	}

	// Otherwise there is a new frame to display
	else
	{
		// At the moment we can only handle images which have undergone
		// only translation.
		assert(m_transform.IsOnlyTranslation());

		// If the output of the decoder is YUVA, then we assume that
		// part of the image may be transparent.
		if (GetStreamConsumer()->GetStreamFormat() == FORMAT_YUVA)
			m_isTransparent = TRUE;
		else
			m_isTransparent = FALSE;

		// Set the bounding rectangle of this image
		DecoderImage const* pImgDetails = (DecoderImage*)dataPtr;
		int left = pImgDetails->offsetX;
		int top  = -pImgDetails->offsetY;
		m_transform.Apply(left, top);
		m_preRenderOffsetX =  pImgDetails->offsetX;
		m_preRenderOffsetY = -pImgDetails->offsetY;

		// If the output format is YUV, we need to ensure that the left and
		// top are even
		if ((GetStreamConsumer()->GetStreamFormat() == FORMAT_YUV) ||
			(GetStreamConsumer()->GetStreamFormat() == FORMAT_YUVA))
		{
			left = 2 * (int)(left / 2);
			top  = 2 * (int)(top / 2);
		}

		// Set the dimensions of the bounding rectangle
		m_boundingRect.SetDimensions((int)left, (int)top,
									 (int)pImgDetails->width,
									 (int)pImgDetails->height);

		GetStream()->Release(FALSE);
		m_updateRequired = TRUE;

		// Also handle the end of the stream
		if (endOfStream)
		{
			TRACE_PRINT("DrawableImage::PreRender(): end of stream detected.\n");
			OnEndOfStream();
		}
	}
} /* DrawableImage::PreRender */


// =====================================================================

//static int frameNum = 0;

//
// DrawableImage::Draw
//
// Draw the part of the contents of the Movie2D node which lies within
// the given bounding rectangle.  This function will need to perform
// the drawing, even if m_updateRequire is FALSE.
//
// The dimensions of the image will already have been read, and are
// given by the dimensions of the bounding rectangle.
//

void DrawableImage::Draw(BRect const* pClipBRect)
{
	assert(pClipBRect != NULL);

	// Only proceed once a stream has been attached
	if (GetStream() == NULL)
		return;

	// Modify the image size to only show the part which lies within
	// the given bounding rectangle.
	BRect visibleRect = m_boundingRect;
	visibleRect.Intersection(pClipBRect); 

	// Only proceed if there is actually something to display
	if (visibleRect.IsEmpty())
		return;

	// Get a pointer to the data containing the frame
	// (the PreRender() method needs to know if this fetch has
	// returned a non-recurrent block)
	int dataSize;
	DWORD rdwTime;
	BOOL endOfStream;
	LPVOID dataPtr = GetStream()->Fetch(dataSize, rdwTime, m_recurrentInDraw,
										endOfStream, TRUE, FALSE,
										-m_startUpDelay);

	// If there is no data, return
	if (dataPtr == NULL)
	{
		TRACE_PRINT("DrawableImage::Draw(): NULL data pointer returned.\n");
		return;
	}

	// If data block returned has 0 size, return
	else if (dataSize == 0)
	{
		TRACE_PRINT("DrawableImage::Draw(): 0 data size returned.\n");
		GetStream()->Release();
		return;
	}
	DecoderImage const* pImgDetails = (DecoderImage*)dataPtr;

	// Display the image, base on its format.  The image parameters
	// are set to point to only the visible portion of the image
	if ((GetStreamConsumer()->GetStreamFormat() == FORMAT_YUV) ||
		(GetStreamConsumer()->GetStreamFormat() == FORMAT_YUVA))
	{
		SrcYUVAImage image;
		image.width	 = min(visibleRect.width,  pImgDetails->width);
		image.height = min(visibleRect.height, pImgDetails->height);
		image.pitch	 = pImgDetails->pitch;
		image.left	 = visibleRect.left;
		image.top	 = visibleRect.top;

		// If the image is not recurrent, we need to allow for the possible
		// change in the image offset values received from the decoder
		if (!m_recurrentInDraw)
		{
			image.left +=  pImgDetails->offsetX - m_preRenderOffsetX;
			image.top  += -pImgDetails->offsetY - m_preRenderOffsetY;
		}

		// Ensure that the dimensions are even
		image.left   = 2 * (int)(image.left   / 2);
		image.top    = 2 * (int)(image.top    / 2);
		image.width  = 2 * (int)(image.width  / 2);
		image.height = 2 * (int)(image.height / 2);

/**
		frameNum++;
		char name[50];
		sprintf(name, "frame%03d-%dx%d.yuv", frameNum,
				visibleRect.width, visibleRect.height);
		FILE* tmpfile = fopen(name,"wb");
		assert(tmpfile != NULL);
		fwrite((unsigned char*)dataPtr + sizeof(DecoderImage),
			   pImgDetails->width * pImgDetails->height * 3 / 2, 1, tmpfile);
		fclose(tmpfile);
**/

		// Get a pointer to the top-left pixel in the Y band
		int bandYsize = pImgDetails->width * pImgDetails->height;
/**
		int offset = 0;
		if (m_boundingRect.top > visibleRect.top)
			offset = (m_boundingRect.top - visibleRect.top) * image.pitch;
		if (visibleRect.left > m_boundingRect.left)
			offset += visibleRect.left - m_boundingRect.left;
**/
		int offset	 = (m_boundingRect.top - visibleRect.top) * image.pitch
						+ (visibleRect.left - m_boundingRect.left);
		dataPtr		 = (unsigned char*)dataPtr + sizeof(DecoderImage);
		image.pY	 = (unsigned char*)dataPtr + offset;

		// Get a pointer to the top-left pixel in the A band
		if (GetStreamConsumer()->GetStreamFormat() == FORMAT_YUV)
			image.pA = NULL;
		else
			image.pA = (unsigned char*)dataPtr + bandYsize * 3 / 2 + offset;

		// Get a pointer to the top-left pixel in the U and V band
/**
		offset = 0;
		if (m_boundingRect.top > visibleRect.top)
			offset = (int)((m_boundingRect.top - visibleRect.top) / 2) * image.pitch;
		if (visibleRect.left > m_boundingRect.left)
			offset += (visibleRect.left - m_boundingRect.left) / 2;
**/
		offset		 = (int)((m_boundingRect.top - visibleRect.top) / 2) * image.pitch
						+ (visibleRect.left - m_boundingRect.left) / 2;
		image.pU	 = (unsigned char*)dataPtr + bandYsize + offset;
		image.pV	 = (unsigned char*)dataPtr + bandYsize * 5 / 4 + offset;

		GetVisualRenderer()->DisplayYUV420AImage(image);
	}

	else if ((GetStreamConsumer()->GetStreamFormat() == FORMAT_RGB) ||
			 (GetStreamConsumer()->GetStreamFormat() == FORMAT_BGR))
	{
		SrcRGBAImage image;
		image.width	 = visibleRect.width;
		image.height = visibleRect.height;
		image.pitch	 = pImgDetails->pitch;
		image.left	 = visibleRect.left;
		image.top	 = visibleRect.top;

		// Get a pointer to the top-left pixel
		int offset	 = (m_boundingRect.top - visibleRect.top) * image.pitch
						+ (visibleRect.left - m_boundingRect.left) * 3;
		if (GetStreamConsumer()->GetStreamFormat() == FORMAT_RGB)
		{
			image.pR	 = (unsigned char*)dataPtr + sizeof(DecoderImage) + offset;
			image.pG	 = image.pR + 1;
			image.pB	 = image.pR + 2;
		}
		else
		{
			image.pB	 = (unsigned char*)dataPtr + sizeof(DecoderImage) + offset;
			image.pG	 = image.pB + 1;
			image.pR	 = image.pB + 2;
		}

		image.pA	 = NULL;
		image.ptrInc = 3;

		GetVisualRenderer()->DisplayRGBAImage(image);
	}

	else
	{
		assert(0);		// Shouldn't get here!
	}

#ifdef WRITE_TRACE
	if (dataSize == 176 * 144 * 3 + 1 + sizeof(DecoderImage))
		fprintf(traceFile, "H263: ");
	else if (dataSize == 176 * 144 * 3 / 2 + sizeof(DecoderImage))
		fprintf(traceFile, "H263: ");
	else
		fprintf(traceFile, "JPEG: ");
	DWORD currentTime = GetStreamConsumer()->GetExecutive()->GetClock()->GetTime();
	TRACE_PRINT3("Displaying for time %ld at %ld (%ld)\n", rdwTime,
				 currentTime, rdwTime - currentTime);
#endif

	// Release this unit from the stream, now that it is finished with,
	// but leave the data there in case nothing new has arrived when this
	// function is next called.
	GetStream()->Release(FALSE);

	// Also handle the end of the stream
	if (endOfStream)
	{
		OnEndOfStream();
		TRACE_PRINT("DrawableImage::Draw(): end of stream detected.\n");
	}
} /* DrawableImage::Draw */


// =====================================================================
