///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  DrawableNode.cpp
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of methods of the class which encapsulates all
//   of the features of a node which can be drawn by the visual renderer.
//   This file contains the implementation of methods common to all nodes
//   which are drawable.
//
// Created: August, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "DrawableNode.h"
#include "VisRend.h"
#include "trace.h"


// The storage for the static data member of the DrawableNode class
int DrawableNode::m_numDrawablesCreated = 0;


// =====================================================================

//
// DrawableNode::DrawableNode
//
// Class constructor - initialise the associated variables, and record
// the order in which this node was created (used to determine the
// order to drawn nodes).
//

DrawableNode::DrawableNode(MediaObject* pMO) : Traversable(pMO)
{
	m_drawOrder = 0;
	m_isTransparent = TRUE;
	m_prevBoundingRect.MakeEmpty();
	m_combinedBoundingRect.MakeEmpty();
	m_isSwitchedOn = TRUE;

	m_createOrder = m_numDrawablesCreated++;
} /* DrawableNode::DrawableNode */


// =====================================================================

//
// DrawableNode::InitPreRender
//
// The method will store a copy of all of the effects parameters
// locally in this node, and then register this node with the visual
// renderer so that it is drawn.  If this node has been switched off,
// FALSE is returned, in which case the calling function must return
// immediately.
//

BOOL DrawableNode::InitPreRender(Effects const& effects)
{
	m_oldTransform = m_transform;
	m_transform = effects.transform;
	m_drawOrder = effects.drawOrder;
	m_isSwitchedOn = effects.isSwitchedOn;

	GetVisualRenderer()->AddToDrawList(this);

	// If this node is switched off, then set its bounding rectangle
	// to empty, and update the overall changed bounding rectangle.
	// FALSE is returned.
	if (!m_isSwitchedOn)
	{
		m_boundingRect.MakeEmpty();
		UpdateBoundingRect();
		return FALSE;
	}
	else
		return TRUE;
} /* DrawableNode::InitPreRender */


// =====================================================================

//
// DrawableNode::UpdateBoundingRect
//
// Call this function from the PreRender() method of a derived
// class, once the m_boundingRect data member has been set.
// This will combine the current bounding rectangle with the
// previous.
//

void DrawableNode::UpdateBoundingRect()
{
	// If the previous and current bounding rectangles are different,
	// then this node will be transparent.
	if (m_prevBoundingRect != m_boundingRect)
		m_isTransparent = TRUE;
	else
		m_isTransparent = FALSE;

	// Calculate the combined bounding rectangle of the previous and
	// current bounding rectangles
	m_combinedBoundingRect = m_prevBoundingRect;
	m_combinedBoundingRect.Union(m_boundingRect);
	m_prevBoundingRect = m_boundingRect;
} /* DrawableNode::UpdateBoundingRect */


// =====================================================================
