///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  DrawableNode.h
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The definition of a class which encapsulates all of the features
//   of a node which can be drawn by the visual renderer.  I would have
//   liked to call this class Drawable, but X11 uses this name already.
//
// Created: August, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#ifndef DRAWABLE_NODE_H
#define DRAWABLE_NODE_H


#include "Traversable.h"
#include "BRect.h"


class DrawableNode : public Traversable
{
	//
	// Public functions
	//
public:

	DrawableNode(MediaObject* pMO);
	// Class constructor

	virtual float GetDrawingOrder() const { return m_drawOrder; }
	// Returns the drawing order value for this node.  The default
	// implementation returns the appropriate data member.

	int GetCreateOrder() const { return m_createOrder; }
	// Returns a number which indicates when this object was created.
	// The larger the number, the later the node was created.  This
	// can be used to determine which node appeared before another in
	// the order given in the source file.

	virtual BOOL NeedsDrawing() const = 0;
	// Returns TRUE if this node needs to be drawn (because it has
	// changed since it was last drawn).

	virtual BOOL IsTransparent() const { return m_isTransparent; }
	// Return TRUE if this node is transparent.  ie. it does not
	// completely fill its bounding rectangle, either because not
	// every pixel is written and/or and alpha blending component
	// is used.  The default implementation returns TRUE if the
	// previous and current bounding boxes are different.  A derived
	// class which implements this method, must include the result
	// of this method from this class.

	virtual void Draw(BRect const* pClipBRect) = 0;
	// Draw the part of the node which is enclosed inside the given
	// bounding rectangle, or draw the entire object if no bounding
	// rectangle is given.
	//
	// The bounding rectange uses the BIFS coordinate system, with the
	// origin of the output display is in the middle of the window, with
	// increasing X and Y to the right and up.

	virtual BOOL IsPointOver(int x, int y) { return GetBoundingRect()->IsPointIn(x, y); }
	// Return TRUE if the given point, in BIFS coordinates, is above the
	// current shape.  Return FALSE otherwise.  The default implementation
	// uses the bounding rectangle of the drawable.

	const BRect* GetBoundingRect() const { return &m_boundingRect; }
	// Returns the bounding rectangle which completely encloses this
	// object.  The bounding rectange uses the BIFS coordinate system,
	// with the origin of the output display is in the middle of the
	// window, with increasing X and Y to the right and up.

	const BRect* GetCombinedBoundingRect() const { return &m_combinedBoundingRect; }
	// Returns the bounding rectangle which completely encloses this
	// object, at it current and previous locations.  The bounding
	// rectange uses the BIFS coordinate system, with the origin of
	// the output display is in the middle of the window, with
	// increasing X and Y to the right and up.


	//
	// Protected implementation details
	//
protected:
	TransformMatrix2D m_oldTransform;
	TransformMatrix2D m_transform;

	// The bounding rectangle used by the drawable node for this render
	BRect m_boundingRect;

	BOOL InitPreRender(Effects const& effects);
	// The method will store a copy of all of the effects parameters
	// locally in this node, and then register this node with the visual
	// renderer so that it is drawn.  If this node has been switched off,
	// FALSE is returned, in which case the calling function must return
	// immediately.

	void UpdateBoundingRect();
	// Call this function from the PreRender() method of a derived
	// class, once the m_boundingRect data member has been set.
	// This will combine the current bounding rectangle with the
	// previous.

	BOOL IsSwitchedOn() const { return m_isSwitchedOn; }
	// Returns FALSE if this node is switched off, as the result of a
	// Switch node, otherwise it returns TRUE.


	//
	// Private implementation details
	//
private:
	BOOL m_isTransparent;
	int m_createOrder;
	static int m_numDrawablesCreated;
	float m_drawOrder;
	BOOL m_isSwitchedOn;

	// The bounding rectangle used by the drawable node for the
	// previous render
	BRect m_prevBoundingRect;

	// The union of the two bounding boxes, which is used when
	// determining the bounding box of this node, taking into account
	// the fact that it might have moved or changed shape since it
	// was last drawn.
	BRect m_combinedBoundingRect;

};


#endif // DRAWABLE_NODE_H
