///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  DrawableText.cpp
//
// Authors: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of a class which implements functionality common
//   to nodes which need to display text, namely the Text2D and
//   StreamingText nodes.
//
// Created: October, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "DrawableText.h"


// =====================================================================

//
// DrawableText::DrawableText
//
// Class constructor - initial array of lines of text.
//

DrawableText::DrawableText(MediaObject* pMO) : GeometryNode(pMO)
{
	m_numLines = 0;
	m_lines = NULL;
	m_numLinesAllocated = 0;
} /* DrawableText::DrawableText */


// =====================================================================

//
// DrawableText::~DrawableText
//
// Class destructor - delete array of lines of text.
//

DrawableText::~DrawableText()
{
	if (m_lines != NULL)
		free(m_lines);
} /* DrawableText::~DrawableText */


// =====================================================================

//
// DrawableText::GetFStype
//
// Return the 'type' field from the FontStyle2D node.  If this no is
// not present, use the default value.
//
const char* DrawableText::GetFStype() const
{
	if (GetFontStyle2D() == NULL)
		return "";
	else
		return (LPCSTR)(const ZString&)GetFontStyle2D()->type;
} /* DrawableText::GetFStype */


// =====================================================================

//
// DrawableText::GetFSstyle
//
// Return the 'style' field from the FontStyle2D node.  If this no is
// not present, use the default value.
//

int DrawableText::GetFSstyle() const
{
	if (GetFontStyle2D() == NULL)
		return 0;
	else
		return GetFontStyle2D()->style;
} /* DrawableText::GetFSstyle */


// =====================================================================

//
// DrawableText::GetFSsize
//
// Return the 'size' field from the FontStyle2D node.  If this no is
// not present, use the default value.
//

int DrawableText::GetFSsize() const
{
	if (GetFontStyle2D() == NULL)
		return -1;
	else
		return GetFontStyle2D()->size;
} /* DrawableText::GetFSsize */


// =====================================================================

//
// DrawableText::GetFShSpacing
//
// Return the 'hSpacing' field from the FontStyle2D node.  If this no is
// not present, use the default value.
//

int DrawableText::GetFShSpacing() const
{
	if (GetFontStyle2D() == NULL)
		return 100;
	else
		return GetFontStyle2D()->hSpacing;
} /* DrawableText::GetFShSpacing */


// =====================================================================

//
// DrawableText::GetFSvSpacing
//
// Return the 'vSpacing' field from the FontStyle2D node.  If this no is
// not present, use the default value.
//

int DrawableText::GetFSvSpacing() const
{
	if (GetFontStyle2D() == NULL)
		return 100;
	else
		return GetFontStyle2D()->vSpacing;
} /* DrawableText::GetFSvSpacing */


// =====================================================================

//
// DrawableText::GetFSlanguage
//
// Return the 'language' field from the FontStyle2D node.  If this no is
// not present, use the default value.
//

const char* DrawableText::GetFSlanguage() const
{
	if (GetFontStyle2D() == NULL)
		return "";
	else
		return (LPCSTR)(const ZString&)GetFontStyle2D()->language;
} /* DrawableText::GetFSlanguage */


// =====================================================================

//
// DrawableText::PreRender
//
// This function will be called when the dimensions of the text string,
// stored in m_data, need to be determined.  The m_boundingRect data
// member will be set to store the new bounding rectangle.  The base
// class PreRender() and UpdateBoundingRect() methods WILL NOT be
// called.
//

void DrawableText::PreRender(Effects const& /*effects*/)
{
	// Determine the properties of the font to use
	m_fontName = GetFStype();
	if ((m_fontName == NULL) || (m_fontName[0] == 0))
		m_fontName = "Times";

	m_fontSize = GetFSsize();
	if (m_fontSize == -1)
		m_fontSize = 12;

	m_bold = (GetFSstyle() & 1) ? TRUE : FALSE;
	m_italic = (GetFSstyle() & 2) ? TRUE : FALSE;
	m_underline = (GetFSstyle() & 4) ? TRUE : FALSE;

	// The following loop will only be done twice if the length
	// field of a Text2D node is given.  The first pass will use
	// either the default font or the font specified.  If a second
	// pass is done, the font size will have been changed to adjust
	// the width of the text to match that specified in the Text2D node.
	int pass = 1;
	while (pass <= 2)
	{
		// Set the font which is to be used to calculate the size of the
		// text string when rendered
		VisualRenderer* pVR = GetVisualRenderer();
		void* pFont = pVR->SetFont(m_fontName, m_fontSize, m_bold,
								   m_italic, m_underline);

		m_boundingRect.MakeEmpty();

		// Determine the dimension of every line of text in the string
		int offset = 0;
		int startOffset = 0;
		BOOL finished = FALSE;
		m_numLines = 0;
		while (!finished)
		{
			// Look for the end of a line (& also the end of the string)
			if ((m_data[offset] == 13) || (m_data[offset] == 0))
			{
				// Make a new line structure to store this line info into
				if (m_numLines == m_numLinesAllocated)
				{
					m_numLinesAllocated += 10;
					int newSize = m_numLinesAllocated * sizeof(LineOfText);
					m_lines = (LineOfText*) realloc(m_lines, newSize);
				}

				// Determine the start and length of this line
				m_lines[m_numLines].start    = startOffset;
				m_lines[m_numLines].numChars = offset - startOffset;

				// and now determine the dimensions of this line in the
				// current font
				pVR->GetTextSize(pFont, &(m_data[startOffset]),
								 m_lines[m_numLines].numChars,
								 m_lines[m_numLines].width,
								 m_lines[m_numLines].height);

				// Update the size of the bounding rectangle
				if ((int)m_lines[m_numLines].width > m_boundingRect.width)
					m_boundingRect.width = (int)m_lines[m_numLines].width;
				m_boundingRect.height += (int)m_lines[m_numLines].height;

				// Prepare for the next line
				m_numLines++;
				if (m_data[offset] == 13)
				{
					// skip over the line break character
					offset++;
					startOffset = offset;
				}
				else if (m_data[offset] == 0)
					finished = TRUE;
			}
			else
				// Otherwise move onto the next character
				offset++;
		}

		// Unset the font now that we have finished with it.
		pVR->UnsetFont(pFont);

		// If this is the first pass, and a length field is given,
		// then determine what size font is required to have the
		// text displayed with this width
		int length = GetLength();
		if ((pass == 1) && (length != -1))
		{
			assert(length != 0);

			// Assume that by scaling the font size, the width become correct
			m_fontSize = m_fontSize * length / m_boundingRect.width;

			pass++;
		}
		else
			// Otherwise we have finished
			pass = 3;
	}

	// Expand the bounding rectangle to take into account the shadow
	if (GetShadow())
	{
		int sx = GetShadPosX();
		int sy = GetShadPosY();

		m_boundingRect.width  += abs(sx);
		m_boundingRect.height += abs(sy);
		if (sx < 0)
			m_boundingRect.left = sx;
		if (sy > 0)
			m_boundingRect.top = sy;
	}

	// At the moment we can only handle text which has undergone
	// only translation.
	assert(m_transform.IsOnlyTranslation());

	// Translate the text string, according to the transform parameters
	float left = m_boundingRect.left;
	float top  = m_boundingRect.top;
	m_transform.Apply(left, top);
	m_boundingRect.left = left + 0.5;
	m_boundingRect.top  = top + 0.5;
} /* DrawableText::PreRender */


// =====================================================================

//
// DrawableText::Draw
//
// Render the text string stored in the m_data field, using the following
// appearance fields:
//   diffuseColor, shadow & shadPos.
//

void DrawableText::Draw(BRect const* pClipBRect)
{
	// Set the font which is to be used
	VisualRenderer* pVR = GetVisualRenderer();
	void* pFont = pVR->SetFont(m_fontName, m_fontSize, m_bold,
							   m_italic, m_underline);

	// Determine where to start drawing the text
	int x = m_boundingRect.left;
	int y = m_boundingRect.top;
	int sx = GetShadPosX();
	int sy = GetShadPosY();
	if (GetShadow())
	{
		if (sx < 0)
			x -= sx;
		if (sy > 0)
			y -= sy;
	}

	// Display each of the lines in the string
	for (int line = 0; line < m_numLines; line++)
	{
		// Draw the shadow - if required
		if (GetShadow())
		{
			// Use green as the shadow for the moment
			Color3f shadColor(0, 0.5, 0);
			pVR->DisplayText(pFont, &(m_data[m_lines[line].start]),
							 m_lines[line].numChars,
							 x + (int)GetShadPosX(),
							 y + (int)GetShadPosY(),
							 shadColor, pClipBRect);
		}

		// Draw the text itself
		pVR->DisplayText(pFont, &(m_data[m_lines[line].start]),
						 m_lines[line].numChars, x, y,
						 GetDiffuseColor(), pClipBRect);

		// Move onto the next line
		y -= m_lines[line].height;
	}

	// Unset the font now that we have finished with it.
	pVR->UnsetFont(pFont);
} /* DrawableText::Draw */


// =====================================================================
