# This perl script will insert the description of functions contained in a
# header file into a HTML document.
#
# The functions must be in the format:
#
#   function definition
#   // comments
#
# The function can be split across multiple lines, as long as the name of
# the function is on the first line.  Likewise, the comment can also be
# on multiple lines, as long as there are no blank lines between the function
# definition and the comment.
#
# The functions and their associated comments are inserted into the HTML
# document at the place specifed by the following two lines:
#
#   <!-- StartInsertCode: insert info -->
#
#   <!-- EndInsertCode: -->
#
# The lines between these comments are discarded, being replaced instead
# by the appropriate contents from the header file.
#
# The insertion info takes the following form:
#
#   HeaderFile function1|function2|function3
#
# where HeaderFile is the name of the header file to read from, and the
# functions to insert are separated by |
#
# Matthew Leditschke
# CSELT
# October, 1997
#


#
# InsertCode
#
# Given the name of the header file, and the | separated list of functions
# to insert.
#
# Insert the function definitions and comments into the output file.
#

sub InsertCode
{
    # Get the two parameters given to this subroutine
    $headerFile = $_[0];
    $functionList = $_[1];

    print OUTFILE "<!-- The following lines have been automatically copied into here from $headerFile, do not edit. -->\n";

    # Open the header file for reading
    open(HDRFILE,$headerFile)
	|| die("InsertCodeIntoHTML: Error opening '$headerFile' for reading. $!\n");

    $inFunction = 0;
    $inComment = 0;

    $functionList = " (".$functionList.")\\s*\\(";

    # Go through the header file, looking for the functions to insert
    while (<HDRFILE>)
    {
	# A blank line terminates a function definition
	if (/^\s*$/)
	{
	    if ($inComment)
	    {
		print OUTFILE "<P>\n\n";
	    }
	    $inFunction = 0;
	    $inComment = 0;
	}

	# A comment line with no words is a paragraph break
	elsif ((/\/\/\s*$/) && ($inComment))
	{
	    print OUTFILE "<P>\n\n";
	}

	# A comment is inserted as plain text
	elsif (/\/\/ (.+)$/)
	{
	    if ($inFunction)
	    {
		$text = $1;

		# If this is the first comment line, finish the <PRE> section
		if (!$inComment)
		{
		    print OUTFILE "</PRE>\n";
		}

		# Make references to methods in <TT> format
		$text =~ s/(\S+\(\))/<TT>$1<\/TT>/;

		$inComment = 1;
		print OUTFILE $text."\n";
	    }
	}

	# Otherwise if this is in a function, it must be part of the definition
	elsif ($inFunction)
	{
	    # Remove any trailing " = 0;" or ";"
	    s/ = 0;$//;
	    s/;$//;

	    # Expand tabs
	    s/\011/    /g;

	    print OUTFILE;
	}

	# Look for the start of a function definition
	elsif (/$functionList/)
	{
	    $inFunction = 1;
	    $inComment = 0;

	    # Remove any trailing " = 0;" or ";"
	    s/ = 0;$//;
	    s/;$//;

	    # Expand tabs
	    s/\011/    /g;

	    print OUTFILE "<PRE>\n";
	    print OUTFILE;
	}
    }

    close(HDRFILE);
}



# Check that the command line is correct
if ($#ARGV != 0)
{
    die("Usage: InsertCodeIntoHTML htmlfile\n");
}


# Open the input file
$infileName = $ARGV[0];
open(INFILE,$infileName)
 || die("InsertCodeIntoHTML: Error opening '$infileName' for reading. $!\n");

# Open a temporary file to write the modified input file to
$tempFileName = $infileName.".tmp";
open(OUTFILE,">".$tempFileName)
 || die("InsertCodeIntoHTML: Error opening '$tempFileName' for writing. $!\n");

$inInsertion = 0;

# Go through the input file, looking for insert markers
while (<INFILE>)
{
    # Look for the start of an insertion
    if (/<!-- StartInsertCode: (\S+)\s+(\S+) -->/)
    {
	print OUTFILE;
	$inInsertion = 1;
	InsertCode($1, $2);
    }

    # Look for the end of an insertion
    elsif (/<!-- EndInsertCode: -->/)
    {
	print OUTFILE;
	$inInsertion = 0;
    }

    # Otherwise only copy the line to the output if not in an insertion
    elsif (!$inInsertion)
    {
	print OUTFILE;
    }
}

# Close the files, and replace the given file with the temporary one
# which has just been created
close INFILE;
close OUTFILE;
unlink($infileName);
rename($tempFileName, $infileName);
