///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  IndexedLineSet2DProxy.cpp
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of the proxy object used by the presenter
//   to render the IndexedLineSet2D node.
//
// Created: August, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-CSELT.h"
#include "IndexedLineSetProxy.h"
#include "ColorProxy.h"
#include "Coordinate2DProxy.h"
#include "trace.h"


// =====================================================================

//
// PresenterCSELT::InitIndexedLineSet2D
//
// Initialise the proxy data structure and attach it to the render data
// pointer of the given media object node.
//

void PresenterCSELT::InitIndexedLineSet2D(IndexedLineSet2D* pObj)
{
	pObj->SetPresenterData(new IndexedLineSet2DProxy(pObj));
}


// ======================================================================

//
// IndexedLineSet2DProxy::Print
//
// Print out the fields and corresponding values for the media object
// derived node that this object is a proxy for, indented according to
// the given depth.
//

void IndexedLineSet2DProxy::Print(FILE* outfile, int depth) const
{
	fprintf(outfile, "IndexedLineSet2DProxy {\n");

	fprintf(outfile, "%scolor \n", IndentPrefix(depth + 1));
	if (GetColor() == NULL)
		fprintf(outfile, "[]\n");
	else
	((ColorProxy*)(GetColor()->GetPresenterData()))->Print(outfile, depth + 1);

	fprintf(outfile, "%scoord \n", IndentPrefix(depth + 1));
	if (GetCoordinate2D() == NULL)
		fprintf(outfile, "[]\n");
	else
	((Coordinate2DProxy*)(GetCoordinate2D()->GetPresenterData()))->Print(outfile, depth + 1);

	fprintf(outfile, "%scolorIndex \n", IndentPrefix(depth + 1));
	if (m_nColorLength <= 0)
		fprintf(outfile, "[]\n");
	else
	{
		for (int i = 0; i < GetIndexLineSet()->colorIndex.GetLength(); i++)
			fprintf(outfile, " %d", (int)GetIndexLineSet()->colorIndex[i]);
		fprintf(outfile, " ]\n");
	}

	fprintf(outfile, "%scolorPerVertex %s\n", IndentPrefix(depth + 1),
			BoolToStr(GetIndexLineSet()->colorPerVertex));

	fprintf(outfile, "%scoordIndex \n", IndentPrefix(depth + 1));
	if (m_nCoordLength <= 0)
		fprintf(outfile, "[]\n");
	else
	{
		for (int i = 0; i < GetIndexLineSet()->coordIndex.GetLength(); i++)
			fprintf(outfile, " %d", (int)GetIndexLineSet()->coordIndex[i]);
		fprintf(outfile, " ]\n");
	}

	fprintf(outfile, "%s}\n", IndentPrefix(depth));
} /* IndexedLineSet2DProxy::Print */


// ======================================================================

//
// IndexedLineSet2DProxy::PreRender
//
//

void IndexedLineSet2DProxy::PreRender(Effects const& effects)
{
	if (!InitPreRender(effects))
		return;

	// If the transparency for this node is 1, then it doesn't need
	// to be drawn, and its bounding rectangle is empty
	if (GetTransparency() > 1 - ALMOST_ZERO)
	{
		m_boundingRect.MakeEmpty();
		UpdateBoundingRect();
		return;
	}

	// Allocate memory for the local copy of the points (only if the
	// current array is not large enough).
	m_nLength = GetCoordinate2D()->point.GetLength();
	if (m_nLength > m_nAllocatedLength)
	{
		// Delete the previous arrays - if required
		if (m_lpPoints)
			delete [] m_lpPoints;
		if (m_shadows)	
			delete [] m_shadows;

		// Create the new arrays
		m_lpPoints = new PointF[m_nLength];
		m_shadows = new PointF[m_nLength];
		m_nAllocatedLength = m_nLength;
	}

	m_nColorLength = GetIndexLineSet()->colorIndex.GetLength();
	if (m_nColorLength > m_nPrevColorLength)
	{
		// update the previous length
		m_nPrevColorLength = m_nColorLength;

		// Delete the previous arrays - if required
		// and create the new one
		if (m_pColorIndex) 
			delete[] m_pColorIndex;
		m_pColorIndex = new int[m_nColorLength];
	}
	// Put in the color values
	for (int i = 0; i < m_nColorLength; i++)
		m_pColorIndex[i] = (int)(GetIndexLineSet()->colorIndex[i]);

	m_nCoordLength = GetIndexLineSet()->coordIndex.GetLength();
	if (m_nCoordLength > m_nPrevCoordLength)
	{
		// update the previous length
		m_nPrevCoordLength = m_nCoordLength;

		// Delete the previous arrays - if required
		// and create the new one
		if (m_pCoordIndex) 
			delete[] m_pCoordIndex;
		m_pCoordIndex = new int[m_nCoordLength];
	}
	// Put in the coordinate values
	for (i = 0; i < m_nCoordLength; i++)
		m_pCoordIndex[i] = (int)(GetIndexLineSet()->coordIndex[i]);

	// Determine where the points are
	for (i =0; i < m_nLength; i++)
	{
		m_lpPoints[i].x = GetCoordinate2D()->point[i].x; 
		m_lpPoints[i].y = GetCoordinate2D()->point[i].y;
	}

	// Determine where the corresponding shadow corners are
	for (i = 0; i < m_nLength; i++)
	{
		m_shadows[i].x = m_lpPoints[i].x + GetShadPosX();
		m_shadows[i].y = m_lpPoints[i].y + GetShadPosY();
	}

	// Determine where these points lie after the transformation 
	// has been applied.
	for (i = 0; i < m_nLength; i++)
	{
		m_transform.Apply(m_lpPoints[i].x, m_lpPoints[i].y);
		m_transform.Apply(m_shadows[i].x, m_shadows[i].y);
	}

	// Set the bounding box
	m_boundingRect.SetDimensions(m_lpPoints, m_nLength);

	// If a shadow is to be drawn, determine where the shadow points
	// are transformed to, and extend the bounding rectangle
	if (GetShadow())
	{
		BRect shadowRect;
		shadowRect.SetDimensions(m_shadows, m_nLength);
		m_boundingRect.Union(shadowRect);
	}

	UpdateBoundingRect();
} /* IndexedLineSet2DProxy::PreRender */


// ============================================================================

//
// IndexedLineSet2DProxy::Draw
//
// Draw the lines.
//

void IndexedLineSet2DProxy::Draw(BRect const* pClipBRect)
{
	assert(pClipBRect != NULL);
	VisualRenderer* pVR = GetVisualRenderer();

	// The amount of the lines to draw lies within both the clipping
	// rectangle and the lines own bounding rectangle.
	BRect visibleRect = *pClipBRect;
	visibleRect.Intersection(GetBoundingRect());
	if (visibleRect.IsEmpty())
		return;

	int i = 0;


	// If a shadow is required, draw it now
	if (GetShadow())
	{
		// Determine where to draw the shadow, allowing for the transformation
		int shadowPosX = GetShadPosX();
		int shadowPosY = GetShadPosY();
		m_transform.Apply(shadowPosX, shadowPosY);

		// Only draw the shadow if it is not directly underneath the polyLines...
		if ((shadowPosX != 0) || (shadowPosY != 0))
		{
			Color3f shadowColor(0.3, 0.8, 0.45);

			if (m_nCoordLength > 0)
			{
				// Join the lines according to the coordIndex array
				for (i = 0; i < m_nCoordLength; i += 3)
					pVR->DisplayLines(visibleRect, 
									  m_shadows[m_pCoordIndex[i]].x,
									  m_shadows[m_pCoordIndex[i]].y,
									  m_shadows[m_pCoordIndex[i + 1]].x,
									  m_shadows[m_pCoordIndex[i + 1]].y,
									  shadowColor, GetWidth());
			}
			else
			{
				// Join the lines sequentially
				for (i = 0; i < m_nLength - 1; i++)
					pVR->DisplayLines(visibleRect, 
									  m_shadows[i].x,
									  m_shadows[i].y,
									  m_shadows[i + 1].x,
									  m_shadows[i + 1].y,
									  shadowColor, GetWidth());

				// Now join the last vertex with the first
				pVR->DisplayLines(visibleRect, 
								  m_shadows[m_nLength - 1].x,
								  m_shadows[m_nLength - 1].y,
								  m_shadows[0].x,
								  m_shadows[0].y,
								  shadowColor, GetWidth());
			}
		}
	}

	// Draw the lines. 
	// If the color node isn't NULL, use it
	if (GetColor())
	{
		if (m_nCoordLength > 0)
		{
			if (GetVertex())
			{
				if (m_nColorLength > 0)
				{
				}
				else
				{
				}
			}
			else
			{
				if (m_nColorLength > 0)
				{
					// Join the lines according to the coordIndex array
					// And use the colors indicised by the colorIndex array
					int count;
					for (i = 0, count = 0; i < m_nCoordLength; i += 3, count++)
						pVR->DisplayLines(visibleRect, 
										  m_lpPoints[m_pCoordIndex[i]].x,
										  m_lpPoints[m_pCoordIndex[i]].y,
										  m_lpPoints[m_pCoordIndex[i + 1]].x,
										  m_lpPoints[m_pCoordIndex[i + 1]].y,
										  GetColor()->color[m_pColorIndex[count]],
										  GetWidth());
				}
				else
				{
					// Join the lines according to the coordIndex array
					// And use the colors indicised sequentially
					int count;
					for (i = 0, count = 0; i < m_nCoordLength; i += 3, count++)
						pVR->DisplayLines(visibleRect, 
										  m_lpPoints[m_pCoordIndex[i]].x,
										  m_lpPoints[m_pCoordIndex[i]].y,
										  m_lpPoints[m_pCoordIndex[i + 1]].x,
										  m_lpPoints[m_pCoordIndex[i + 1]].y,
										  GetColor()->color[count],
										  GetWidth());
				}
			}
		}
		else
		{
			if (GetVertex())
			{
				if (m_nColorLength > 0)
				{
				}
				else
				{
				}
			}
			else
			{
				if (m_nColorLength > 0)
				{
					// Join the lines sequentially
					// And use the colors indicised by the colorIndex array
					for (i = 0; i < m_nLength - 1; i++)
						pVR->DisplayLines(visibleRect, 
										  m_lpPoints[i].x,
										  m_lpPoints[i].y,
										  m_lpPoints[i + 1].x,
										  m_lpPoints[i + 1].y,
										  GetColor()->color[m_pColorIndex[i]],
										  GetWidth());

					// Now join the last vertex with the first
					pVR->DisplayLines(visibleRect, 
									  m_lpPoints[m_nLength - 1].x,
									  m_lpPoints[m_nLength - 1].y,
									  m_lpPoints[0].x,
									  m_lpPoints[0].y,
									  GetColor()->color[m_pColorIndex[i]],
									  GetWidth());
				}
				else
				{
					// Join the lines sequentially
					// And use the colors indicised sequentially
					for (i = 0; i < m_nLength - 1; i++)
						pVR->DisplayLines(visibleRect, 
										  m_lpPoints[i].x,
										  m_lpPoints[i].y,
										  m_lpPoints[i + 1].x,
										  m_lpPoints[i + 1].y,
										  GetColor()->color[i],
										  GetWidth());

					// Now join the last vertex with the first
					pVR->DisplayLines(visibleRect, 
									  m_lpPoints[m_nLength - 1].x,
									  m_lpPoints[m_nLength - 1].y,
									  m_lpPoints[0].x,
									  m_lpPoints[0].y,
									  GetColor()->color[i],
									  GetWidth());
				}
			}
		}
	}
	// Draw the lines 
	// and use the Material diffuseColor
	else
	{
		Color3f sameColor = GetDiffuseColor();

		if (m_nCoordLength > 0)
		{
			if (GetVertex())
			{
			}
			else
			{
				// Join the lines according to the coordIndex array
				int count;
				for (i = 0, count = 0; i < m_nCoordLength; i += 3, count++)
					pVR->DisplayLines(visibleRect, 
									  m_lpPoints[m_pCoordIndex[i]].x,
									  m_lpPoints[m_pCoordIndex[i]].y,
									  m_lpPoints[m_pCoordIndex[i + 1]].x,
									  m_lpPoints[m_pCoordIndex[i + 1]].y,
									  sameColor,
									  GetWidth());
			}
		}
		else
		{
			if (GetVertex())
			{
			}
			else
			{
				// Join the lines sequentially
				// And use the Material diffusColor
				for (i = 0; i < m_nLength - 1; i++)
					pVR->DisplayLines(visibleRect, 
									  m_lpPoints[i].x,
									  m_lpPoints[i].y,
									  m_lpPoints[i + 1].x,
									  m_lpPoints[i + 1].y,
									  sameColor,
									  GetWidth());

				// Now join the last vertex with the first
				pVR->DisplayLines(visibleRect, 
								  m_lpPoints[m_nLength - 1].x,
								  m_lpPoints[m_nLength - 1].y,
								  m_lpPoints[0].x,
								  m_lpPoints[0].y,
								  sameColor,
								  GetWidth());
			}
		}
	}
} /* IndexedLineSet2DProxy::Draw */


// ===========================================================================
