///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  Movie2DProxy.cpp
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of the proxy object used by the presenter
//   to render the Movie2D node.
//
// Created: August, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-CSELT.h"
#include "Movie2DProxy.h"
#include "Stream.h"
#include "trace.h"


// =====================================================================

//
// PresenterCSELT::InitMovie2D
//
// Initialise the proxy data structure and attach it to the render data
// pointer of the given media object node.
//

void PresenterCSELT::InitMovie2D(Movie2D* pObj)
{
	pObj->SetPresenterData(new Movie2DProxy(pObj));
}


// =====================================================================

//
// Movie2DProxy::Movie2DProxy
//
// Class constructor.
//

Movie2DProxy::Movie2DProxy(MediaObject* pMO) : DrawableImage(pMO)
{
	m_firstTimeAtStartTime = TRUE;
	m_firstTimeAtStopTime  = TRUE;
	m_endOfTimeReached = FALSE;
	m_startTime = 0;
	m_timesStarted = 0;
} /* Movie2DProxy::Movie2DProxy */


// =====================================================================

//
// Movie2DProxy::Print
//
// Print out the fields and corresponding values for the media object
// derived node that this object is a proxy for, indented according to
// the given depth.
//

void Movie2DProxy::Print(FILE* outfile, int depth) const
{
	fprintf(outfile, "%sMovie2D {\n", IndentPrefix(depth));
	const char* indentStr = IndentPrefix(depth + 1);
	fprintf(outfile, "%sobjectDescriptorID %d\n", indentStr,
			(int)GetMovie2D()->GetObjectDescriptorID());
	fprintf(outfile, "%sloop %s\n", indentStr, BoolToStr(GetMovie2D()->loop));
	fprintf(outfile, "%sspeed %.3f\n", indentStr, (float)GetMovie2D()->speed);
	fprintf(outfile, "%sstartTime %.2lf\n", indentStr, (double)(Time)GetMovie2D()->startTime);
	fprintf(outfile, "%sstopTime %.2lf\n", indentStr, (double)(Time)GetMovie2D()->stopTime);
	fprintf(outfile, "%s}\n", IndentPrefix(depth));
} /* Movie2DProxy::Print */


// =====================================================================

//
// Movie2DProxy::OnEndOfStream
//
// This method is called when the end of a stream is detected in either
// the DrawableImage::PreRender() method or the DrawableImage::Draw()
// method.
//

void Movie2DProxy::OnEndOfStream()
{
	// Call the base class's method
	DrawableImage::OnEndOfStream();

	// Set flags and events which are required at the end of a stream.
	m_firstTimeAtStopTime = FALSE;
	m_firstTimeAtStartTime = TRUE;
	m_endOfTimeReached = TRUE;
	GetMovie2D()->isActive = FALSE;
} /* Movie2DProxy::OnEndOfStream */


// =====================================================================

//
// Movie2DProxy::PreRender
//
// Decides when the start and stop the movie, based on the values of
// the startTime and stopTime fields in the Movie2D node.  If the node
// is active, this method calls the PreRender() and UpdateBoundingRect()
// methods from the base class.
//

void Movie2DProxy::PreRender(Effects const& effects)
{
	// This gets the time since the start of the scene
	// (convert the time to seconds, as used by BIFS)
	double currentTime = (double)GetPresenter()->GetTime() / 1000;

	// If the start time has changed since this was last called, reset things
	if ((double)GetMovie2D()->startTime != m_startTime)
		m_endOfTimeReached = FALSE;
	m_startTime = (double)GetMovie2D()->startTime;

	// Only proceed if the end time has not been reached
	if (m_endOfTimeReached)
		return;

	// Only proceed if the start time has been reached
	if (currentTime < m_startTime)
		return;

	// Determine if the stop time has been reached
	if ((currentTime >= (double)GetMovie2D()->stopTime) &&
		((double)GetMovie2D()->stopTime >= m_startTime))
	{
		// If this is the first time, generate an isActive=FALSE event,
		// but only if the stream was previously active
		if ((m_firstTimeAtStopTime) && ((Bool)GetMovie2D()->isActive))
			OnEndOfStream();
	}

	// otherwise the movie needs to be played
	else
	{
		// If this is the first time, generate an isActive=TRUE event
		if (m_firstTimeAtStartTime)
		{
			m_firstTimeAtStartTime = FALSE;
			m_firstTimeAtStopTime = TRUE;
			GetMovie2D()->isActive = TRUE;
			m_timesStarted++;

			// If this is not the first time that this has been
			// started, we need to fully disconnect this stream, and
			// then reconnect this stream to the demux
			Movie2D* pMovie2D = GetMovie2D();
			if (m_timesStarted > 1)
			{
				pMovie2D->GetStream()->Terminate();
				pMovie2D->CloseStream();
				pMovie2D->SetStream(pMovie2D->m_copyOfESdesc,
									pMovie2D->m_copyOfFileName);
			}

			// we also need to start the stream
			pMovie2D->OpenStream();

			// and record when we started playing (in msec)
			m_startUpDelay = 1000 * m_startTime;

			// allow a bit of time for the decoder to start up
			// (there should really be another way to do this!)
			m_startUpDelay += 300;

			TRACE_PRINT1("H263: Startup delay = %ld ms\n", m_startUpDelay);
		}

		DrawableImage::PreRender(effects);
		UpdateBoundingRect();
	}
} /* Movie2DProxy::PreRender */


// =====================================================================
