///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  Nodes-CSELT.cpp
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   This file contains the implementation of the renderer independent
//   portions of the nodes.
//
// Created: August, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "Nodes-CSELT.h"
#include "WrldInfo.h"
#include "NodeCode.h"


// ========================================================================

//
// Interpolate
//
// Given two keys, and their corresponding values, and a fraction value
// which lies between the two key values, return the interpolated value.
//

static float Interpolate(float key1, float keyValue1,
						 float key2, float keyValue2,
						 float fraction)
{
	// Check that the fraction lies between the two key values
	assert((fraction >= key1) && (fraction <= key2));

	// Determine how far between the keys this fraction is,
	// allowing for (nearly) identical key values.
	float intervalFraction;
	float keyDiff = key2 - key1;
	if ((keyDiff > -1e-10) && (keyDiff < 1e-10))
		intervalFraction = 0;
	else
		intervalFraction = (fraction - key1) / keyDiff;

	// Interpolate between the key values using this fraction
	float newValue = (keyValue2 - keyValue1) * intervalFraction + keyValue1;

	return newValue;
} /* Interpolate */


// ========================================================================

//
// Set up the BIFS tables for the Appearance2D node.
//

BIFS_IMPLEMENT_NODE_START(Appearance2D)
	BIFS_FIELD(Appearance2D, material,			0, 0, 0, 0, 0, 0)
	BIFS_FIELD(Appearance2D, texture,			1, 1, 1, 0, 0, 0)
	BIFS_FIELD(Appearance2D, textureTransform,	2, 2, 2, 0, 0, 0)
	BIFS_FIELD(Appearance2D, properties,		3, 3, 3, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Appearance2D)


// ========================================================================

//
// Set up the BIFS tables for the AudioSource node.
//

BIFS_IMPLEMENT_NODE_START(AudioSource)
	BIFS_FIELD(AudioSource, objectDescriptorID,	0, 0, 0, 0, 0, 0)
	BIFS_FIELD(AudioSource, pitch,				1, 1, 1, 0, 0, 0)
	BIFS_FIELD(AudioSource, startTime,			2, 2, 2, 0, 0, 0)
	BIFS_FIELD(AudioSource, stopTime,			3, 3, 3, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(AudioSource)


// ========================================================================

//
// Set up the BIFS tables for the Background2D node.
//

BIFS_IMPLEMENT_NODE_START(Background2D)
	BIFS_EVENT(Background2D, set_bind,			     0,    0, 0, 0)
	BIFS_FIELD(Background2D, objectDescriptorID, 0,  1, 0, 0, 0, 0)
	BIFS_FIELD(Background2D, isBound,			 1, -1, 1, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Background2D)


// ========================================================================

//
// Set up the BIFS tables for the Circle node.
//

BIFS_IMPLEMENT_NODE_START(Circle)
	BIFS_FIELD(Circle, radius, 0, 0, 0, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Circle)


// ========================================================================

//
// Set up the BIFS tables for the Color node.
//

BIFS_IMPLEMENT_NODE_START(Color)
	BIFS_FIELD(Color, color, 0, 0, 0, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Color)


// ========================================================================

//
// Set up the BIFS tables for the ColorInterpolator node.
//

BIFS_IMPLEMENT_NODE_START(ColorInterpolator)
	BIFS_EVENT(ColorInterpolator, set_fraction,	     0,    0, 0, 0)
	BIFS_FIELD(ColorInterpolator, key,			 0,  1, 0, 0, 0, 0)
	BIFS_FIELD(ColorInterpolator, keyValue,		 1,  2, 1, 0, 0, 0)
	BIFS_FIELD(ColorInterpolator, value_changed, 2, -1, 2, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(ColorInterpolator)


// ========================================================================

//
// ColorInterpolator::Onset_fraction
//
// This function is called when the set_fraction data member has changed,
// and it has already been updated when this function is called.  When
// this happens, calculate the value that corresponds with this fraction,
// and update the value_changed data member.
//

void ColorInterpolator::Onset_fraction()
{
	// Ensure that the key and keyValue arrays are the same length,
	// and that they are not empty.
	assert(key.GetLength() == keyValue.GetLength());
	assert(key.GetLength() != 0);

	// The given fraction is less than the specified range
	if ((float)set_fraction < key[0])
		value_changed = keyValue[0];

	// The given fraction is greater than the specified range
	else if ((float)set_fraction >= key[key.GetLength() - 1])
		value_changed = keyValue[key.GetLength() - 1];

	// The given fraction lies in the specified range of key values
	else
	{
		for (int k = 1; k < key.GetLength(); k++)
		{
			// Find the key values the fraction lies between
			if (((float)set_fraction >= key[k - 1]) &&
				((float)set_fraction < key[k]))
			{
				// and then interpolate the new value
				Color3f newValue;
				newValue.r = Interpolate(key[k - 1], keyValue[k - 1].r,
										 key[k], keyValue[k].r,
										 (float)set_fraction);
				newValue.g = Interpolate(key[k - 1], keyValue[k - 1].g,
										 key[k], keyValue[k].g,
										 (float)set_fraction);
				newValue.b = Interpolate(key[k - 1], keyValue[k - 1].b,
										 key[k], keyValue[k].b,
										 (float)set_fraction);
				value_changed = newValue;
				break;
			}
		}
	}
} /* ColorInterpolator::Onset_fraction */


// ========================================================================

//
// Set up the BIFS tables for the Coordinate2D node.
//

BIFS_IMPLEMENT_NODE_START(Coordinate2D)
	BIFS_FIELD(Coordinate2D, point,		0, 0, 0, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Coordinate2D)


// ========================================================================

//
// Set up the BIFS tables for the Curve2D node.
//

BIFS_IMPLEMENT_NODE_START(Curve2D)
	BIFS_FIELD(Curve2D, points,		0, 0, 0, 0, 0, 0)
	BIFS_FIELD(Curve2D, fineness,	1, 1, 1, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Curve2D)


// ========================================================================

//
// Set up the BIFS tables for the FontStyle2D node.
//

BIFS_IMPLEMENT_NODE_START(FontStyle2D)
	BIFS_FIELD(FontStyle2D, type,		0, 0, 0, 0, 0, 0)
	BIFS_FIELD(FontStyle2D, style,		1, 1, 1, 0, 0, 0)
	BIFS_FIELD(FontStyle2D, size,		2, 2, 2, 0, 0, 0)
	BIFS_FIELD(FontStyle2D, hSpacing,	3, 3, 3, 0, 0, 0)
	BIFS_FIELD(FontStyle2D, vSpacing,	4, 4, 4, 0, 0, 0)
	BIFS_FIELD(FontStyle2D, language,	5, 5, 5, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(FontStyle2D)


// ========================================================================

//
// Set up the BIFS tables for the Group2D node.
//

BIFS_IMPLEMENT_NODE_START(Group2D)
    BIFS_EVENT(Group2D, addChildren,	    0,     0, 0, 0)
    BIFS_EVENT(Group2D, removeChildren,     1,     0, 0, 0)
	BIFS_CHILDREN(Group2D, children,	0,  2,  0, 0)
	BIFS_FIELD(Group2D, bboxCenter,		1, -1, -1, 0, 0, 0)
	BIFS_FIELD(Group2D, bboxSize,		2, -1, -1, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Group2D)


// ========================================================================

//
// Set up the BIFS tables for the Image2D node.
//

BIFS_IMPLEMENT_NODE_START(Image2D)
	BIFS_FIELD(Image2D, objectDescriptorID,	0, 0, 0, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Image2D)


// ========================================================================

//
// Set up the BIFS tables for the Image2D node.
//

BIFS_IMPLEMENT_NODE_START(IndexedLineSet2D)
//	BIFS_EVENT(IndexedLineSet2D, set_colorIndex,	    0,     0, 0, 0)
//	BIFS_EVENT(IndexedLineSet2D, set_coordIndex,	    1,     0, 0, 0)
	BIFS_FIELD(IndexedLineSet2D, color,				0,  2,  0, 0, 0, 0)
	BIFS_FIELD(IndexedLineSet2D, coord,				1,  3,  1, 0, 0, 0)
	BIFS_FIELD(IndexedLineSet2D, colorIndex,		2, -1, -1, 0, 0, 0)
	BIFS_FIELD(IndexedLineSet2D, colorPerVertex,	3, -1, -1, 0, 0, 0)
	BIFS_FIELD(IndexedLineSet2D, coordIndex,		4, -1, -1, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(IndexedLineSet2D)


// ========================================================================

//
// Set up the BIFS tables for the Layout node.
//

BIFS_IMPLEMENT_NODE_START(Layout)
	BIFS_CHILDREN(Layout, children,		0, 0, 0, 0)
	BIFS_FIELD(Layout, mode,			1, 1, 1, 0, 0, 0)
	BIFS_FIELD(Layout, wrap,			2, 2, 2, 0, 0, 0)
	BIFS_FIELD(Layout, width,			3, 3, 3, 0, 0, 0)
	BIFS_FIELD(Layout, height,			4, 4, 4, 0, 0, 0)
	BIFS_FIELD(Layout, lineSpacing,		5, 5, 5, 0, 0, 0)
	BIFS_EVENT(Layout, addChildren,		   6,    0, 0, 0)
	BIFS_EVENT(Layout, removeChildren,	   7,    0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Layout)


// ========================================================================

void Layout::OnaddChildren()
{
}


// ========================================================================

void Layout::OnremoveChildren()
{
}

// ========================================================================

//
// Set up the BIFS tables for the Material2D node.
//

BIFS_IMPLEMENT_NODE_START(Material2D)
	BIFS_FIELD(Material2D, diffuseColor, 0, 0, 0, 0, 0, 0)
	BIFS_FIELD(Material2D, transparency, 1, 1, 1, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Material2D)


// ========================================================================

//
// Set up the BIFS tables for the Movie2D node.
//

BIFS_IMPLEMENT_NODE_START(Movie2D)
	BIFS_FIELD(Movie2D, loop,				0,  0, 0, 0, 0, 0)
	BIFS_FIELD(Movie2D, speed,				1,  1, 1, 0, 0, 0)
	BIFS_FIELD(Movie2D, startTime,			2,  2, 2, 0, 0, 0)
	BIFS_FIELD(Movie2D, stopTime,			3,  3, 3, 0, 0, 0)
	BIFS_FIELD(Movie2D, objectDescriptorID,	4,  4, 4, 0, 0, 0)
	BIFS_FIELD(Movie2D, duration_changed,	5, -1, 5, 0, 0, 0)
	BIFS_FIELD(Movie2D, isActive,			6, -1, 6, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Movie2D)


// ========================================================================

//
// Set up the BIFS tables for the PointSet2D node.
//

BIFS_IMPLEMENT_NODE_START(PointSet2D)
	BIFS_FIELD(PointSet2D, color,			 0, 0, 0, 0, 0, 0)
	BIFS_FIELD(PointSet2D, coord,			 1, 1, 1, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(PointSet2D)


// ========================================================================

//
// Set up the BIFS tables for the Position2DInterpolator node.
//

BIFS_IMPLEMENT_NODE_START(Position2DInterpolator)
	BIFS_EVENT(Position2DInterpolator, set_fraction,      0,    0, 0, 0)
	BIFS_FIELD(Position2DInterpolator, key,			  0,  1, 0, 0, 0, 0)
	BIFS_FIELD(Position2DInterpolator, keyValue,	  1,  2, 1, 0, 0, 0)
	BIFS_FIELD(Position2DInterpolator, value_changed, 2, -1, 2, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Position2DInterpolator)


// ========================================================================

//
// Position2DInterpolator::Onset_fraction
//
// This function is called when the set_fraction data member has changed,
// and it has already been updated when this function is called.  When
// this happens, calculate the value that corresponds with this fraction,
// and update the value_changed data member.
//

void Position2DInterpolator::Onset_fraction()
{
	// Ensure that the key and keyValue arrays are the same length,
	// and that they are not empty.
	assert(key.GetLength() == keyValue.GetLength());
	assert(key.GetLength() != 0);

	// The given fraction is less than the specified range
	if ((float)set_fraction < key[0])
		value_changed = keyValue[0];

	// The given fraction is greater than the specified range
	else if ((float)set_fraction >= key[key.GetLength() - 1])
		value_changed = keyValue[key.GetLength() - 1];

	// The given fraction lies in the specified range of key values
	else
	{
		for (int k = 1; k < key.GetLength(); k++)
		{
			// Find the key values the fraction lies between
			if (((float)set_fraction >= key[k - 1]) &&
				((float)set_fraction < key[k]))
			{
				// and then interpolate the new value
				Vec2f newValue;
				newValue.x = Interpolate(key[k - 1], keyValue[k - 1].x,
										 key[k], keyValue[k].x,
										 (float)set_fraction);
				newValue.y = Interpolate(key[k - 1], keyValue[k - 1].y,
										 key[k], keyValue[k].y,
										 (float)set_fraction);
				value_changed = newValue;
				break;
			}
		}
	}
} /* Position2DInterpolator::Onset_fraction */


// ========================================================================

//
// Set up the BIFS tables for the Properties2D node.
//

BIFS_IMPLEMENT_NODE_START(Properties2D)
	BIFS_FIELD(Properties2D, filled,	0, 0, 0, 0, 0, 0)
	BIFS_FIELD(Properties2D, shadow,	1, 1, 1, 0, 0, 0)
	BIFS_FIELD(Properties2D, shadPos,	2, 2, 2, 0, 0, 0)
	BIFS_FIELD(Properties2D, dotted,	3, 3, 3, 0, 0, 0)
	BIFS_FIELD(Properties2D, width,		4, 4, 4, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Properties2D)


// ========================================================================

//
// Set up the BIFS tables for the Rectangle node.
//

BIFS_IMPLEMENT_NODE_START(Rectangle)
	BIFS_FIELD(Rectangle, size, 0, 0, 0, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Rectangle)


// ========================================================================

//
// Set up the BIFS tables for the ScalarInterpolator node.
//

BIFS_IMPLEMENT_NODE_START(ScalarInterpolator)
	BIFS_EVENT(ScalarInterpolator, set_fraction,	    0,    0, 0, 0)
	BIFS_FIELD(ScalarInterpolator, key,				0,  1, 0, 0, 0, 0)
	BIFS_FIELD(ScalarInterpolator, keyValue,		1,  2, 1, 0, 0, 0)
	BIFS_FIELD(ScalarInterpolator, value_changed,	2, -1, 2, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(ScalarInterpolator)


// ========================================================================

//
// ScalarInterpolator::Onset_fraction
//
// This function is called when the set_fraction data member has changed,
// and it has already been updated when this function is called.  When
// this happens, calculate the value that corresponds with this fraction,
// and update the value_changed data member.
//

void ScalarInterpolator::Onset_fraction()
{
	// Ensure that the key and keyValue arrays are the same length,
	// and that they are not empty.
	assert(key.GetLength() == keyValue.GetLength());
	assert(key.GetLength() != 0);

	// The given fraction is less than the specified range
	if ((float)set_fraction < key[0])
		value_changed = keyValue[0];

	// The given fraction is greater than the specified range
	else if ((float)set_fraction >= key[key.GetLength() - 1])
		value_changed = keyValue[key.GetLength() - 1];

	// The given fraction lies in the specified range of key values
	else
	{
		for (int k = 1; k < key.GetLength(); k++)
		{
			// Find the key values the fraction lies between
			if (((float)set_fraction >= key[k - 1]) &&
				((float)set_fraction < key[k]))
			{
				// and then interpolate the new value
				float newValue = Interpolate(key[k - 1], keyValue[k - 1],
											 key[k], keyValue[k],
											 (float)set_fraction);
				value_changed = newValue;
				break;
			}
		}
	}
} /* ScalarInterpolator::Onset_fraction */


// ========================================================================

//
// Set up the BIFS tables for the Scroll node.
//

BIFS_IMPLEMENT_NODE_START(Scroll)
    BIFS_EVENT(Scroll, addChildren,		   0,    0, 0, 0)
    BIFS_EVENT(Scroll, removeChildren,	   1,    0, 0, 0)
	BIFS_CHILDREN(Scroll, children,		0, 2, 0, 0)
	BIFS_FIELD(Scroll, smoothScroll,	1, 3, 1, 0, 0, 0)
	BIFS_FIELD(Scroll, loop,			2, 4, 2, 0, 0, 0)
	BIFS_FIELD(Scroll, scrollXRate,		3, 5, 3, 0, 0, 0)
	BIFS_FIELD(Scroll, scrollYRate,		4, 6, 4, 0, 0, 0)
	BIFS_FIELD(Scroll, width,			5, 7, 5, 0, 0, 0)
	BIFS_FIELD(Scroll, height,			6, 8, 6, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Scroll)


// ========================================================================

//
// Set up the BIFS tables for the Shape2D node.
//

BIFS_IMPLEMENT_NODE_START(Shape2D)
	BIFS_FIELD(Shape2D, appearance,	0, 0, 0, 0, 0, 0)
	BIFS_FIELD(Shape2D, geometry,	1, 1, 1, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Shape2D)


// ========================================================================

//
// Set up the BIFS tables for the Sound2D node.
//

BIFS_IMPLEMENT_NODE_START(Sound2D)
	BIFS_FIELD(Sound2D, location,	0, 0, 0, 0, 0, 0)
	BIFS_FIELD(Sound2D, intensity,	1, 1, 1, 0, 0, 0)
	BIFS_FIELD(Sound2D, sound,		2, 2, 2, 0, 0, 0)
	BIFS_FIELD(Sound2D, spatialize,	3, 3, 3, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Sound2D)


// ========================================================================

//
// Set up the BIFS tables for the Switch node.
//

BIFS_IMPLEMENT_NODE_START(Switch)
	BIFS_CHILDREN(Switch, choice,	0, 0, 0, 0)
	BIFS_FIELD(Switch, whichChoice,	1, 1, 1, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Switch)


// ========================================================================

//
// Set up the BIFS tables for the Text2D node.
//

BIFS_IMPLEMENT_NODE_START(Text2D)
	BIFS_FIELD(Text2D, data,		0, 0, 0, 0, 0, 0)
	BIFS_FIELD(Text2D, length,		1, 1, 1, 0, 0, 0)
	BIFS_FIELD(Text2D, fontStyle,	2, 2, 2, 0, 0, 0)
	BIFS_FIELD(Text2D, ucs_2,		3, 3, 3, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Text2D)


// ========================================================================

//
// Set up the BIFS tables for the TextureTransform node.
//

BIFS_IMPLEMENT_NODE_START(TextureTransform)
	BIFS_NULL_ENTRY
BIFS_IMPLEMENT_NODE_END(TextureTransform)


// ========================================================================

//
// Set up the BIFS tables for the TimeSensor node.
//

BIFS_IMPLEMENT_NODE_START(TimeSensor)
	BIFS_FIELD(TimeSensor, cycleInterval,	 0,  0, 0, 0, 0, 0)
	BIFS_FIELD(TimeSensor, enabled,			 1,  1, 1, 0, 0, 0)
	BIFS_FIELD(TimeSensor, loop,			 2,  2, 2, 0, 0, 0)
	BIFS_FIELD(TimeSensor, startTime,		 3,  3, 3, 0, 0, 0)
	BIFS_FIELD(TimeSensor, stopTime,		 4,  4, 4, 0, 0, 0)
	BIFS_FIELD(TimeSensor, cycleTime,		 5, -1, 5, 0, 0, 0 )
	BIFS_FIELD(TimeSensor, fraction_changed, 6, -1, 6, 0, 0, 0 )
	BIFS_FIELD(TimeSensor, isActive,		 7, -1, 7, 0, 0, 0 )
	BIFS_FIELD(TimeSensor, time,			 8, -1, 8, 0, 0, 0 )
BIFS_IMPLEMENT_NODE_END(TimeSensor)


// ========================================================================

//
// Set up the BIFS tables for the TouchSensor2D node.
//

BIFS_IMPLEMENT_NODE_START(TouchSensor2D)
	BIFS_FIELD(TouchSensor2D, enabled,				0,  0, 0, 0, 0, 0)
	BIFS_FIELD(TouchSensor2D, hitPoint_changed,		1, -1, 1, 0, 0, 0)
	BIFS_FIELD(TouchSensor2D, hitTexCoord_changed,	2, -1, 2, 0, 0, 0)
	BIFS_FIELD(TouchSensor2D, isActive,				3, -1, 3, 0, 0, 0)
	BIFS_FIELD(TouchSensor2D, isOver,				4, -1, 4, 0, 0, 0)
	BIFS_FIELD(TouchSensor2D, touchTime,			5, -1, 5, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(TouchSensor2D)


// ========================================================================

//
// Set up the BIFS tables for the Transform2D node.
//

BIFS_IMPLEMENT_NODE_START(Transform2D)
    BIFS_EVENT(Transform2D, addChildren,		    0,     0, 0, 0)
    BIFS_EVENT(Transform2D, removeChildren,		    1,     0, 0, 0)
	BIFS_FIELD(Transform2D, center,				0,  2,  0, 0, 0, 0)
	BIFS_CHILDREN(Transform2D, children,		1,  3,  1, 0)
	BIFS_FIELD(Transform2D, rotationAngle,		2,  4,  2, 0, 0, 0)
	BIFS_FIELD(Transform2D, scale,				3,  5,  3, 0, 0, 0)
	BIFS_FIELD(Transform2D, scaleOrientation,	4,  6,  4, 0, 0, 0)
	BIFS_FIELD(Transform2D, drawingOrder,		5,  7,  5, 0, 0, 0)
	BIFS_FIELD(Transform2D, translation,		6,  8,  6, 0, 0, 0)
	BIFS_FIELD(Transform2D, bboxCenter,			7, -1, -1, 0, 0, 0)
	BIFS_FIELD(Transform2D, bboxSize,			8, -1, -1, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(Transform2D)


// ========================================================================

//
// Set up the BIFS tables for the XLayout node.
//

BIFS_IMPLEMENT_NODE_START(XLayout)
	BIFS_CHILDREN(XLayout, children,	0, -1, -1, 0)
	BIFS_FIELD(XLayout, width,			1,  0,  0, 0, 0, 0)
	BIFS_FIELD(XLayout, height,			2,  1,  1, 0, 0, 0)
	BIFS_FIELD(XLayout, groups,			3, -1, -1, 0, 0, 0)
	BIFS_FIELD(XLayout, constraint,		4, -1, -1, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(XLayout)


// ========================================================================

//
// Set up the BIFS tables for the WorldInfo node.
//

BIFS_IMPLEMENT_NODE_START(WorldInfo)
	BIFS_FIELD(WorldInfo, info,		0, -1, -1, 0, 0, 0)
	BIFS_FIELD(WorldInfo, title,	1, -1, -1, 0, 0, 0)
BIFS_IMPLEMENT_NODE_END(WorldInfo)


// ========================================================================
