///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File: Presenter-CSELT.cpp
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of a class which represents a Presenter,
//   composed of a visual and audio rendering engine.
//
// Created: April, 1997
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "Presenter-CSELT.h"
#include "Executiv.h"
#include "MediaObj.h"
#include "Caps.h"
#include "Clock.h"
#include "trace.h"
#include "DrawableNode.h"
#include "ThreadList.h"
#include "UserEventHandlerProxy.h"
#include "MPEG4Player.h"
#include "MainFrm.h"


#ifdef WRITE_TRACE
FILE* traceFile;
#endif


// ========================================================================

//
// Presenter::Create
//
// Return a pointer to a presenter.  Use this to construct
// a presenterr appropriate for the current platform.  A pointer
// to the executive, and a platform specific parameter are given.
//

Presenter* Presenter::Create(Executive* pExecutive, void* pInfo)
{
	return new PresenterCSELT(pExecutive, (HWND)pInfo);
} /* Presenter::Create */


// ========================================================================

//
// PresenterCSELT::PresenterCSELT
//
// Class constructor
//  - instantiates the visual and audio renderers
//

PresenterCSELT::PresenterCSELT(Executive* pExecutive, HWND hWnd)
{
	TRACE_PRINT("Entering PresenterCSELT::PresenterCSELT()\n");

	m_bActive = FALSE;
	m_pScene = NULL;
	m_actualFrameRate = 0;
	m_currentTime = 0;

	// The default frame rate is 15 frames per second
	m_frameRate = 15;
	m_newFrameRate = 15;

	// Remember the pointer to the Executive
	m_pExecutive = pExecutive;

	// Instantiate the renderers
	m_pVisualRenderer = VisualRenderer::Create(this, (void*)hWnd);
	m_pAudioRenderer  = AudioRenderer::Create(this, (void*)hWnd);

	TRACE_PRINT("Leaving PresenterCSELT::PresenterCSELT()\n");
} /* PresenterCSELT::PresenterCSELT */


// ========================================================================

//
// PresenterCSELT::~PresenterCSELT
//
// Class destructor - calls terminate if it hasn't already been called.
//

PresenterCSELT::~PresenterCSELT()
{
	TRACE_PRINT("Entering PresenterCSELT::~PresenterCSELT()\n");

	// Terminate the presenter (just to make sure that it is done)
	if (m_bActive)
		Terminate();

	// Delete the visual and audio renderers
	delete m_pVisualRenderer;
	delete m_pAudioRenderer;

	TRACE_PRINT("Leaving PresenterCSELT::~PresenterCSELT()\n");
} /* PresenterCSELT::~PresenterCSELT */


// ========================================================================

//
// PresenterCSELT::GetCapabilities
//
// Combine the capabilities of the video and audio renderers.
//

void PresenterCSELT::GetCapabilities (Capabilities* pCaps)
{
	int i, j;
	Capabilities caps;

	// Get the capabilities of the two renderers
	m_pVisualRenderer->GetCapabilities(pCaps); 
	m_pAudioRenderer->GetCapabilities(&caps);
	
	// Combine the capabilities into a single object
	j = pCaps->GetFormatCapabilityLength();
	for (i = 0; i < caps.GetFormatCapabilityLength(); i++)
		pCaps->AddFormatCapability(caps.GetFormatCapability(i), j++);
} /* PresenterCSELT::GetCapabilities */


// ========================================================================

//
// PresenterCSELT::Init
//
// Initialise the renderers.  As both the audio and visual renderers
// run in their own threads, then the presenter does not need to.
// This method will be called once for every file that is opened.
//

void PresenterCSELT::Init(MediaObject* pScene)
{
	TRACE_PRINT("Entering PresenterCSELT::Init()\n");

	// Remember the pointer to the root scene object
	m_pScene = pScene;

	// Flag that the presenter is now active
	m_bActive = TRUE;

	// Initialise the renderers
	m_pVisualRenderer->Init();
	m_pAudioRenderer->Init();

	// Start the main thread of the presenter
	ZThread<PresenterCSELT> thread(Main, this);
	int hThread = thread.Begin();
  	if (hThread == -1)
    	throw;
	else
		AddToThreadList("Presenter", hThread);

	TRACE_PRINT("Leaving PresenterCSELT::Init()\n");
} /* PresenterCSELT::Init */


// ========================================================================

//
// PresenterCSELT::Terminate
//
// Terminates the visual and audio rendering, performs cleanups.
// This method will be called once for every file that is closed.
//

void PresenterCSELT::Terminate()
{
	TRACE_PRINT("Entering PresenterCSELT::Terminate():\n");

	// Flag that the presenter is no longer active
	m_bActive = FALSE;

	// Wait for this thread to terminate
	m_mtxThread.P ();
	m_mtxThread.V ();

	// Stop the visual and audio renders
	m_pVisualRenderer->Terminate();
	m_pAudioRenderer->Terminate();

	TRACE_PRINT("Leaving PresenterCSELT::Terminate():\n");
} /* PresenterCSELT::Terminate */


// ========================================================================

//
// PresenterCSELT::OnUserEvent
//
// Called by the executive when a user event occurs.  Add the details of
// this event to the list of events that have arrived to be process,
// since the scene was last drawn.
//
// This function will be called by the thread which is running the frame
// application.
//

void PresenterCSELT::OnUserEvent(USER_EVENT event, int x, int y,
								 UINT controlKeysStatus)
{
	// Only deal with events when the presenter is active
	if (m_bActive)
	{
		TRACE_PRINT2("Pres: Received event at (%d,%d)\n", x, y);

		m_mtxEventArray.P();
		m_eventArray += new EventInfo(event, x, y, controlKeysStatus);
		m_mtxEventArray.V();
	}
} /* PresenterCSELT::OnUserEvent */


// ========================================================================

//
// PresenterCSELT::RegisterEventHandler
//
// Add the given UserEventHandlerProxy node to the list
// of nodes which need to be informed of user events.
//

void PresenterCSELT::RegisterEventHandler(UserEventHandlerProxy* pNode)
{
	m_mtxEventHandlersArray.P();

	// Only add this node if it is not already in the list
	BOOL inList = FALSE;
	for (int i = 0; i < m_eventHandlersArray.GetLength(); i++)
		if (m_eventHandlersArray[i] == pNode)
			inList = TRUE;

	if (!inList)
		m_eventHandlersArray += pNode;

	m_mtxEventHandlersArray.V();
} /* PresenterCSELT::RegisterEventHandler */


// ========================================================================

//
// PresenterCSELT::UnregisterEventHandler
//
// Remove the given UserEventHandlerProxy node from the
// list of nodes which need to be informed of user events.
//

void PresenterCSELT::UnregisterEventHandler(UserEventHandlerProxy* pNode)
{
	m_mtxEventHandlersArray.P();
	m_eventHandlersArray -= pNode;
	m_mtxEventHandlersArray.V();
} /* PresenterCSELT::UnregisterEventHandler */


// ========================================================================

//
// PresenterCSELT::ProcessEvents
//
// Go through all of the events that are in the event array, and for each
// event do the following:
//   - Translate (x,y) to BIFS coordinates
//   - Go through the list of drawable nodes, looking for the first which
//     contains the given point.
//   - For the found node, search for a TouchSensor node to handle this
//     event.
//
// This function will be called by the thread which is running the
// presenter.
//

void PresenterCSELT::ProcessEvents()
{
	m_mtxEventArray.P();

	while (m_eventArray.GetLength() > 0)
	{
		// Get the first event
		EventInfo* pEvent = m_eventArray[0];
		pEvent->pass = 1;

		TRACE_PRINT2("Pres: Processing event at (%d,%d)\n",
					 pEvent->m_x, pEvent->m_y);

		// Translate the coordinates from screen to BIFS
		pEvent->m_x -= m_pVisualRenderer->GetScreenWidth() / 2;
		pEvent->m_y  = m_pVisualRenderer->GetScreenHeight() / 2 - pEvent->m_y;

		// Look through the drawable nodes for a node under this point
		DrawableNode* pDrawable;
		pDrawable = m_pVisualRenderer->FindNode(pEvent->m_x, pEvent->m_y);

		// If a node has been found, look for a sensor node to handle
		// this event
		if (pDrawable != NULL)
			pDrawable->FindTouchSensor(pEvent);

		// Call the OnUserEvent method of all nodes which have registered
		pEvent->pass = 2;
		for (int i = 0; i < m_eventHandlersArray.GetLength(); i++)
			m_eventHandlersArray[i]->OnUserEvent(pEvent);

		// We have finished with this event, remove it from the array
		m_eventArray.DeleteFrom(0);
		delete pEvent;
	}

	m_mtxEventArray.V();
} /* PresenterCSELT::ProcessEvents */


// =====================================================================

//
// PresenterCSELT::Main
//
// This is the main loop of the presenter thread.  Render the scene
// at the specified frame rate.
//

void PresenterCSELT::Main()
{
	// Mark the thread as running
	m_mtxThread.P();

	TRACE_PRINT("Entering PresenterCSELT::Main():\n");

	// These variables relate to the timing of the presenter
	unsigned long scenesDrawnSinceRateChage = 0;
	unsigned long startingTime = GetExecutive()->GetClock()->GetTime();
	float sceneUpdatePeriod = 1000.0 / m_frameRate;
	ResetFrameRateCalculation();
	DWORD scenesDrawnSinceStart = 0;

	// The default effects given to PreRender()
	Effects defEffects;
	defEffects.transform.SetIdentity();
	defEffects.drawOrder = 0;
	defEffects.isSwitchedOn = TRUE;

	// Erase the display
	m_pVisualRenderer->Clear();
	m_pVisualRenderer->Paint();

#ifdef WRITE_TRACE
	// Print out the scene before starting
	((Traversable*)m_pScene->GetPresenterData())->Print(traceFile, 0);
#endif

	// This loop runs while the presenter is active, which is up to
	// either the next Terminate() call, or all nodes have finished.
	while (m_bActive)
	{
		// Get a pointer to the current root proxy object (will it change?)
		Traversable* pRootProxy = (Traversable*)m_pScene->GetPresenterData();

		// If the frame rate has changed, reset the timing related variables.
		if (m_frameRate != m_newFrameRate)
		{
//			TRACE_PRINT2("Changing frame rate from %.2f to %.2f\n",
//						 m_frameRate, m_newFrameRate);
			m_frameRate = m_newFrameRate;
			sceneUpdatePeriod = 1000.0 / m_frameRate;
			scenesDrawnSinceRateChage = 0;
			startingTime = GetExecutive()->GetClock()->GetTime();
		}

		// Update the current time value, which is the time since the
		// presenter was started.
		m_currentTime = GetExecutive()->GetClock()->GetTime();

		// Find the nodes which have had user events occur above them
		ProcessEvents();

		// Update the TimeSensor nodes in the scene
		pRootProxy->UpdateTimeSensor();

		// Render the scene
		TRACE_PRINT("Before scene PreRender.\n");
		m_pVisualRenderer->ResetDrawList();
		pRootProxy->PreRender(defEffects);
		TRACE_PRINT("After scene PreRender, before scene draw.\n");
		m_pVisualRenderer->DrawScene();
		scenesDrawnSinceStart++;
		TRACE_PRINT("After scene draw.\n");

		// If this is the first time that the scene has been rendered,
		// then we need to start the audio renderer
		if (scenesDrawnSinceStart == 1)
			m_pAudioRenderer->Start();

		// If all of the nodes in the scene have finished, then the
		// Presenter is also finished.
		if (pRootProxy->IsFinished())
		{
			TRACE_PRINT("Pres: The scene is finished.\n");
			m_bActive = FALSE;
			break;
		}

		// Determine how long to wait before starting the next composition
		//  - do this by determining the number of clock ticks which have
		//    elapsed since starting, and the number of the clock tick counter
		//    when the next composition should start.
		//
		// The time from the start of the thread is used instead of the time
		// from the start of this composition, to get more accurate timing,
		// avoiding jitter in small scale time readings.
		scenesDrawnSinceRateChage++;
		unsigned int elapsedTC = GetExecutive()->GetClock()->GetTime() - startingTime;
		unsigned int startNextTC = scenesDrawnSinceRateChage * (unsigned int)sceneUpdatePeriod;

		// Determine the current actual frame rate.
		UpdateFrameRateCalculation(elapsedTC);

		// Only sleep if ahead of schedule
		TRACE_PRINT1("Pres: Sleeping for %d ms.\n", startNextTC - elapsedTC);
		if (startNextTC > elapsedTC)
			Sleep(startNextTC - elapsedTC);
	}

	TRACE_PRINT("Leaving PresenterCSELT::Main():\n");

	// Flag that the thread is terminating
	m_mtxThread.V();

	GetApp()->OnSceneFinished();

} /* PresenterCSELT::Main */


// ========================================================================

//
// PresenterCSELT::WaitTillFinished
//
// This function will only return when all of the nodes in the
// scene have finished, which means that the Presenter thread
// has finished.
//

void PresenterCSELT::WaitTillFinished()
{
	m_mtxThread.P ();
	m_mtxThread.V ();
} /* PresenterCSELT::WaitTillFinished */


// ========================================================================

//
// PresenterCSELT::ResetFrameRateCalculation
//
// Reset the frame rate calculation variables.
//

void PresenterCSELT::ResetFrameRateCalculation()
{
	m_nextElapsedTime = NUM_ELAPSED_TIMES - 1;
	for (int i = 0; i < (int)NUM_ELAPSED_TIMES; i++)
		m_elapsedTimes[i] = 0;
	m_actualFrameRate = 0;
} /* PresenterCSELT::ResetFrameRateCalculation */


// =====================================================================

//
// PresenterCSELT::UpdateFrameRateCalculation
//
// Update the frame rate calculation, by giving the next time at
// which a frame has been presented.
//

void PresenterCSELT::UpdateFrameRateCalculation(unsigned int displayTime)
{
	// Put the given time into the ring buffer of the most recent
	// NUM_ELAPSED_TIMES times.
	m_nextElapsedTime = (m_nextElapsedTime + 1) % NUM_ELAPSED_TIMES;
	m_elapsedTimes[m_nextElapsedTime] = displayTime;

	// Look back through the recent display times, for a frame which
	// was displayed at least a second (1000 ms) ago.  This is used to
	// determine the average frame rate for the past second.
	int i = m_nextElapsedTime;
	int frames = 0;
	while ((displayTime - m_elapsedTimes[i] < 1000) &&
		   (frames < NUM_ELAPSED_TIMES))
	{
		i = (i == 0) ? (NUM_ELAPSED_TIMES - 1) : (i - 1);
		frames++;
	}

	// And now calculate the frame rate
	if (displayTime == m_elapsedTimes[i])
		m_actualFrameRate = m_frameRate;
	else
		m_actualFrameRate = 1000.0 * (float)frames / (float)(displayTime - m_elapsedTimes[i]);
} /* PresenterCSELT::UpdateFrameRateCalculation */

// =====================================================================

//
// PresenterCSELT::DeleteObject
//
// The generic rendering proxy object deleter.
//

void PresenterCSELT::DeleteObject(MediaObject* pObject)
{
	Traversable* pProxy = (Traversable*)(pObject->GetPresenterData());
	if (pProxy != NULL)
		delete pProxy;
} /* PresenterCSELT::DeleteObject */


// =====================================================================


// Emacs uses these...
// Local Variables:
// mode: c++
// tab-width: 4
// End:
