///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File: Presenter-CSELT.h
//
// Authors: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   This class defines a Presenter, composed of a visual and audio
//   rendering engine.
//
// Created: August, 1997
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#ifndef PRESENTER_CSELT_H
#define PRESENTER_CSELT_H


#include "BasePresenter.h"
#include "AudRend-CSELT.h"
#include "VisRend.h"
#include "ztl.h"
#include "Im1defs.h"


class Executive;
class NodeDepthInfo;
class Capabilities;
class MediaObject;
class AudioRenderer;
class VisualRenderer;
class UserEventHandlerProxy;


//
// A class which stores all of the details of an event.
//
class EventInfo
{
public:
	EventInfo(USER_EVENT event, int x, int y, UINT controlKeysStatus)
	{
		m_event = event;
		m_x = x;
		m_y = y;
		m_controlKeysStatus = controlKeysStatus;
	}

	USER_EVENT m_event;
	int m_x;
	int m_y;
	UINT m_controlKeysStatus;
	int pass;
};


//
// The Presenter class
//
class PresenterCSELT : public Presenter
{
	//
	// Public functions
	//
public:

	// Class constructor
	//  - instantiates the visual and audio renderers
	PresenterCSELT(Executive* pExecutive, HWND hWnd);

	// Class destructor - calls terminate if it hasn't already been called
	~PresenterCSELT();

	// Get the capabilities of the video and audio renderers
	void GetCapabilities (Capabilities* pPresentationCaps);

	// Initialise the renderers, and start presenting the scene
	// (whose root object is pointed to by pScene).  This method
	// will be called once for every file that is opened.
	void Init(MediaObject* pScene);

	// Stops rendering, performs cleanups and terminates the object
	// thread.  This method will be called once for every file that
	// is closed.
	void Terminate();

	// Return the current presenter time, which is the time since the
	// presenter was started, and this is only updated before a scene
	// is displayed.  The time returned is in milliseconds.
	DWORD GetTime() { return m_currentTime; }

	// Called by the executive when a user event occurs.
	void OnUserEvent(USER_EVENT event, int x, int y,
					 UINT uControlKeysStatus);

	// Return TRUE if the renderers were initialised successfully.
	BOOL InitialisedOK()
	{ return (m_pAudioRenderer->InitialisedOK() && m_pVisualRenderer->InitialisedOK()); }

	// This is the main loop of the presenter thread.  Render the scene
	// at the specified frame rate.
	void Main();
	static void StartThread(void* params);

	// This function will only return when all of the nodes in the
	// scene have finished, which means that the Presenter thread
	// has finished.
	void WaitTillFinished();

	// Add the given UserEventHandlerProxy node to the list
	// of nodes which need to be informed of user events.
	void RegisterEventHandler(UserEventHandlerProxy* pNode);

	// Remove the given UserEventHandlerProxy node from the
	// list of nodes which need to be informed of user events.
	void UnregisterEventHandler(UserEventHandlerProxy* pNode);

	// Tell the visual renderer to display frames at this rate.  The
	// default rate for the visual renderer is 15 frames per second.
	void SetTargetFrameRate(float newFrameRate)
		{ if (newFrameRate > 0) m_newFrameRate = newFrameRate; }

	// Get the visual renderer's current target frame rate.
	float GetTargetFrameRate() const
		{ return m_newFrameRate; }

	// Return the actual frame rate that the visual renderer is producing.
	// This will differ from the target frame rate if there is not enough
	// computing power.  The rate returned will have been obtained by
	// averaging over the last second.
	float GetActualFrameRate() const
	{ return m_actualFrameRate; }

	// The following functions are called by the constructors of the
	// nodes, and are used to create the proxy object associated with
	// every node which is used to render it.
	// All of these functions take a single argument, being a pointer
	// to the node.
	void InitAppearance2D(class Appearance2D*);
	void InitAudioSource(class AudioSource*);
	void InitBackground2D(class Background2D*);
	void InitCircle(class Circle*);
	void InitColor(class Color*);
	void InitColorInterpolator(class ColorInterpolator*);
	void InitCoordinate2D(class Coordinate2D*);
	void InitCurve2D(class Curve2D*);
	void InitFontStyle2D(class FontStyle2D*);
	void InitGroup2D(class Group2D*);
	void InitImage2D(class Image2D*);
	void InitIndexedLineSet2D(class IndexedLineSet2D*);
	void InitLayout(class Layout*);
	void InitMaterial2D(class Material2D*);
	void InitMovie2D(class Movie2D*);
	void InitPointSet2D(class PointSet2D*);
	void InitPosition2DInterpolator(class Position2DInterpolator*);
	void InitProperties2D(class Properties2D*);
	void InitRectangle(class Rectangle*);
	void InitScalarInterpolator(class ScalarInterpolator*);
	void InitScroll(class Scroll*);
	void InitShape2D(class Shape2D*);
	void InitSound2D(class Sound2D*);
	void InitSwitch(class Switch*);
	void InitText2D(class Text2D*);
	void InitTimeSensor(class TimeSensor*);
	void InitTouchSensor2D(class TouchSensor2D*);
	void InitTransform2D(class Transform2D*);
	void InitWorldInfo(class WorldInfo*);
	void InitXLayout(class XLayout*);

	// This is the generic object destroyer, to be called by the
	// destructors of the nodes.
	void DeleteObject(class MediaObject* pObject);

	//
	// Private implementation details
	//
private:
	void ProcessEvents();

	// Flags the presenter thread when it is time to finish
	BOOL m_bActive;

	// This mutex is used by Terminate() to wait until the thread
	// has finished.
	ZMutex m_mtxThread;

	// The output frame rate (frames per second)
	float m_frameRate;

	// This holds the new value of the frame rate until it is made
	// the current frame rate.
	float m_newFrameRate;

	// This stores the actual frame rate as calculated by the main loop
	// of the presenter.
	float m_actualFrameRate;

	// The time that has elapsed since the presenter was started.
	DWORD m_currentTime;

	// Reset the frame rate calculation variables
	void ResetFrameRateCalculation();

	// Update the frame rate calculation, by giving the next time at
	// which a frame has been presented.
	void UpdateFrameRateCalculation(unsigned int nextTime);

	// Data members pertaining to the calculation of the actual frame
	// rate.
#define NUM_ELAPSED_TIMES 25
	unsigned int m_nextElapsedTime;
	unsigned int m_elapsedTimes[NUM_ELAPSED_TIMES];

	// The array of events that need to be processed
	ZArray<EventInfo*> m_eventArray;
	ZMutex m_mtxEventArray;

	// The array of nodes to be notified of user events
	ZArray<UserEventHandlerProxy*> m_eventHandlersArray;
	ZMutex m_mtxEventHandlersArray;

};


#endif // PRESENTER_CSELT_H


// Emacs uses these...
// Local Variables:
// mode: c++
// tab-width: 4
// End:
