///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  Rectangle.cpp
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of the AudioSource node.
//
// Created: August, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-CSELT.h"
#include "RectangleProxy.h"
//#include "Material2D.h"
#include "trace.h"


// =====================================================================

//
// PresenterCSELT::InitRectangle
//
// Initialise the proxy data structure and attach it to the render data
// pointer of the given media object node.
//

void PresenterCSELT::InitRectangle(Rectangle* pObj)
{
	pObj->SetPresenterData(new RectangleProxy(pObj));
}


// ======================================================================

//
// RectangleProxy::Print
//
// Print out the fields and corresponding values for the media object
// derived node that this object is a proxy for, indented according to
// the given depth.
//

void RectangleProxy::Print(FILE* outfile, int depth) const
{
	fprintf(outfile, "Rectangle {\n");
	fprintf(outfile, "%ssize %.2f %.2f\n", IndentPrefix(depth + 1),
			((Vec2f)GetRectangle()->size).x,
			((Vec2f)GetRectangle()->size).y);
	fprintf(outfile, "%s}\n", IndentPrefix(depth));
} /* RectangleProxy::Print */


// ======================================================================

//
// RectangleProxy::PreRender
//
//

void RectangleProxy::PreRender(Effects const& effects)
{
	if (!InitPreRender(effects))
		return;

	// If the transparency for this node is 1, or it is not filled and
	// the line width is 0, then it doesn't need to be drawn, and its
	// bounding rectangle is empty.
	if ((GetTransparency() > 1 - ALMOST_ZERO) ||
		((GetFilled() == FALSE) && (GetWidth() == 0)))
	{
		m_boundingRect.MakeEmpty();
		UpdateBoundingRect();
		return;
	}

	// Determine where the four corners of the rectangle are
	m_corners[0].x = -((Vec2f)GetRectangle()->size).x / 2;
	m_corners[0].y = ((Vec2f)GetRectangle()->size).y / 2;
	m_corners[1].x = ((Vec2f)GetRectangle()->size).x / 2;
	m_corners[1].y = ((Vec2f)GetRectangle()->size).y / 2;
	m_corners[2].x = ((Vec2f)GetRectangle()->size).x / 2;
	m_corners[2].y = -((Vec2f)GetRectangle()->size).y / 2;
	m_corners[3].x = -((Vec2f)GetRectangle()->size).x / 2;
	m_corners[3].y = -((Vec2f)GetRectangle()->size).y / 2;

	// Determine where the corresponding shadow corners are
	for (int i = 0; i < 4; i++)
	{
		m_shadows[i].x = m_corners[i].x + GetShadPosX();
		m_shadows[i].y = m_corners[i].y + GetShadPosY();
	}

	// Determine where these points lie after the transformation has been
	// applied.
	for (i = 0; i < 4; i++)
	{
		m_transform.Apply(m_corners[i].x, m_corners[i].y);
		m_transform.Apply(m_shadows[i].x, m_shadows[i].y);
	}

	// A rotated rectangle will be treated much like a polygon
	if (m_transform.ContainsRotation())
	{
		// Find the bounding box for this rotated rectangle
		m_boundingRect.SetDimensions(m_corners, 4);

		// If a shadow is to be drawn, determine where the shadow points
		// are transformed to, and extend the bounding rectangle
		if (GetShadow())
		{
			BRect shadowRect;
			shadowRect.SetDimensions(m_shadows, 4);
			m_boundingRect.Union(shadowRect);
		}
	}

	// A non-rotated rectangle is handled separately
	else
	{
		// The bounding rectangle of the rectangle itself
		m_drawRect.SetDimensions(m_corners[0].x, m_corners[0].y, 
								 m_corners[2].x - m_corners[0].x + 1,
								 m_corners[0].y - m_corners[2].y + 1);
		m_boundingRect = m_drawRect;

		// Extend this bounding rectangle to include the shadow, if required
		if (GetShadow())
		{
			m_shadowRect.SetDimensions(m_shadows[0].x, m_shadows[0].y, 
									   m_shadows[2].x - m_shadows[0].x + 1,
									   m_shadows[0].y - m_shadows[2].y + 1);
			m_boundingRect.Union(m_shadowRect);
		}
	}

	UpdateBoundingRect();
} /* RectangleProxy::PreRender */


// ============================================================================

//
// RectangleProxy::Draw
//
// Draw the rectangle.  If the rectangle is rotated, it is drawn as a
// polygon, otherwise it is drawn as a rectangle.  Transparency is
// only supported for non-rotated filled rectangles, which do not
// have a shadow.
//

void RectangleProxy::Draw(BRect const* pClipBRect)
{
	assert(pClipBRect != NULL);
	VisualRenderer* pVR = GetVisualRenderer();

	// The amount of the rectangle to draw lies within both the clipping
	// rectangle and the rectangle's own bounding rectangle.
	BRect visibleRect = *pClipBRect;
	visibleRect.Intersection(GetBoundingRect());
	if (visibleRect.IsEmpty())
		return;

	// Get the transparency value to use
	float transparency = GetTransparency();

	// If a shadow is required, draw it now
	if (GetShadow())
	{
		// Determine where to draw the shadow, allowing for the transformation
		int shadowPosX = GetShadPosX();
		int shadowPosY = GetShadPosY();
		m_transform.Apply(shadowPosX, shadowPosY);

		// Only draw the shadow if it is not directly underneath the rectangle...
		if ((shadowPosX != 0) || (shadowPosY != 0))
		{
			Color3f shadowColor(0, 0.5, 0);

			if (m_transform.ContainsRotation())
				pVR->DisplayPolygon(visibleRect, m_shadows, 4,
									shadowColor, GetWidth(), GetFilled());

			else
				pVR->DisplayRectangle(visibleRect, m_shadowRect,
									  shadowColor, GetWidth(),
									  GetFilled(), 0.0);

			// As a shadow has been drawn, transparency is not supported
			transparency = 0;
		}
	}

	// And now draw the rectangle itself
	if (m_transform.ContainsRotation())
		pVR->DisplayPolygon(visibleRect, m_corners, 4, GetDiffuseColor(),
							GetWidth(), GetFilled());
	else
		pVR->DisplayRectangle(visibleRect, m_drawRect, GetDiffuseColor(),
							  GetWidth(), GetFilled(), transparency);
} /* RectangleProxy::Draw */


// ===========================================================================

//
// RectangleProxy::IsTransparent
//
// The Rectangle is not transparent if the following conditions hold:
//  - DrawableNode::IsTransparent() is FALSE,
//  - it is filled,
//  - there is no transparency,
//  - the rectangle is not rotated
//

BOOL RectangleProxy::IsTransparent() const
{
	if ((!DrawableNode::IsTransparent()) &&
		(GetFilled()) && (GetTransparency() == 0) &&
		(!m_transform.ContainsRotation()))
		return FALSE;
	else
		return TRUE;
} /* RectangleProxy::IsTransparent */


// ===========================================================================
