///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File: AudioSourceProxy.cpp
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of a simple AudioSourceProxy node used for
//   testing purposes.
//
// Created: August, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#ifdef WIN32
#include <io.h>
#else
#include <sys/types.h>
#include <sys/stat.h>
#endif
#include "../AudRend-CSELT.h"
#include "AudioSourceProxy.h"


// ============================================================================

//
// AudioSourceProxy::AudioSourceProxy
//
// Class constructor.
//

AudioSourceProxy::AudioSourceProxy()
{
	m_bufferSize = 0;
	m_buffer = NULL;
	m_pAudRend = NULL;
} /* AudioSourceProxy::AudioSourceProxy */


// ============================================================================

//
// AudioSourceProxy::~AudioSourceProxy
//
// Class destructor.
//

AudioSourceProxy::~AudioSourceProxy()
{
	if (m_pAudRend != NULL)
		m_pAudRend->RemoveSourceNode(this);

	if (m_buffer != NULL)
		delete m_buffer;
} /* AudioSourceProxy::~AudioSourceProxy */


// ============================================================================

//
// AudioSourceProxy::Init
//
// Open the given file, with the associated audio format.  Return
// TRUE if opened OK, otherwise return FALSE.
//

int AudioSourceProxy::Init(char* fileName, DWORD samplesPerSec,
						   DWORD bitsPerSample, WORD numChannels,
						   AudioRenderer* pAudRend)
{
	// Open the given file
	m_file = fopen(fileName, "rb");
	if (m_file == NULL)
		return 0;

	// Determine the length of the file
#ifdef WIN32	
	m_bytesLeftInFile = _filelength(_fileno(m_file));
#else
	struct stat buf;
	fstat(fileno(m_file), &buf);
	m_bytesLeftInFile = buf.st_size;
#endif

	// Tell the audio renderer about this node
	m_pAudRend = pAudRend;
	m_pAudRend->AddSourceNode(this, 0, samplesPerSec, bitsPerSample,
							  numChannels);

	// Make up a buffer to read data from the file (30ms)
	m_bufferSize = samplesPerSec * bitsPerSample / 8 * numChannels * 30 / 1000;
	m_buffer = new char [m_bufferSize];

	return 1;
} /* AudioSourceProxy::Init */


// ============================================================================

//
// AudioSourceProxy::GetData
//
// This function is used by the audio renderer to get more data
// to play.  The audio render will ask for a given number of bytes
// (which will be a multiple of the number of bytes required for a
// single sample), and the audio source node will return a pointer
// to some data, and the number of bytes of data pointed to.
//
// The number of bytes returned can be less than that asked for.
// If fewer bytes were returned to the audio renderer than expected,
// then the audio renderer will ask for more.  If zero bytes are
// returned, then the audio renderer will assume that no data can be
// given, and will insert silence.
//

void AudioSourceProxy::GetData(DWORD nBytesRequired, LPBYTE& rpData,
							   DWORD& rnBytesGiven)
{
	// If the end of the file has already been reached, return 0 bytes
	if (m_bytesLeftInFile <= 0)
	{
		rnBytesGiven = 0;
		rpData = NULL;
	}
	else
	{
		// Otherwise determine the number of bytes to read, being
		// the smallest of the buffer size, the number requested and
		// the number of bytes left.
		rnBytesGiven = m_bufferSize;
		if (rnBytesGiven > nBytesRequired)
			rnBytesGiven = nBytesRequired;
		if (rnBytesGiven > (DWORD)m_bytesLeftInFile)
			rnBytesGiven = (DWORD)m_bytesLeftInFile;

		fread(m_buffer, 1, rnBytesGiven, m_file);
		rpData = (LPBYTE)m_buffer;

		m_bytesLeftInFile -= rnBytesGiven;
	}
} /* AudioSourceProxy::GetData */


// ============================================================================

//
// AudioSourceProxy::DataUsed
//
// This function is used by the audio renderer to indicate that it
// has finished with some data given to it by GetData().  The
// parameters given to this function will be the same as those
// returned by GetData().
//

void AudioSourceProxy::DataUsed(LPBYTE /*pData*/, DWORD /*nBytesGiven*/)
{
	
} /* AudioSourceProxy::DataUsed */


// ============================================================================
