///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  Text2DProxy.cpp
//
// Authors: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of the proxy object used by the presenter
//   to render the Text2D node.
//
// Created: September, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-CSELT.h"
#include "Text2DProxy.h"
#include "FontStyle2DProxy.h"


// =====================================================================

//
// PresenterCSELT::InitText2D
//
// Initialise the proxy data structure and attach it to the render data
// pointer of the given media object node.
//

void PresenterCSELT::InitText2D(Text2D* pObj)
{
	pObj->SetPresenterData(new Text2DProxy(pObj));
}


// =====================================================================

//
// Text2DProxy::Print
//
// Print out the fields and corresponding values for the media object
// derived node that this object is a proxy for, indented according to
// the given depth.
//

void Text2DProxy::Print(FILE* outfile, int depth) const
{
	fprintf(outfile, "Text2D {\n");
	const char* indentStr = IndentPrefix(depth + 1);

	fprintf(outfile, "%sdata \"%s\"\n", indentStr,
			(LPCSTR)(ZString)GetText2D()->data);

	fprintf(outfile, "%slength %.2f\n", indentStr,
			(float)GetText2D()->length);

	fprintf(outfile, "%sfontStyle ", indentStr);
	if (GetFontStyle2D() == NULL)
		fprintf(outfile, "[]\n");
	else
		((FontStyle2DProxy*)(GetFontStyle2D()->GetPresenterData()))->Print(outfile, depth + 1);

	fprintf(outfile, "%sucs_2 %s\n", indentStr,
			BoolToStr(GetText2D()->ucs_2));

	fprintf(outfile, "%s}\n", IndentPrefix(depth));
} /* Text2DProxy::Print */


// =====================================================================

//
// Text2DProxy::PreRender
//
// Call the DrawableNode class's method to remember the effects and to
// register with the visual renderer.  Then, if either the Text2D node
// or its associated FontStyle2D node has changed since this was last
// drawn, determine the new layout of the text.
//

void Text2DProxy::PreRender(Effects const& effects)
{
	if (!InitPreRender(effects))
		return;

	// If the transparency for this node is 1, then it doesn't need
	// to be drawn, and its bounding rectangle is empty
	if (GetTransparency() > 1 - ALMOST_ZERO)
	{
		m_boundingRect.MakeEmpty();
		UpdateBoundingRect();
		return;
	}

	// Only proceed if the text string is not empty
	char const* pData = (LPCSTR)((const ZString&)GetText2D()->data);
	if ((pData == NULL) || (pData[0] == 0))
	{
		m_boundingRect.MakeEmpty();
		UpdateBoundingRect();
		return;
	}

	// Only layout the text if either the Text2D or FontStyle2D node
	// has changed.
/* (always do this until ROUTEs have been fixed to update the time)
	DWORD presTime = GetPresenter()->GetTime();
	if ((GetText2D()->GetUpdateTime() >= presTime) ||
		((GetFontStyle2D() != NULL) &&
		 (GetFontStyle2D()->GetUpdateTime() >= presTime)))
 */
	{
		// Make a local copy of the text string for the base class to use
		m_data = pData;

		// and then determine where the text string is to be displayed
		DrawableText::PreRender(effects);
	}

	UpdateBoundingRect();
} /* Text2DProxy::PreRender */


// =====================================================================

//
// Text2DProxy::Draw
//
// Draw the text string associated with this node
//

void Text2DProxy::Draw(BRect const* pClipBRect)
{
	DrawableText::Draw(pClipBRect);
} /* Text2DProxy::Draw */


// =====================================================================
