///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  Transform2DProxy.cpp
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of the proxy object used by the presenter
//   to render the Transform2D node.
//
// Created: September, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-CSELT.h"
#include "Transform2DProxy.h"
#include "trace.h"


// =====================================================================

//
// PresenterCSELT::InitTransform2D
//
// Initialise the proxy data structure and attach it to the render data
// pointer of the given media object node.
//

void PresenterCSELT::InitTransform2D(Transform2D* pObj)
{
	pObj->SetPresenterData(new Transform2DProxy(pObj));
}


// =====================================================================

//
// Transform2DProxy::PreRender
//
// Add the transform parameters of this node to the given effects, and
// then apply the PreRender method to all of the children of this node.
//

void Transform2DProxy::PreRender(Effects const& effects)
{
	Effects newEffects;

	//
	// Determine the matrix for this transform
	//
	TransformMatrix2D C, Cinv;		// Center of rotation
	C.SetTranslation(((Vec2f)GetTransform2D()->center).x,
					 ((Vec2f)GetTransform2D()->center).y);
	Cinv.SetTranslation(-((Vec2f)GetTransform2D()->center).x,
						-((Vec2f)GetTransform2D()->center).y);
//	TRACE_PRINT("\nC:\n");
//	C.Print(traceFile);
//	TRACE_PRINT("\nCinv:\n");
//	Cinv.Print(traceFile);

	TransformMatrix2D R;			// Rotation
	R.SetRotation((float)GetTransform2D()->rotationAngle);
//	TRACE_PRINT("\nR:\n");
//	R.Print(traceFile);

	TransformMatrix2D S;			// Scale
	S.SetScale(((Vec2f)GetTransform2D()->scale).x,
			   ((Vec2f)GetTransform2D()->scale).y);
//	TRACE_PRINT("\nS:\n");
//	S.Print(traceFile);

	TransformMatrix2D SR, SRinv;	// Scale orientation
	SR.SetRotation((float)GetTransform2D()->scaleOrientation);
	SRinv.SetRotation(-(float)GetTransform2D()->scaleOrientation);
//	TRACE_PRINT("\nSR:\n");
//	SR.Print(traceFile);
//	TRACE_PRINT("\nSRinv:\n");
//	SRinv.Print(traceFile);

	TransformMatrix2D T;			// Translation
	T.SetTranslation(((Vec2f)GetTransform2D()->translation).x,
					 ((Vec2f)GetTransform2D()->translation).y);
//	TRACE_PRINT("\nT:\n");
//	T.Print(traceFile);

	// Determine the new transform matrix and drawing order
	newEffects.transform = effects.transform * T * C * R * SR * S * SRinv * Cinv;
	newEffects.drawOrder = (float)GetTransform2D()->drawingOrder;

	// Copy over the isSwitchOn field unchanged
	newEffects.isSwitchedOn = effects.isSwitchedOn;

	// Apply these new effects to the children nodes
	PreRenderChildren(newEffects);
} /* Transform2DProxy::PreRender */


// =====================================================================

//
// Transform2DProxy::Print
//
// Print out the fields and corresponding values for the media object
// derived node that this object is a proxy for, indented according to
// the given depth.
//

void Transform2DProxy::Print(FILE* outfile, int depth) const
{
	fprintf(outfile, "%sTransform2D {\n", IndentPrefix(depth));

	// Print out the fields of this node
	const char* indentStr = IndentPrefix(depth + 1);
	fprintf(outfile, "%scenter %.2f %.2f\n", indentStr,
			((Vec2f)GetTransform2D()->center).x,
			((Vec2f)GetTransform2D()->center).y);
	fprintf(outfile, "%srotationAngle %.4f\n", indentStr,
			(float)GetTransform2D()->rotationAngle);
	fprintf(outfile, "%sscale %.2f %.2f\n", indentStr,
			((Vec2f)GetTransform2D()->scale).x,
			((Vec2f)GetTransform2D()->scale).y);
	fprintf(outfile, "%sscaleOrientation %.2f\n", indentStr,
			(float)GetTransform2D()->scaleOrientation);
	fprintf(outfile, "%sdrawingOrder %.2f\n", indentStr,
			(float)GetTransform2D()->drawingOrder);
	fprintf(outfile, "%stranslation %.2f %.2f\n", indentStr,
			((Vec2f)GetTransform2D()->translation).x,
			((Vec2f)GetTransform2D()->translation).y);

	// Print out the children of this node
	fprintf(outfile, "%schildren", indentStr);
	PrintChildren(outfile, depth + 1);

	fprintf(outfile, "%s}\n", IndentPrefix(depth));
} /* Transform2DProxy::Print */


// =====================================================================
