///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  Traversable.h
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The definition of the base object used by the visual renderer,
//   which encapsulates the features of a node which is traversable.
//
// Created: August, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#ifndef TRAVERSABLE_H
#define TRAVERSABLE_H


#include "MediaObj.h"
#include "Executiv.h"
#include "Presenter-CSELT.h"
#include "TransformMatrix2D.h"


class VisualRenderer;
class AudioRenderer;


//
// This class is used to store all of the attributes which effect how
// a node in the scene graph can be displayed.
//
class Effects
{
public:
	TransformMatrix2D transform;
	float drawOrder;
	BOOL isSwitchedOn;
};


class Traversable
{
	//
	// Public functions
	//
public:

	Traversable(MediaObject* pMO);
	// Class constructor.  Set the pointer to the media object that
	// this object is attached to, and get pointers to the presenter
	// and the visual and audio renderers, which can be used by
	// derived classes.

	virtual ~Traversable() {}
	// We need a virtual destructor

	virtual void UpdateTimeSensor() {}
	// This function will recursively traverse the scene tree,
	// updating all TimeSensor nodes.  This traversal of the scene
	// tree will be done before the PreRender traversal, as the time
	// sensor nodes will generate changes to fields which need to be
	// reflected in the drawing of the scene.
	//
	// The default implementation is to do nothing.  Grouping nodes
	// and the TimeSensor node will override this.

	virtual void PreRender(Effects const& effects) = 0;
	// This function will recursively traverse the scene tree,
	// determining where every object will appear (cumulation of
	// all transformation matricies).
	//
	// During this traversal, a node which is to be drawn has to call
	// VisualRenderer::AddToDrawList().  Even if the node has not
	// changed since PreRender() was last called, the node must
	// still be registered with the VisualRenderer.

	virtual BOOL IsFinished() const = 0;
	// Return TRUE if this node has finished, or FALSE otherwise.  Once
	// all nodes in the scene are finished, the Presenter stops rendering
	// the scene.

	virtual void FindTouchSensor(EventInfo const* pEvent);
	// Look for a touch sensor node.  Calling this function will do
	// one of two things.  If this is a leave node, it will call
	// FindTouchSensor() of its enclosing grouping node.  If this is
	// a grouping node, then OnUserEvent() will be called for all
	// nodes, stopping when the first TRUE return value is found.
	// If not, then FindTouchSensor() of the enclosing grouping node
	// is called, until the top of the scene tree is reached.

	virtual BOOL OnUserEvent(EventInfo const*) { return FALSE; }
	// Return TRUE if this is a TouchSensor node (which will handle
	// the given event), otherwise return FALSE.  The default
	// implementation returns FALSE.

	virtual void Print(FILE* outfile, int depth) const = 0;
	// Print out the fields and corresponding values for the media
	// object derived node that this object is a proxy for, indented
	// according to the given depth.  This method is intended to be
	// primarily used for debugging.

	MediaObject* GetObject() const { return m_pMediaObject; }
	// Return a pointer to the MediaObject derived object that this
	// object is attached to.


	//
	// Protected implementation details
	//
protected:

	const char* IndentPrefix(int depth) const;
	// Return a string containing the correct number of spaces for the given
	// level of indentation.  This function is called by the Print() method
	// of each node when printing out the scene tree.

	const char* BoolToStr(Bool b) const;
	// Return a string which displays the value of the given Bool.

	PresenterCSELT* GetPresenter() const { return m_pPresenter; }
	// Return a pointer to the presenter

	VisualRenderer* GetVisualRenderer() const { return m_pVisualRenderer; }
	// Return a pointer to the visual renderer

	AudioRenderer* GetAudioRenderer() const { return m_pAudioRenderer; }
	// Return a pointer to the audio renderer


	//
	// Private implementation details
	//
private:

	// A pointer to the MediaObject derived object that this
	// object is attached to.
	MediaObject* m_pMediaObject;

	// Remember the presenter that this object can use.
	PresenterCSELT* m_pPresenter;

	// Remember the visual renderer that this object can use.
	VisualRenderer* m_pVisualRenderer;

	// Remember the audio renderer that this object can use.
	AudioRenderer* m_pAudioRenderer;
};


#endif // TRAVERSABLE_H
