///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File: VisRend-CSELT-DD.cpp
//
// Author: Matthew Leditschke <CSELT, Matthew.Leditschke@cselt.it>
//
// Description:
//   The implementation of a class which performs visual rendering using
//   DirectDraw under Windows 95/NT.
//
// Created: August, 1997
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "VisRend-CSELT-DD.h"
#include "Copy.h"
#include "Caps.h"
#include "Executiv.h"
#include "Presenter-CSELT.h"
#include "Clock.h"
#include "dx-errors.h"
#include "MPEG4Player.h"
#include "FatalError.h"
#include "trace.h"

// =====================================================================

//
// VisualRendererDD::VisualRendererDD
//
// Class constructor - given a pointer to the presenter and a handle
// to the window to display in.
//

VisualRendererDD::VisualRendererDD(Presenter* pPresenter, HWND hWnd)
{
	TRACE_PRINT("Entering VisualRendererDD::VisualRendererDD()\n");

	m_initOK = TRUE;

	// Remember the handle to the window in which the drawing takes place
	m_hWnd = hWnd;

	// Remember a pointer to the presenter that this renderer is part of
	m_pPresenter = pPresenter;

	// Read from the dialog box the size of the window
//	m_outputWidth  = GetDialogBox()->GetOutputWidth();
//	m_outputHeight = GetDialogBox()->GetOutputHeight();
	m_outputWidth  = 352;
	m_outputHeight = 288;

	// Initialise the DirectDraw routines
	m_pDirectDraw = new DirectDraw;
	char* errorString = m_pDirectDraw->Init(hWnd, m_outputWidth, m_outputHeight);

	// If an error has occurred, display a message and then exit the
	// application
	if (errorString != NULL)
	{
		CFatalError fatalErrorDialog;
		fatalErrorDialog.m_message = errorString;
		fatalErrorDialog.DoModal();

		m_initOK = FALSE;
	}

	// Erase the output window
	if (m_initOK)
		Clear();

	TRACE_PRINT("Leaving VisualRendererDD::VisualRendererDD()\n");
} /* VisualRendererDD::VisualRendererDD */


// =====================================================================

//
// VisualRendererDD::~VisualRendererDD
//
// Class destructor - shuts down the DirectDraw routines.
//

VisualRendererDD::~VisualRendererDD()
{
	TRACE_PRINT("Entering VisualRendererDD::~VisualRendererDD()\n");

	m_pDirectDraw->Shutdown();
	delete m_pDirectDraw;

	TRACE_PRINT("Leaving VisualRendererDD::~VisualRendererDD()\n");
} /* VisualRendererDD::~VisualRendererDD */


// =====================================================================

//
// VisualRendererDD::Init
//
// Initialise the renderer.  Call the base class implementation.
//

void VisualRendererDD::Init()
{
	VisualRenderer::Init();
} /* VisualRendererDD::Init */


// ========================================================================

//
// VisualRendererDD::Terminate
//
// Stops rendering and performs cleanups.
// Erase the contents of the window if required.
//

void VisualRendererDD::Terminate()
{
	TRACE_PRINT("Entering VisualRendererDD::Terminate():\n");

	VisualRenderer::Terminate();

	if (m_clearOnEnd)
	{
		TRACE_PRINT("Erasing the contents of the window.\n");
		Clear();
		Paint();
	}

	TRACE_PRINT("Leaving VisualRendererDD::Terminate():\n");
} /* VisualRendererDD::Terminate */


// ========================================================================

//
// VisualRendererDD::Paint
//
// Copy the given region of the offscreen bitmap to the corresponding
// region of the onscreen window.  The region is given in BIFS coordinates.
// If the given rectangle pointer is NULL, the entire window is redrawn.
//

void VisualRendererDD::Paint(void* pInfo)
{
	BRect* pBRect = (BRect*)pInfo;

	// Determine the source rectangle
	RECT srcRect;
	if (pBRect == NULL)
		SetRect(&srcRect, 0, 0, m_outputWidth, m_outputHeight);
	else
	{
		// Convert from BIFS coords to window coords
		srcRect.left   = pBRect->left + m_outputWidth / 2;
		srcRect.top    = m_outputHeight / 2 - pBRect->top;
		srcRect.right  = srcRect.left + pBRect->width;
		srcRect.bottom = srcRect.top  + pBRect->height;
	}

	// Get the screen coordinates of the window to draw in
	RECT destRect;
	GetWindowRect(m_hWnd, &destRect);

// Reduce by 1 in every direction to allow for the border
//	destRect.left += 1;
	destRect.top  += 1;
//	destRect.right  -= 1;
	destRect.bottom -= 1;

	// If the rectangle is given, modify the destination rectangle
	// to include only this area.
	if (pBRect != NULL)
	{
		// Convert from BIFS coordinates to screen coordinates
		destRect.left  += pBRect->left + m_outputWidth / 2;
		destRect.top   += m_outputHeight / 2 - pBRect->top;
		destRect.right  = destRect.left + pBRect->width;
		destRect.bottom = destRect.top  + pBRect->height;
	}

	// Copy the given rectangle from the offscreen surface to the onscreen
	m_pDirectDraw->RenderSurface(&srcRect, &destRect);

//	TRACE_PRINT("In VisualRenderer::Paint():\n");
} /* VisualRendererDD::Paint */


// ========================================================================

//
// VisualRendererDD::GetCapabilities
//
// Get the capabilities of this visual renderer.
//

void VisualRendererDD::GetCapabilities (Capabilities* pCaps)
{
	pCaps->AddFormatCapability(FORMAT_RGB, 0);
	pCaps->AddFormatCapability(FORMAT_YUV, 1);
	pCaps->AddFormatCapability(FORMAT_YUVA, 2);
} /* VisualRendererDD::GetCapabilities */


// ========================================================================

//
// VisualRendererDD::Clear
//
// Completely erase the specifed region of the offscreen drawing
// area, but do not display the surface.  If the given rectangle
// is NULL, erase the entire offscreen area.
//

void VisualRendererDD::Clear(BRect const* pBRect)
{
	RECT updateRect;
	if (pBRect == NULL)
		m_pDirectDraw->EraseSurface(NULL);
	else
	{
		// Convert from BIFS coords to window coords
		updateRect.left   = pBRect->left + m_outputWidth / 2;
		updateRect.top    = m_outputHeight / 2 - pBRect->top;
		updateRect.right  = updateRect.left + pBRect->width;
		updateRect.bottom = updateRect.top  + pBRect->height;

		m_pDirectDraw->EraseSurface(&updateRect);
	}
} /* VisualRendererDD::Clear */


// ========================================================================

//
// VisualRendererDD::DisplayRGBAImage
//
// Display the given RGBA image, which includes where to display
// the top-left pixel in the image.  This uses the BIFS coordinate
// system, with the origin of the output display is in the middle
// of the window, with increasing X and Y to the right and up.
//

void VisualRendererDD::DisplayRGBAImage(SrcRGBAImage const& srcImage)
{
	// Change the (left,top) point from having the origin in the
	// middle of the screen to having the origin at the top-left
	// of the output window
	SrcRGBAImage flippedSrcImage = srcImage;
	flippedSrcImage.top = m_outputHeight / 2 - flippedSrcImage.top;
	flippedSrcImage.left += m_outputWidth / 2;

	// Lock the surface for writing
	DDSURFACEDESC ddSurfaceDesc;
	m_pDirectDraw->LockSurface(ddSurfaceDesc);
	SurfaceDesc surfaceDesc;
	surfaceDesc.dwWidth   = ddSurfaceDesc.dwWidth;
	surfaceDesc.dwHeight  = ddSurfaceDesc.dwHeight;
	surfaceDesc.lPitch    = ddSurfaceDesc.lPitch;
	surfaceDesc.lpSurface = ddSurfaceDesc.lpSurface;

	// Copy the image (or a portion of it)
	switch (m_pDirectDraw->GetRGBpixelFormat())
	{
	case RGB555:
		CopyRGBAToRGB555(flippedSrcImage, surfaceDesc);
		break;
	case RGB565:
		CopyRGBAToRGB565(flippedSrcImage, surfaceDesc);
		break;
	case RGB24bpp:
		CopyRGBAToRGB24bpp(flippedSrcImage, surfaceDesc);
		break;
	case RGB32bpp:
		CopyRGBAToRGB32bpp(flippedSrcImage, surfaceDesc);
		break;
	case BGR32bpp:
		CopyRGBAToBGR32bpp(flippedSrcImage, surfaceDesc);
		break;
	default:
		assert(0);		// Shouldn't get here.
		break;
	}

	// Finished writing to the surface, so unlock it
	m_pDirectDraw->UnlockSurface();
} /* VisualRendererDD::DisplayRGBAImage */


// ========================================================================

//
// VisualRendererDD::DisplayYUV420AImage
//
// Display the given YUV4:2:0 image, which includes where to display
// the top-left pixel in the image.  This uses the BIFS coordinate
// system, with the origin of the output display is in the middle
// of the window, with increasing X and Y to the right and up.
//

void VisualRendererDD::DisplayYUV420AImage(SrcYUVAImage const& srcImage)
{
	// The offsets must be even
	assert ((srcImage.left % 2) == 0);
	assert ((srcImage.top % 2) == 0);

	// The source width and height must be even - for now
	assert ((srcImage.width % 2) == 0);
	assert ((srcImage.height % 2) == 0);

	// Change the (left,top) point from having the origin in the
	// middle of the screen to having the origin at the top-left
	// of the output window
	SrcYUVAImage flippedSrcImage = srcImage;
	flippedSrcImage.top = m_outputHeight / 2 - flippedSrcImage.top;
	flippedSrcImage.left += m_outputWidth / 2;

	// Lock the surface for writing
	DDSURFACEDESC ddSurfaceDesc;
	m_pDirectDraw->LockSurface(ddSurfaceDesc);
	SurfaceDesc surfaceDesc;
	surfaceDesc.dwWidth   = ddSurfaceDesc.dwWidth;
	surfaceDesc.dwHeight  = ddSurfaceDesc.dwHeight;
	surfaceDesc.lPitch    = ddSurfaceDesc.lPitch;
	surfaceDesc.lpSurface = ddSurfaceDesc.lpSurface;

	// Copy the image (or a portion of it)
	switch (m_pDirectDraw->GetRGBpixelFormat())
	{
	case RGB555:
		CopyYUVToRGB555(flippedSrcImage, surfaceDesc);
		break;
	case RGB565:
		CopyYUVToRGB565(flippedSrcImage, surfaceDesc);
		break;
	case RGB24bpp:
		CopyYUVToRGB24bpp(flippedSrcImage, surfaceDesc);
		break;
	case RGB32bpp:
		CopyYUVToRGB32bpp(flippedSrcImage, surfaceDesc);
		break;
	case BGR32bpp:
		CopyYUVToBGR32bpp(flippedSrcImage, surfaceDesc);
		break;
	default:
		assert(0);		// Shouldn't get here.
		break;
	}

	// Finished writing to the surface, so unlock it
	m_pDirectDraw->UnlockSurface();
} /* VisualRendererDD::DisplayYUV420AImage */


// ========================================================================

//
// VisualRendererDD::DisplayCircle
//
// Display the given shape, which includes where to display
// the top-left pixel in the image.  This uses the BIFS coordinate
// system, with the origin of the output display is in the middle
// of the window, with increasing X and Y to the right and up.
//
void VisualRendererDD::DisplayCircle(BRect const& pClipBRect, 
									 PointF const* lpPoints,
									 Color3f pixelColor, int width, 
									 BOOL filled)
{
	HDC hdc;
	HRGN hrgn;
	m_pDirectDraw->GetHDC(hdc);

	int left = pClipBRect.left + m_outputWidth / 2;
	int top = m_outputHeight / 2 - pClipBRect.top;
	int right = left + pClipBRect.width;
	int bottom = top + pClipBRect.height;
	hrgn = CreateRectRgn(left, top, right, bottom); 
    SelectClipRgn(hdc, hrgn); 

	HGDIOBJ hOldBrush = 0;
	HGDIOBJ hOldPen = 0;
	CBrush brush;

	int color = RGB(255 * pixelColor.r, 255 * pixelColor.g, 255 * pixelColor.b);
	CPen pen(PS_SOLID, width, color);
	hOldPen = SelectObject(hdc, pen);

	int xPos1 = lpPoints[0].x + m_outputWidth / 2; 
	int yPos1 = m_outputHeight / 2 - lpPoints[0].y; 
	int xPos2 = lpPoints[1].x + m_outputWidth / 2; 
	int yPos2 = m_outputHeight / 2 - lpPoints[1].y; 

	if (filled)
	{
		brush.CreateSolidBrush(color);
		hOldBrush = SelectObject(hdc, brush);
	}
	else
		hOldBrush = SelectObject(hdc, GetStockObject(HOLLOW_BRUSH));

	Ellipse(hdc, xPos1, yPos1, xPos2, yPos2);

	if (hOldBrush != 0)
		SelectObject(hdc, hOldBrush);

	if (hOldPen != 0)
		SelectObject(hdc, hOldPen);

    SelectClipRgn(hdc, NULL); 
	m_pDirectDraw->ReleaseHDC(hdc);

	TRACE_PRINT("After DisplayCircle\n");
} /* VisualRendererDD::DisplayCircle */


// ========================================================================

//
// VisualRendererDD::DisplayCurve
//
// Display the given shape, which includes where to display
// the top-left pixel in the image.  This uses the BIFS coordinate
// system, with the origin of the output display is in the middle
// of the window, with increasing X and Y to the right and up.
//
void VisualRendererDD::DisplayCurve(BRect const& pClipBRect, 
									PointF const* lpPoints,
								    Color3f pixelColor, 
									int width, int length)
{
	HDC hdc;
	m_pDirectDraw->GetHDC(hdc);

	int left = pClipBRect.left + m_outputWidth / 2;
	int top = m_outputHeight / 2 - pClipBRect.top;
	int right = left + pClipBRect.width;
	int bottom = top + pClipBRect.height;
	HRGN hrgn = CreateRectRgn(left, top, right, bottom); 
    SelectClipRgn(hdc, hrgn); 

	POINT* points = new POINT[length];

	HGDIOBJ hOldPen = 0;
	int color = RGB(255 * pixelColor.r, 255 * pixelColor.g, 255 * pixelColor.b);
	CPen pen(PS_SOLID, width, color);
	hOldPen = SelectObject(hdc, pen);

	for (int i = 0; i < length; i++)
	{
		points[i].x = lpPoints[i].x + m_outputWidth / 2; 
		points[i].y = m_outputHeight / 2 - lpPoints[i].y; 
	}
	PolyBezier(hdc, points, length);

	delete [] points;

	if (hOldPen != 0)
		SelectObject(hdc, hOldPen);

    SelectClipRgn(hdc, NULL); 
	m_pDirectDraw->ReleaseHDC(hdc);

	TRACE_PRINT("After DisplayCurve\n");
}


// ========================================================================

//
// VisualRendererDD::DisplayPolygon
//
// Draw the portion of the given polygon which lies within the given
// clipping rectangle.  The polygon is given by an array of points
// (locations given in BFS coordinates).  The colour, line width
// and filled flag are also given.
//

void VisualRendererDD::DisplayPolygon(BRect const& clipBRect,
									  PointF const* lpPoints, int numPoints,
									  Color3f const& pixelColor,
									  int width, BOOL filled)
{
	HDC hdc;
	m_pDirectDraw->GetHDC(hdc);

	// Set up a clipping region, in screen coordinates
	int left = clipBRect.left + m_outputWidth / 2;
	int top = m_outputHeight / 2 - clipBRect.top;
	int right = left + clipBRect.width;
	int bottom = top + clipBRect.height;
	HRGN hrgn = CreateRectRgn(left, top, right, bottom); 
    SelectClipRgn(hdc, hrgn); 

	// Set up the pen to draw this polygon with
	int color = RGB(255 * pixelColor.r, 255 * pixelColor.g, 255 * pixelColor.b);
	CPen pen(PS_SOLID, width, color);
	HGDIOBJ hOldPen = SelectObject(hdc, pen);

	// Make a local copy of the array of points, converting from BIFS to
	// screen coordinates at the same time.
	POINT* points = new POINT [numPoints];
	for (int i = 0; i < numPoints; i++)
	{
		points[i].x = lpPoints[i].x + m_outputWidth /2;
		points[i].y = m_outputHeight / 2 - lpPoints[i].y;
	}

	// Set the brush to either filled or hollow, if the polygon is to
	// be filled or not.
	HGDIOBJ hOldBrush;
	CBrush brush;
	if (filled)
	{
		brush.CreateSolidBrush(color);
		hOldBrush = SelectObject(hdc, brush);
	}
	else
		hOldBrush = SelectObject(hdc, GetStockObject(HOLLOW_BRUSH));

	// Draw the polygon
	Polygon(hdc, points, numPoints);

	// Release stuff
	delete [] points;
	if (hOldBrush != 0)
		SelectObject(hdc, hOldBrush);

	if (hOldPen != 0)
		SelectObject(hdc, hOldPen);

    SelectClipRgn(hdc, NULL); 
	m_pDirectDraw->ReleaseHDC(hdc);
} /* VisualRendererDD::DisplayPolygon */


// ========================================================================

//
// VisualRendererDD::DisplayRectangle
//
// Draw a rectangle, whose size and location is given by drawBRect,
// only showing that portion which lies within the given clipping
// rectangle.  The colour, line width, filled flag and transparency
// are also given.  All coordinates are given in BIFS coordinates.
//
// If the rectangle is filled, then the rectangle is drawn by writing
// directly to the frame store.  If not, the windows routines are used.
//

void VisualRendererDD::DisplayRectangle(BRect const& clipBRect,
										BRect const& drawBRect,
										Color3f const& pixelColor,
										int width, BOOL filled,
										float transparency)
{
	// Determine how much of the rectangle is to be drawn
	BRect visibleRect = drawBRect;
	visibleRect.Intersection(clipBRect);
	visibleRect.top = m_outputHeight / 2 - visibleRect.top;
	visibleRect.left += m_outputWidth / 2;

	// Change the coordinates of the rectangle to draw from BIFS
	// to screen coordinates.
	BRect screenRect = drawBRect;
	screenRect.top = m_outputHeight / 2 - screenRect.top;
	screenRect.left += m_outputWidth / 2;

	// Change the clip rectangle from BIFS to screen coordinates.
	BRect screenClip = clipBRect;
	screenClip.top = m_outputHeight / 2 - screenClip.top;
	screenClip.left += m_outputWidth / 2;

	// If the rectangle is filled, write directly to the frame store
	if (filled)
	{
		// Lock the surface for writing
		DDSURFACEDESC ddSurfaceDesc;
		m_pDirectDraw->LockSurface(ddSurfaceDesc);
		SurfaceDesc surfaceDesc;
		surfaceDesc.dwWidth   = ddSurfaceDesc.dwWidth;
		surfaceDesc.dwHeight  = ddSurfaceDesc.dwHeight;
		surfaceDesc.lPitch    = ddSurfaceDesc.lPitch;
		surfaceDesc.lpSurface = ddSurfaceDesc.lpSurface;

		// Copy the image (or a portion of it)
		switch (m_pDirectDraw->GetRGBpixelFormat())
		{
		case RGB555:
			FillRGB555(visibleRect, surfaceDesc, pixelColor, transparency);
			break;
		case RGB565:
			FillRGB565(visibleRect, surfaceDesc, pixelColor, transparency);
			break;
		case RGB24bpp:
			FillRGB24bpp(visibleRect, surfaceDesc, pixelColor, transparency);
			break;
		case RGB32bpp:
			FillRGB32bpp(visibleRect, surfaceDesc, pixelColor, transparency);
			break;
		case BGR32bpp:
			FillBGR32bpp(visibleRect, surfaceDesc, pixelColor, transparency);
			break;
		default:
			assert(0);		// Shouldn't get here.
			break;
		}

		// Finished writing to the surface, so unlock it
		m_pDirectDraw->UnlockSurface();
	}

	// Only the outline of the rectangle is drawn, use windows routines
	else
	{
		HDC hdc;
		m_pDirectDraw->GetHDC(hdc);
		CDC dc;
		dc.Attach(hdc);

		// Set up a clipping region, in screen coordinates
		HRGN hrgn = CreateRectRgn(screenClip.left, screenClip.top,
								  screenClip.left + screenClip.width,
								  screenClip.top + screenClip.height); 
		SelectClipRgn(hdc, hrgn); 

		// Set up the pen to draw this polygon with
		int color = RGB(255 * pixelColor.r, 255 * pixelColor.g, 255 * pixelColor.b);
		CPen pen(PS_SOLID, width, color);
		HGDIOBJ hOldPen = SelectObject(hdc, pen);

		// Set up a hollow brush
		HGDIOBJ hOldBrush = SelectObject(hdc, GetStockObject(HOLLOW_BRUSH));

		// Draw the frame of the rectangle
		dc.Rectangle(screenRect.left, screenRect.top,
					 screenRect.left + screenRect.width,
					 screenRect.top + screenRect.height);

		// Release stuff
		if (hOldBrush != 0)
			SelectObject(hdc, hOldBrush);

		if (hOldPen != 0)
			SelectObject(hdc, hOldPen);

		dc.Detach();
		SelectClipRgn(hdc, NULL); 
		m_pDirectDraw->ReleaseHDC(hdc);
	}
} /* VisualRendererDD::DisplayRectangle */


// ========================================================================

void VisualRendererDD::DisplayLines(BRect const& clipBRect, 
									int const& startPos1, int const& startPos2,
									int const& endPos1, int const& endPos2,
								    Color3f const& pixelColor, 
								    int width)
{
	HDC hdc;
	m_pDirectDraw->GetHDC(hdc);

	// Set up a clipping region, in screen coordinates
	int left = clipBRect.left + m_outputWidth / 2;
	int top = m_outputHeight / 2 - clipBRect.top;
	int right = left + clipBRect.width;
	int bottom = top + clipBRect.height;
	HRGN hrgn = CreateRectRgn(left, top, right, bottom); 
    SelectClipRgn(hdc, hrgn); 

	// Set up the pen to draw this polygon with
	int color = RGB(255 * pixelColor.r, 255 * pixelColor.g, 255 * pixelColor.b);
	CPen pen(PS_SOLID, width, color);
	HGDIOBJ hOldPen = SelectObject(hdc, pen);

	// Make a local copy of the array of points, converting from BIFS to
	// screen coordinates at the same time.
	int xPos1, xPos2, yPos1, yPos2;
	xPos1 = startPos1 + m_outputWidth /2;
	yPos1 = m_outputHeight / 2 - startPos2;
	xPos2 = endPos1 + m_outputWidth /2;
	yPos2 = m_outputHeight / 2 - endPos2;

	// Move the current position to the start position
	MoveToEx(hdc, xPos1, yPos1, NULL);

	// Draw the line
	LineTo(hdc, xPos2, yPos2);

	// Release stuff
	if (hOldPen != 0)
		SelectObject(hdc, hOldPen);

    SelectClipRgn(hdc, NULL); 
	m_pDirectDraw->ReleaseHDC(hdc);
} /* VisualRendererDD::DisplayLines */


// ========================================================================

// A structure storing details about a font and its associated device
// context.
typedef struct
{
	HDC hDC;
	HFONT hFont;
	HGDIOBJ hPrevFont;
} FontInfo;


// =====================================================================

//
// VisualRendererDD::SetFont
//
// Return a renderer specific handle to the requested font.
//

void* VisualRendererDD::SetFont(const char* fontName, int fontSize,
								BOOL bold, BOOL italic, BOOL underline)
{
	FontInfo* pFontInfo = new FontInfo;
	m_pDirectDraw->GetHDC(pFontInfo->hDC);

	// Create a description of the required font,
	LOGFONT font;
	font.lfHeight = -MulDiv(fontSize, GetDeviceCaps(pFontInfo->hDC, LOGPIXELSY), 72);
	font.lfWidth = 0;
	font.lfEscapement = 0;
	font.lfOrientation = 0;
	font.lfWeight = (bold) ? FW_BOLD : FW_NORMAL;
	font.lfItalic = italic;
	font.lfUnderline = underline;
	font.lfStrikeOut = FALSE;
	font.lfCharSet = ANSI_CHARSET;
	font.lfOutPrecision = OUT_DEFAULT_PRECIS;
	font.lfClipPrecision = CLIP_DEFAULT_PRECIS;
	font.lfQuality = DEFAULT_QUALITY;
	font.lfPitchAndFamily = DEFAULT_PITCH | FF_DONTCARE;
	strcpy(font.lfFaceName, fontName);

	// and now create the font,
	pFontInfo->hFont = CreateFontIndirect(&font);
	assert(pFontInfo->hFont != NULL);

	// and make it the current font.
	pFontInfo->hPrevFont = SelectObject(pFontInfo->hDC, pFontInfo->hFont);

	return pFontInfo;
} /* VisualRendererDD::SetFont */


// =====================================================================

//
// VisualRendererDD::UnsetFont
//
// Stop using the font which was set using SetFont().  Every call
// to SetFont() must be accompanied by a call to UnsetFont().
//


void VisualRendererDD::UnsetFont(void* pFont)
{
	assert(pFont != NULL);
	FontInfo* pFontInfo = (FontInfo*)pFont;

	// Unselect the font (by putting back the previous one), delete it,
	// and then release the device context.
	SelectObject(pFontInfo->hDC, pFontInfo->hPrevFont);
	DeleteObject(pFontInfo->hFont);
	m_pDirectDraw->ReleaseHDC(pFontInfo->hDC);

	delete pFont;
} /* VisualRendererDD::UnsetFont */


// =====================================================================

//
// VisualRendererDD::GetTextSize
//
// Return the width and the height of the given text string, using
// a font return by SetFont().
//

void VisualRendererDD::GetTextSize(void* pFont, const char* string,
								   int length, unsigned int& width,
								   unsigned int& height)
{
	assert(pFont != NULL);
	FontInfo* pFontInfo = (FontInfo*)pFont;

	SIZE stringSize;
	GetTextExtentPoint32(pFontInfo->hDC, string, length, &stringSize);
	width  = stringSize.cx;
	height = stringSize.cy;
} /* VisualRendererDD::GetTextSize */


// =====================================================================

//
// VisualRendererDD::DisplayText
//
// Display a string with the given length, at the specifed location
// and using the given colour, clipping the output to the given
// rectangle.  The x,y point and the clip rectangle are given in
// BIFS coordinates.
//
// Only one line of text will be given at a time.
//

void VisualRendererDD::DisplayText(void* pFont, const char* string,
								   int length, int x, int y,
								   Color3f const& color,
								   BRect const* pClipBRect)
{
	assert(pFont != NULL);
	FontInfo* pFontInfo = (FontInfo*)pFont;

	// Set the colour of the text string
	SetTextColor(pFontInfo->hDC,
				 RGB(color.r * 255, color.g * 255, color.b * 255));

	// There will be no background colour
	SetBkMode(pFontInfo->hDC, TRANSPARENT);

	// And set the text alignment within the clipping rectangle
	SetTextAlign(pFontInfo->hDC, TA_TOP | TA_LEFT);

	// Set up the clipping rectangle (& convert BIFS to screen coords)
	RECT clipRect;
	clipRect.left   = x + m_outputWidth / 2;
	clipRect.top    = m_outputHeight / 2 - y;
	clipRect.right  = clipRect.left + pClipBRect->width;
	clipRect.bottom = clipRect.top + pClipBRect->height;

	// Draw the text string
	DrawText(pFontInfo->hDC, string, length, &clipRect,
			 DT_LEFT | DT_SINGLELINE | DT_TOP);
} /* VisualRendererDD::DisplayText */


// ========================================================================

//
// VisualRendererDD::DisplayPoint
//
// Set the specified pixel to the given colour, using the provided level
// of transparency.  The location of the pixel to set is given in BIFS
// coordinates, with the origin of the output display is in the middle
// of the window, with increasing X and Y to the right and up.
//
void VisualRendererDD::DisplayPoint(int xPos, int yPos, Color3f const* pPixelColor,
									float transparency)
{
	// If the transparency is high enough so that the point is see through, return
	if (transparency > (float)254 / (float)255)
		return;

	// Convert the pixel location from BIFS to screen coordinates
	yPos = m_outputHeight / 2 - yPos;
	xPos += m_outputWidth / 2;

	// Lock the surface for writing
	DDSURFACEDESC ddSurfaceDesc;
	m_pDirectDraw->LockSurface(ddSurfaceDesc);
	SurfaceDesc surfaceDesc;
	surfaceDesc.dwWidth   = ddSurfaceDesc.dwWidth;
	surfaceDesc.dwHeight  = ddSurfaceDesc.dwHeight;
	surfaceDesc.lPitch    = ddSurfaceDesc.lPitch;
	surfaceDesc.lpSurface = ddSurfaceDesc.lpSurface;

	// Determine the new pixel colour
	int red   = pPixelColor->r * 255 + 0.5;
	int green = pPixelColor->g * 255 + 0.5;
	int blue  = pPixelColor->b * 255 + 0.5;

	// Copy the image (or a portion of it)
	switch (m_pDirectDraw->GetRGBpixelFormat())
	{
	case RGB555:
		{
			// Get a pointer to the pixel to change
			unsigned short* pDest = (unsigned short*)((unsigned char*)((unsigned char*)surfaceDesc.lpSurface + yPos * surfaceDesc.lPitch + 2 * xPos));

			if (transparency > ALMOST_ZERO)
			{
				// Get the RGB values underneath the pixel
				int oldRed   = (*pDest & 0x7c00) >> 7;
				int oldGreen = (*pDest & 0x03e0) >> 2;
				int oldBlue  = (*pDest & 0x001f) << 3;

				// Apply the transparency
				ApplyTransparencyToRGB(red, green, blue, transparency, oldRed, oldGreen, oldBlue);
			}

			// Combine the values, 5 bits red, 5 bits green and 5 bits blue
			*pDest++ = MAKE_RGB555_PIXEL(red, green, blue);
		}
		break;

	case RGB565:
		{
			// Get a pointer to the pixel to change
			unsigned short* pDest = (unsigned short*)((unsigned char*)((unsigned char*)surfaceDesc.lpSurface + yPos * surfaceDesc.lPitch + 2 * xPos));

			if (transparency > ALMOST_ZERO)
			{
				// Get the RGB values underneath the pixel
				int oldRed   = (*pDest & 0xf800) >> 8;
				int oldGreen = (*pDest & 0x07e0) >> 3;
				int oldBlue  = (*pDest & 0x001f) << 3;

				// Apply the transparency
				ApplyTransparencyToRGB(red, green, blue, transparency, oldRed, oldGreen, oldBlue);
			}

			// Combine the values, 5 bits red, 6 bits green and 5 bits blue
			*pDest++ = MAKE_RGB565_PIXEL(red, green, blue);
		}
		break;

	case RGB24bpp:
		{
			// Get a pointer to the pixel to change
			unsigned char* pDest = (unsigned char*)surfaceDesc.lpSurface
									+ yPos * surfaceDesc.lPitch	+ 3 * xPos;

			if (transparency > ALMOST_ZERO)
			{
				// Get the RGB values underneath the pixel
				int oldRed   = pDest[2];
				int oldGreen = pDest[1];
				int oldBlue  = pDest[0];

				// Apply the transparency
				ApplyTransparencyToRGB(red, green, blue, transparency, oldRed, oldGreen, oldBlue);
			}

			// Store the 8 bit values for blue, green and red
			*pDest++ = (unsigned char) blue;
			*pDest++ = (unsigned char) green;
			*pDest++ = (unsigned char) red;
		}
		break;

	case RGB32bpp:
	case BGR32bpp:
		{
			// Get a pointer to the pixel to change
			unsigned long* pDest = (unsigned long*)((unsigned char*)((unsigned char*)surfaceDesc.lpSurface + yPos * surfaceDesc.lPitch
									+ 4 * xPos));

			if (transparency > ALMOST_ZERO)
			{
				// Get the RGB values underneath the pixel
				int oldRed   = (*pDest & 0xff0000) >> 16;
				int oldGreen = (*pDest & 0xff00) >> 8;
				int oldBlue  = (*pDest & 0xff);

				// Apply the transparency
				ApplyTransparencyToRGB(red, green, blue, transparency, oldRed, oldGreen, oldBlue);
			}

			// Combine the values, 8 bits red, 8 bits green and 8 bits blue
			if (m_pDirectDraw->GetRGBpixelFormat() == RGB32bpp)
				*pDest++ = MAKE_RGB32bpp_PIXEL(red, green, blue);
			else
				*pDest++ = MAKE_BGR32bpp_PIXEL(red, green, blue);
		}
		break;

	default:
		assert(0);		// Shouldn't get here.
		break;
	}

	// Finished writing to the surface, so unlock it
	m_pDirectDraw->UnlockSurface();
} /* VisualRendererDD::DisplayPoint */


// =====================================================================
