/*
 * This is the definition of an interface to the MPEG-4 decoder which can only
 * decode a single video object.  It is intended that this library will be
 * called from other programs, in particular, from the overall MPEG-4 system
 * terminal.
 *
 * Matthew Leditschke
 * March, 1997.
 * Modified by Franco Casalino and Antonio Arena (CSELT) for read directly from buffer
 *
 * September, 1997
 */


#ifdef __cplusplus
extern "C" {
#endif


/*
 * This is the handle to the decoder.  Type casting is used to hide the details
 * of the decoder from the calling program.
 */
typedef long* Mpeg4DecoderHandle;

/*
 * This is the structure which contains the details of of a frame that has been
 * read from the decoder.  This structure must be initialised before calling
 * Mpeg4DecoderDecode().
 */
struct Mpeg4DecoderFrame
{
	unsigned char* pYdata;
	unsigned char* pUdata;
	unsigned char* pVdata;
	unsigned char* pAdata;
	int alphaIsRectangular;
	int allocatedWidth;
	int allocatedHeight;
	int activeWidth;
	int activeHeight;
	int offsetX;
	int offsetY;
};
typedef struct Mpeg4DecoderFrame Mpeg4DecoderFrame;


/*
 * Functions:
 */

/*
 * Mpeg4DecoderInit:
 *
 * Initialise the MPEG-4 decoding routines, giving the name of the file which
 * contains the bitstream to read from.  This bitstream must contain only one
 * video object.
 *
 * This function returns a pointer to the details of the decoder object
 * initialised.  NULL is returned if an error occurred opening the file.
 */
Mpeg4DecoderHandle Mpeg4DecoderInit (unsigned char *, int);

/*
 * Mpeg4DecoderShutdown:
 *
 * Shutdown the MPEG-4 decoding routines, giving the MPEG-4 decoder object to
 * shutdown.
 */
void Mpeg4DecoderShutdown (Mpeg4DecoderHandle decoderHandle);

/*
 * Mpeg4DecoderRewind:
 *
 * Rewind the bitstream so that the next frame decoded is from the start of the
 * bitstream.  Return a pointer to the new decoder object.
 */
Mpeg4DecoderHandle Mpeg4DecoderRewind (Mpeg4DecoderHandle decoderHandle);

/*
 * Mpeg4DecoderInitFrame:
 *
 * Initialise the given frame structure, to contain an empty frame of no size.
 */
void Mpeg4DecoderInitFrame (Mpeg4DecoderFrame* pFrame);

/*
 * Mpeg4DecoderDeallocFrame:
 *
 * Deallocate the memory associate with the given frame structure.
 */
void Mpeg4DecoderDeallocFrame (Mpeg4DecoderFrame* pFrame);

/*
 * Mpeg4DecoderDecode:
 *
 * Decode the bitstream until a complete frame (VOP) has been decoded, and write
 * this frame into the frame store provided, which will be at least the size
 * of the decoded image, and will contain the offsets for this frame.  This frame
 * store must have been initialised using Mpeg4DecoderInitFrame().
 *
 * Return 1 if the end of the bitstream was reached before a complete frame
 * could be decoded, otherwise return 0.
 */
int Mpeg4DecoderDecode (Mpeg4DecoderHandle decoderHandle,
						Mpeg4DecoderFrame* pFrame);

/*
 * Mpeg4DecoderDecodeHeader:
 *
 * Decode the bitstream until the header of the next frame (VOP) has been decoded.
 * The frame structure will contain the dimensions of the frame, and its offset,
 * but will not contain any allocated memory.  This frame store must have been
 * initialised using Mpeg4DecoderInitFrame().
 *
 * Return 1 if the end of the bitstream was reached before a complete frame
 * could be decoded, otherwise return 0.
 *
 * This function is intended to be used in conjunction with
 * Mpeg4DecoderDecoderBody(), as an alternative to Mpeg4DecoderDecode().
 */
int Mpeg4DecoderDecodeHeader (Mpeg4DecoderHandle decoderHandle,
							  Mpeg4DecoderFrame* pFrame);

/*
 * Mpeg4DecoderDecodeBody:
 *
 * Decode the body of the next frame (VOP), the header of which has been read by
 * Mpeg4DecoderDecoderHeader().  The pointers to the frame store to write the
 * decoded data are given, and are assumed to be big enough.  The width and the
 * height of this frame store are also given.
 *
 * Return 1 if the end of the bitstream was reached before a complete frame
 * could be decoded, otherwise return 0.
 *
 * This function is intended to be used in conjunction with
 * Mpeg4DecoderDecoderHeader(), as an alternative to Mpeg4DecoderDecode().
 */
int Mpeg4DecoderDecodeBody (Mpeg4DecoderHandle decoderHandle,
							unsigned char* Yptr, unsigned char* Uptr,
							unsigned char* Vptr, unsigned char* Aptr,
							unsigned int destWidth, unsigned int destHeight);


void resetBitstream(Mpeg4DecoderHandle, unsigned char *, int);


#ifdef __cplusplus
}
#endif
