/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Decoder
FILE:       BifsDec.cpp
AUTHOR:     Zvi Lifshitz (VDOnet Corp.) and Kevin Barron (KPN Research)
Date:		From 17/6/96
OVERVIEW:   Implementation of BIFSDecoder.
-------------------------------------------------------------------------*/
#include "defs.h"
#include <assert.h>
#include "BifsDec.h"
#include "NodeFld.h"
#include "Root.h"
#include "ESConfig.h"
#include "Executiv.h"
#include "Stream.h"
#include "WrldInfo.h"
#include "Route.h"
#include "StrAssoc.h"

//#define SIMULATE_BIFS_DECODER

#ifdef _DEBUG
#define pBifs this		// used by TRACEF
#endif

/*-----------------------------------------------------------------
BIFSDecoder constructor
-----------------------------------------------------------------*/
BIFSDecoder :: BIFSDecoder (Executive *pExecutive, MediaStream *pBIFS) :
	m_pExecutive (pExecutive),
	m_pBIFS (pBIFS)
{
#ifdef _DEBUG
	m_nIndent = 0;
	m_bNewLine = TRUE;
	m_fTrace = fopen ("trace.txt", "w");
//	m_fTrace = stdout;
#endif
}

BIFSDecoder :: ~BIFSDecoder ()
{
	m_pBIFS -> Terminate ();
#ifdef _DEBUG
	fclose (m_fTrace);
#endif
}

/*-----------------------------------------------------------------
Create the scene from BIFS and returns a pointer to the scene.
The pointer is also store in this object, for later use by UpdateScene().
-----------------------------------------------------------------*/
MediaObject * BIFSDecoder :: BuildScene ()
{
#ifdef _DEBUG
	printf ("Start building scene\n");
#endif

	m_pScene = NULL;
	m_nLastBit = 8;
	m_nSize = 0;
	m_nLastByte = 0;
	m_pData = NULL;

#ifdef SIMULATE_BIFS_DECODER
	/*-----------------------------------------------------------------
	Until BIFSDecoder is implemented, this function is a simulation,
	which manually builds the test scene
	-----------------------------------------------------------------*/
	Transform2D *p, *pScene;
	Shape *pShape;
	MovieTexture *pMovie;
	Sound *pSound;
	NaturalAudio *pClip;
	StreamingText *pText;

	m_pScene = pScene = new Transform2D;
	pScene -> SetExecutive (m_pExecutive);
	pScene -> children += p = new Transform2D;
	p -> SetExecutive (m_pExecutive);
	p -> SetParent (pScene);
	p -> translation = SFVec3F (88, 72, 10);
	p -> children += pShape = new Shape;
	pShape -> SetParent (p);
	pShape -> SetExecutive (m_pExecutive);
	pShape -> appearance = pMovie = new MovieTexture;
	pMovie -> SetParent (pShape);
	pMovie -> SetExecutive (m_pExecutive);
	pMovie -> loop = TRUE;
	pMovie -> speed = 1;
	pMovie -> objectDescriptorID = 32;
	pMovie -> repeatS = FALSE;
	pMovie -> repeatT = FALSE;
	p -> children += pSound = new Sound;
	pSound -> SetParent (p);
	pSound -> SetExecutive (m_pExecutive);
	pSound -> source = pClip = new NaturalAudio;
	pClip -> SetParent (pSound);
	pClip -> SetExecutive (m_pExecutive);
	pClip -> objectDescriptorID = 64;
	pScene -> children += p = new Transform2D;
	p -> SetParent (pScene);
	p -> SetExecutive (m_pExecutive);
	p -> translation = SFVec3f (264, 216, 20);
	p -> children += pShape = new Shape;
	pShape -> SetParent (p);
	pShape -> SetExecutive (m_pExecutive);
	pShape -> appearance = pMovie = new MovieTexture;
	pMovie -> SetParent (pShape);
	pMovie -> SetExecutive (m_pExecutive);
	pMovie -> loop = TRUE;
	pMovie -> speed = 1;
	pMovie -> objectDescriptorID = 33;
	pMovie -> repeatS = FALSE;
	pMovie -> repeatT = FALSE;
	pScene -> children += p = new Transform2D;
	p -> SetParent (pScene);
	p -> SetExecutive (m_pExecutive);
	p -> translation = SFVec3f (250, 50, 30);
	p -> children += pShape = new Shape;
	pShape -> SetParent (p);
	pShape -> SetExecutive (m_pExecutive);
	pShape -> geometry = pText = new StreamingText;
	pText -> SetParent (pShape);
	pText -> SetExecutive (m_pExecutive);
	pText -> objectDescriptorID = 99;
	pText -> maxSize = 100;
	pText -> ucs_2 = FALSE;

#else
	/*-----------------------------------------------------------------
	The BIFS parser architecture is based on the principle that every object,
	node or node field, parses itself.  So to parse the scene, we
	instantiate an SFNode object, and call the object to parse itself.
	The parse function will throw an exception if end of stream is 
	unexpectedly encountered.
	-----------------------------------------------------------------*/
	SFRootNode node;
	SFNode<WorldInfo> worldInfo;
	
	try
	{
		if (ParseInt (1))	// if (hasWorldInfo)
			worldInfo.Parse (this, NULL);
		if (node.Parse (this, NULL))
		{	/*-----------------------------------------------------------------
			If parse was successful, "node" now points to the scene root,
			which is the value we have to return. After saving the root 
			pointer, we set "node" to NULL. This prevents the destruction
			of the tree by SFNode's desctructor.
			-----------------------------------------------------------------*/
			m_pScene = (Root *) (MediaObject *) node;
			m_pScene -> SetContainer (NULL);
			m_pScene -> SetWorldInfo (worldInfo);
			if ((MediaObject *) worldInfo != NULL)
				((MediaObject *) worldInfo) -> SetRoot (m_pScene);

			/*-----------------------------------------------------------------
			Parse Routes
			-----------------------------------------------------------------*/
			if (ParseInt (1))	// if hasROUTEs
				ParseRoutesDescription (m_pScene);

			/*-----------------------------------------------------------------
			Parse SessionStreamAssociation
			-----------------------------------------------------------------*/
			if (ParseInt (1))	// if hasSessionStreamAssociation
			{	SFNode<SessionStreamAssociation> streamAssociation;

				streamAssociation.Parse (this, m_pScene);
				m_pScene ->
					SetSessionStreamAssociation (streamAssociation);
				if ((SessionStreamAssociation *) streamAssociation)
					streamAssociation -> SetStreams ();
			}
		}
	}
	catch (BifsException)
	{
		m_pScene = NULL;
	}

#endif
	return m_pScene;
}

/*-----------------------------------------------------------------
Parses ObjectDescriptors and scene updates from the stream, and update
the scene accordingly.
-----------------------------------------------------------------*/
void BIFSDecoder :: UpdateScene ()
{
#ifdef SIMULATE_BIFS_DECODER
	/*-----------------------------------------------------------------
	Until BIFSDecoder is implemented, this function is a simulation,
	which manually attaches ObjectDescriptors to nodes,
	then fetches units from the BIFS stream.
	-----------------------------------------------------------------*/
	ES_DESCRIPTOR objectJPEG = {10, 0, "visual/JPEG", {10, 1}};
	ES_DESCRIPTOR objectG723 = {10, 0, "audio/G723", {10, 2}};
	ES_DESCRIPTOR objectH263 = {10, 0, "visual/H263", {10, 1}};
	ES_DESCRIPTOR objectTEXT = {10, 0, "visual/TEXT", {10, 1}};
	m_pExecutive -> LockScene ();
	m_pScene -> FindObjectDescriptor (33) -> SetStream (&objectJPEG);
	m_pScene -> FindObjectDescriptor (32) -> SetStream (&objectH263);
	m_pScene -> FindObjectDescriptor (99) -> SetStream (&objectTEXT);
	m_pScene -> FindObjectDescriptor (64) -> SetStream (&objectG723);
	m_pExecutive -> UnlockScene ();

	int nSize;
	DWORD dwTime;

	while (m_pBIFS -> FetchNext (nSize, dwTime, TRUE) != END_OF_STREAM)
	{
		m_pBIFS -> Release (TRUE);
	}

#else

	/*-----------------------------------------------------------------
	Loop till end of stream
	-----------------------------------------------------------------*/
	try
	{	for (;;)
		{
			ParseUpdateFrame ();
		}
	}
	catch (BifsException) {}

#endif
}

/*-----------------------------------------------------------------
Parse a session stream association element
-----------------------------------------------------------------*/
void BIFSDecoder :: ParseSessionStreamAssociation ()
{
	TRACEF ("SessionStreamAssociation {", 0);
	TRACENEWLINE;

	/*-----------------------------------------------------------------
 	Get the number of object_dec to be parsed
	Call the parse function for each instance
	-----------------------------------------------------------------*/
	int nObjectCount = ParseInt (8);
	int i;

	TRACEF ("ObjectCount %d", nObjectCount);
	TRACENEWLINE;
	TRACEINDENT (1);

	for	(i = 0; i < nObjectCount; i++)
	{
		ParseObjectDesc();
	}
	TRACEINDENT (-1);
}


/*-----------------------------------------------------------------
ParseObjectDec will parse each Obj Dec in the stream
----------------------------------------------------------------*/
void BIFSDecoder :: ParseObjectDesc ()
{
	TRACEF ("ObjectDescriptor {", 0);
	TRACENEWLINE;
	TRACEINDENT (1);

	ES_DESCRIPTOR objectDesc;
	
	// Variables needed only to manage this function
	int nObjectDescriptorId	= ParseInt (10);
	int nStreamCount = ParseInt (5);
	BOOL bExtensionFlag = ParseInt (1);

	TRACEF ("ObjectDescriptorId %d", nObjectDescriptorId);
	TRACENEWLINE;
	TRACEF ("StreamCount %d", nStreamCount);
	TRACENEWLINE;
	TRACEF ("extension_flag %d", bExtensionFlag);
	TRACENEWLINE;

	// loop over the number of streams
	for (int i = 0; i < nStreamCount; i++)
	{
		TRACEF ("ES_Descriptor {", 0);
		TRACENEWLINE;
		TRACEINDENT (1);
	
		objectDesc.m_nES_number = ParseInt (5);
		ParseInt (3);		// skip 3 bits (reserved not used)
		ParseInt (8);		// skip 8 bits (Muxinfolength not used)
		
		TRACEF ("ES_number %d", objectDesc.m_nES_number);
		TRACENEWLINE;

		/*-----------------------------------------------------------------
		Assign DectypeString to the structure
		Get the first 8 bits ... it holds the length of the string
		-----------------------------------------------------------------*/
		int nStrLength = ParseInt (8);
		int j = 0;
		while (--nStrLength >= 0)
			objectDesc.m_szDecType [i++] = ParseInt (8);
		objectDesc.m_szDecType [i] = 0;

		TRACEF ("DecTypeString %s", objectDesc.m_szDecType);
		TRACENEWLINE;

		ParseInt (8);		// skip 8 bits (Qos_Descriptor not used)
		
		TRACEF ("ESConfigParams {", 0);
		TRACENEWLINE;
		TRACEINDENT (1);

		objectDesc.m_ESConfigParams.m_nPredefined = ParseInt (8);

		TRACEF ("predefined %d", objectDesc.m_ESConfigParams.m_nPredefined);
		TRACENEWLINE;
		TRACEINDENT (-1);
		TRACEF ("}", 0);

		TRACENEWLINE;
		TRACEINDENT (-1);
		TRACEF ("}", 0);

		// assign the relevant streams
		m_pExecutive -> SetStream (nObjectDescriptorId, &objectDesc, NULL);
	}
	TRACENEWLINE;
	TRACEINDENT (-1);
	TRACEF ("}", 0);
	TRACENEWLINE;
}

/*-----------------------------------------------------------------
Retrieves nBits bits from the BifStream and returns their integer value
Methodology based on GetByte() and GetBits()
-----------------------------------------------------------------*/
int BIFSDecoder :: ParseInt (int nBits)
{
	BOOL bRecurrent;
	int nReturn = 0;
	while (nBits-- > 0)
	{
		/*-----------------------------------------------------------------
		Get next byte, if necessary
		-----------------------------------------------------------------*/
		if (m_nLastBit >= 8)
		{	
			if (m_nSize == 0)
			{
				DWORD dwTime;

				/*-----------------------------------------------------------------
				Fetch a new block of values
				-----------------------------------------------------------------*/
				if (m_pData != NULL)
					m_pBIFS -> Release (TRUE);
				BOOL bEndOfStream;
				m_pData = (LPBYTE) m_pBIFS -> Fetch (
							m_nSize,
							dwTime,
							bRecurrent,
							bEndOfStream,
							FALSE,
							TRUE);
				if (bEndOfStream)
				{
					m_pBIFS -> Terminate ();
					throw BifsException (BIFX_END_OF_STREAM);
				}
			}
	 		m_nLastByte = (int) *m_pData++;
			m_nLastBit = 0;
			m_nSize--;
		}

		/*-----------------------------------------------------------------
		Retrieve the first bit of the byte, shift the byte so next time
		second bit will be the first
		-----------------------------------------------------------------*/
		nReturn <<= 1;
		nReturn |= (m_nLastByte & 0x80) >> 7;
		m_nLastByte <<= 1;
		++m_nLastBit;
	}
	return nReturn;

}

/*-----------------------------------------------------------------
The following is the float parse function 
-----------------------------------------------------------------*/
float BIFSDecoder :: ParseFloat ()
{	
#ifdef WIN32
	char buf [4] = "\0\0\0";
	int i;
	for (i = 0; i < 32; i++)
		buf[3-i/8] |= ParseInt (1) << (7 - i%8);

	return * (float *) buf;

#endif

	return 0.0;
}

/*-----------------------------------------------------------------
The following is the double parse function
-----------------------------------------------------------------*/
double BIFSDecoder :: ParseDouble ()
{
#ifdef WIN32	
	char buf [8] = "\0\0\0\0\0\0\0";
	int i;
	for (i = 0; i < 64; i++)
		buf[7-i/8] |= ParseInt (1) << (7 - i%8);
	
	return * (double *) buf;

#endif

	return 0.0;
}

/*-----------------------------------------------------------------
Trace functions for debug mode
-----------------------------------------------------------------*/
#ifdef _DEBUG
void BIFSDecoder :: Tracef (LPCSTR szFormat, ...)
{
	if (m_bNewLine)
	{	for (int i = m_nIndent; --i >= 0;)
			putc ('\t', m_fTrace);
		m_bNewLine = FALSE;
	}
	va_list argptr;
	va_start (argptr, szFormat);
	vfprintf (m_fTrace, szFormat, argptr);
	va_end (argptr);
	fflush(m_fTrace);		//ML added this line
}
	
void BIFSDecoder :: TraceIndent (int nDirection)
{
	m_nIndent += nDirection;
}

void BIFSDecoder :: TraceNewLine ()
{
	if (!m_bNewLine)
	{	m_bNewLine = TRUE;
		putc ('\n', m_fTrace);
	}
}
#endif

/*-----------------------------------------------------------------
Parse RoutesDescription
-----------------------------------------------------------------*/
void BIFSDecoder :: ParseRoutesDescription (Root *pRoot)
{
	if (ParseInt (1))		// if ListDescription 
		do
			ParseRoute (pRoot);
		while (ParseInt (1));		// while moreROUTEs
	else
		for (int i = ParseInt (10); --i >= 0;)
			ParseRoute (pRoot);
}

/*-----------------------------------------------------------------
Parse one Route and insert it into the scene whose root is pRoot
-----------------------------------------------------------------*/
void BIFSDecoder :: ParseRoute (Root *pRoot, int nID)
{
	/*-----------------------------------------------------------------
	if nID >= we already know the ID
	-----------------------------------------------------------------*/
	if (nID < 0)
	{	if (ParseInt (1))	// if isUpdatable
		{	nID = ParseInt (10);
			TRACEF ("DEF %d ", nID);
		}
	}

	int nNodeID = ParseInt (10);
	TRACEF ("ROUTE %d.", nNodeID);
	MediaObject *pNode = pRoot -> FindNode (nNodeID);
	if (pNode == NULL)
		throw BifsException (BIFX_NODEID_NOT_FOUND);
	if (!pNode -> AddRoute (nID, this))
		throw BifsException (BIFX_PARSING_ERROR);
	TRACENEWLINE;
}

/*-----------------------------------------------------------------
Parse an UpdateFrame
-----------------------------------------------------------------*/
void BIFSDecoder :: ParseUpdateFrame ()
{
	TRACEF ("UpdateScene {", 0);
	TRACENEWLINE;
	TRACEINDENT (1);
	m_nLastBit = 8;
	m_nSize = 0;
	m_nLastByte = 0;
	
	do
	{	
		int nCommandCode = ParseInt (2);

		m_pExecutive -> LockScene ();
		switch (nCommandCode) {
		case 0 :
			ParseInsertionCommand ();
			break;
		case 1:
			ParseDeletionCommand ();
			break;
		case 2:
			ParseReplacementCommand ();
			break;
		case 3:
			ParseSceneReplaceCommand ();
			break;
		}
		m_pExecutive -> UnlockScene ();

		TRACENEWLINE;
	}
	while (ParseInt (1));

	TRACEINDENT (-1);
	TRACEF ("}", 0);
	TRACENEWLINE;
}

/*-----------------------------------------------------------------
Parse Insertion Command
-----------------------------------------------------------------*/
void BIFSDecoder :: ParseInsertionCommand ()
{
	int nID;
	MediaObject *pNode;

	switch (ParseInt (2)) {
	case 0:
		nID = ParseInt (10);
		TRACEF ("Insert node into %d.children ", nID);
		pNode = m_pScene -> FindNode (nID);
		if (!pNode)
			throw BifsException (BIFX_NODEID_NOT_FOUND);
		((GroupNode *) pNode) -> ParseChild (this);
		break;
	case 1:
		nID = ParseInt (10);
		TRACEF ("Insert into %d.", nID);
		pNode = m_pScene -> FindNode (nID);
		if (!pNode)
			throw BifsException (BIFX_NODEID_NOT_FOUND);
		pNode -> InsertValue (this);
		break ;
	case 2:
		ParseRoute (m_pScene);
		break;
	}
}

/*-----------------------------------------------------------------
Parse DeletionCommand
-----------------------------------------------------------------*/
void BIFSDecoder :: ParseDeletionCommand ()
{
	int nID;
	MediaObject *pNode;

	int nType = ParseInt (2);
	nID = ParseInt (10);
	switch (nType) {
	case 0:
		TRACEF ("Delete %d", nID);
		pNode = m_pScene -> FindNode (nID);
		if (!pNode)
			throw BifsException (BIFX_NODEID_NOT_FOUND);
		pNode -> Delete ();
		break;
	case 1:
		TRACEF ("Delete %d.", nID);
		pNode = m_pScene -> FindNode (nID);
		if (!pNode)
			throw BifsException (BIFX_NODEID_NOT_FOUND);
		pNode -> DeleteValue (this);
		break ;
	case 2:
		TRACEF ("Delete ROUTE %d", nID);
		delete m_pScene -> FindRoute (nID);
		break ;
	}
}

/*-----------------------------------------------------------------
Parse replacement command
-----------------------------------------------------------------*/
void BIFSDecoder :: ParseReplacementCommand ()
{
	int nID;
	MediaObject *pNode;

	int nType = ParseInt (2);
	nID = ParseInt (10);
	switch (nType) {
	case 0:
		TRACEF ("Replace %d by ", nID);
		pNode = m_pScene -> FindNode (nID);
		if (!pNode)
			throw BifsException (BIFX_NODEID_NOT_FOUND);
		pNode -> ReplaceNode (this);
		break;
	case 1:
		TRACEF ("Replace %d.", nID);
		pNode = m_pScene -> FindNode (nID);
		if (!pNode)
			throw BifsException (BIFX_NODEID_NOT_FOUND);
		pNode -> ReplaceValue (this);
		break;
	case 2:
		TRACEF ("Replace %d.", nID);
		pNode = m_pScene -> FindNode (nID);
		if (!pNode)
			throw BifsException (BIFX_NODEID_NOT_FOUND);
		pNode -> ReplaceIndexedValue (this);
		break ;
	case 3:
		TRACEF ("Replace ROUTE %d by ", nID);
		delete m_pScene -> FindRoute (nID);
		ParseRoute (m_pScene, nID);
		break ;
	}
}

void BIFSDecoder::ParseSceneReplaceCommand()
{

}
