/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Adaptation Layer
FILE:       DataChan.cpp
AUTHOR:     Zvi Lifshitz (VDOnet Corp.)
OVERVIEW:   Implementation of DataChannel object
-------------------------------------------------------------------------*/
#include "defs.h"
#include "Channel.h"
#include "Stream.h"
#include "Except.h"
#include "Service.h"

/*-----------------------------------------------------------------
Config() binds a MediaStream, copy the config params,
and store values of predefined parameters
-----------------------------------------------------------------*/
void DataChannel :: Config (
	MediaStream *pStream,
	ESCONFIGPARAMS *pConfigParams,
	int nExpectedAULength
)
{
	extern ESCONFIGPARAMS g_arPredefined [];

	m_pStream = pStream;
	m_pBuffer = NULL;
	m_nExpectedAULength = nExpectedAULength;
	m_dwPreCTS = m_dwCTS = m_dwTimeOffset = 0;
	m_nAULength = 0;

	/*-----------------------------------------------------------------
	Store ESconfigParams. If the pointer is null, use the defaults.
	-----------------------------------------------------------------*/
	if (pConfigParams)
		m_configParams = *pConfigParams;
	else
		m_configParams = g_arPredefined [0];
}

/*-----------------------------------------------------------------
Close the stream before deleting the object
-----------------------------------------------------------------*/
DataChannel :: ~DataChannel ()
{
	if (m_pStream)
		m_pStream -> Close ();
}

/*-----------------------------------------------------------------
Notify the service that the channel is ready
-----------------------------------------------------------------*/
void DataChannel :: ChannelReady ()
{
	m_bChannelReady = TRUE;
	if (m_nTag >= 0)
		m_pService -> ChannelsReady (1, &m_nTag);
}

/*-----------------------------------------------------------------
Called from Service when channel is open. Receives the CAT.
-----------------------------------------------------------------*/
void DataChannel :: OnChannelOpen (int nTag)
{
	m_nTag = nTag;
	if (m_bChannelReady)
		m_pService -> ChannelsReady (1, &m_nTag);
}

/*-----------------------------------------------------------------
Parse an AU and post it to the stream
-----------------------------------------------------------------*/
void DataChannel :: OnPacketReceived (LPBYTE pData, int nLength)
{
	/*-----------------------------------------------------------------
	Save packet address. GetBit() and GetByte() will be used
	to fetch information from the packet.
	-----------------------------------------------------------------*/
	m_pData = pData;
	m_nLastBit = 8;			// will cause next bit to be retrieved
							// from a new byte
	ResetByteCount ();		// needed for counting header bytes

	int nSequenceNumber = 0;
	int nObjectClockReference = 0;
	BOOL bDecodingTimeStampFlag = FALSE;
	BOOL bPresentationTimeStampFlag = FALSE;
	BOOL bExpirationTimeStampFlag = FALSE;
	BOOL bAccessUnitLengthFlag = FALSE;
	BOOL bInstantBitrateFlag = FALSE;
	DWORD dwDecodingTimeStamp = 0;
	DWORD dwExpirationTimeStamp = 0;
	int nAccessUnitLength = 0;
	int nInstantBitrate = 0;
	int nDegradationPriority = 0;

	BOOL bAccessUnitStart =
		m_configParams.m_bUseAccessUnitStartFlag ? GetBits (1) :
		m_nAULength == 0;
	BOOL bRandomAccessPoint =
		m_configParams.m_UseRandomAccessPointFlag ? GetBits (1) : 1;
	BOOL bAccessUnitEnd =
		m_configParams.m_UseAccessUnitEndFlag ? GetBits (1) : 1;
	BOOL bOCRflag =
		m_configParams.m_nOCRlength > 0 ? GetBits (1) : 0;
	BOOL bPaddingFlag =
		m_configParams.m_bUsePaddingFlag ? GetBits (1) : 0;

	int nPaddingBits = 0;
	if (bPaddingFlag)
		nPaddingBits = GetBits (3);
	else
	{
		nSequenceNumber =
			GetBits (m_configParams.m_nSeqNumLength);
		if (bOCRflag)
		{
			nObjectClockReference =
				GetBits (m_configParams.m_nOCRlength);
		}
		if (bAccessUnitStart)
		{
			if (m_configParams.m_bUseTimeStamps)
			{	bDecodingTimeStampFlag = GetBits (1);
				bPresentationTimeStampFlag = GetBits (1);
				bExpirationTimeStampFlag = GetBits (1);
			}
			if (m_configParams.m_nAUlength > 0)
				bAccessUnitLengthFlag = GetBits (1);
			if (m_configParams.m_nInstantBitrateLength > 0)
				bInstantBitrateFlag = GetBits (1);
			if (bDecodingTimeStampFlag)
			{
				dwDecodingTimeStamp =
					GetBits (m_configParams.m_nTimeStampLength);
			}
			if (bPresentationTimeStampFlag)
			{
				DWORD dwCTS = GetBits (
					m_configParams.m_nTimeStampLength
				);

				if (dwCTS + m_dwTimeOffset <= m_dwPreCTS && m_dwPreCTS != 0)
					m_dwTimeOffset += 1 << m_configParams.m_nTimeStampLength;
				m_dwPreCTS = dwCTS + m_dwTimeOffset;
				m_dwCTS =
					m_dwPreCTS * 1000 /
					m_configParams.m_nTimeStampResolution;
			}
			if (bExpirationTimeStampFlag)
			{
				dwExpirationTimeStamp =
					GetBits (m_configParams.m_nTimeStampLength);
			}
			if (bAccessUnitLengthFlag)
			{
				nAccessUnitLength =
					GetBits (m_configParams.m_nAUlength);
			}
			if (bInstantBitrateFlag)
			{
				nInstantBitrate =
					GetBits (m_configParams.m_nInstantBitrateLength);
			}
			nDegradationPriority =
				GetBits (m_configParams.m_nDegradationPriorityLength);
		}
	}

	/*-----------------------------------------------------------------
	Get payload length by substructing header length from unit length
	-----------------------------------------------------------------*/
	int nOldLength = m_nAULength;

	nLength -= GetByteCount ();
	m_nAULength += nLength;

	/*-----------------------------------------------------------------
	If this is a new AU, allocate buffer on the stream, otherwise,
	reallocate if necessary.
	If this is AU end, we know exactly how much to allocate. Otherwise,
	we use the size recommended by the decoder.
	-----------------------------------------------------------------*/
	if (m_pStream != NULL)
	{	if (bAccessUnitStart)
			m_pBuffer = (LPBYTE) m_pStream -> Allocate (
				bAccessUnitEnd ? m_nAULength : m_nExpectedAULength
			);
		else if (m_nAULength > m_nExpectedAULength)
			m_pBuffer = (LPBYTE) m_pStream -> Reallocate (m_nAULength);

		if (m_pBuffer == END_OF_STREAM)
			throw DemuxException ();

		if (m_pBuffer == NULL)
			throw DemuxException (DemuxException :: MEMORY_ERROR);
	}
		
	/*-----------------------------------------------------------------
	Copy the AU payload to the buffer
	-----------------------------------------------------------------*/
	if (m_pBuffer)
		memcpy (m_pBuffer + nOldLength, m_pData, nLength);

	/*-----------------------------------------------------------------
	If this is end of AU, dispatch the buffer into the MediaStream.
	-----------------------------------------------------------------*/
	if (bAccessUnitEnd && m_pStream)
	{	m_pStream -> Dispatch (m_nAULength, m_dwCTS);
		m_nAULength = 0;
	}
}

/*-----------------------------------------------------------------
GetByte() retrieves a new byte from the source file
-----------------------------------------------------------------*/
int DataChannel :: GetByte ()
{
	m_nLastByte = *m_pData++;
	m_nLastBit = 8;		// The whole byte is consumed, unless this
						// function is called from GetBits, which will
						// reset m_nLastBit to zero.
	++m_nCount;
	return m_nLastByte;
}

/*-----------------------------------------------------------------
GetBits () retrieve nBits bits from the source file, returning them as 
an int.
-----------------------------------------------------------------*/
int DataChannel :: GetBits (int nBits)
{
	int nResult = 0;

	while (nBits-- > 0)
	{
		/*-----------------------------------------------------------------
		Get next byte, if necessary
		-----------------------------------------------------------------*/
		if (m_nLastBit >= 8)
		{	m_nLastByte = GetByte ();
			m_nLastBit = 0;
		}

		/*-----------------------------------------------------------------
		Retrieve the first bit of the byte, shift the byte so next time
		second bit will be the first
		-----------------------------------------------------------------*/
		nResult <<= 1;
		nResult |= (m_nLastByte & 0x80) >> 7;
		m_nLastByte <<= 1;
		++m_nLastBit;
	}
	return nResult;
}