/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  all
FILE:       ztl.cpp
AUTHOR:     Zvi Lifshitz (VDOnet Corp.)
OVERVIEW:   Implementation of the template library for data containers.
Currently includes:

ZArray<type>
ZString
-------------------------------------------------------------------------*/
#include "defs.h"
#include "ztl.h"

/*-----------------------------------------------------------------
ZArrayBase :: Stress () - make sure we have room for nSize+1 elements
-----------------------------------------------------------------*/
void ZArrayBase :: Stress (int nSize)
{
	int nNewSize = (++nSize) * m_nElementSize;

	/*-----------------------------------------------------------------
	If necessary, extend the buffer by 25% or GROW_BY, whatever is larger,
	but not less than what we need.
	-----------------------------------------------------------------*/
	if (nNewSize > m_nSize)
	{	nNewSize = max (nNewSize, m_nSize + max (GROW_BY, m_nSize / 4));
		if (m_pBuffer == NULL)
			m_pBuffer = malloc (nNewSize);
		else
			m_pBuffer = realloc (m_pBuffer, nNewSize);
		if (m_pBuffer == NULL)
			throw;
		m_nSize = nNewSize;
	}
	if (m_nLength < nSize)
		m_nLength = nSize;
}

/*-----------------------------------------------------------------
Clear all elements
-----------------------------------------------------------------*/
void ZArrayBase :: Clear ()
{
	if (m_pBuffer)
		free (m_pBuffer);
	m_pBuffer = NULL;
	m_nSize = m_nLength = 0;
}

/*-----------------------------------------------------------------
Make room for new element at position i
-----------------------------------------------------------------*/
void ZArrayBase :: MakeRoomAt (int i)
{
	Stress (m_nLength);
	if (m_nLength > i + 1)
		memmove (
			((char *) m_pBuffer) + (i + 1) * m_nElementSize,
			((char *) m_pBuffer) + i * m_nElementSize,
			(m_nLength - i - 1) * m_nElementSize
		);
	memset (((char *) m_pBuffer) + i * m_nElementSize, 0, m_nElementSize);
}

/*-----------------------------------------------------------------
Delete element at position i
-----------------------------------------------------------------*/
void ZArrayBase :: DeleteFrom (int i)
{
	if (m_nLength > i + 1)
		memmove (
			((char *) m_pBuffer) + i * m_nElementSize,
			((char *) m_pBuffer) + (i + 1) * m_nElementSize,
			(m_nLength - i - 1) * m_nElementSize
		);
	--m_nLength;
}

/*-----------------------------------------------------------------
Scans the array and returns a pointer to an element which meets some
criterion, or NULL if none does.
ZCriterion is a base class for objects which have a virtual function,
IsOK(), which receives a pointer to NULL, casts it to pointer to
some object, and checks whether the object meets some criterion
-----------------------------------------------------------------*/
void *ZArrayBase :: Search (ZCriterion *pCriterion)
{
	int i;
	void *p;

	for (i = 0; i < m_nLength; i++)
	{	p = ((char *) m_pBuffer) + i * m_nElementSize;
		if (pCriterion -> IsOK (p))
			return p;
	}
	return NULL;
}

/*-----------------------------------------------------------------
Copy the contents of one array to the other, leaving the source array
empty
-----------------------------------------------------------------*/
void ZArrayBase :: Pump (ZArrayBase &rArray)
{
	m_pBuffer = rArray.m_pBuffer;
	m_nLength = rArray.m_nLength;
	m_nSize = rArray.m_nSize;
	rArray.m_pBuffer = NULL;
	rArray.m_nLength = rArray.m_nSize = 0;
}

/*-----------------------------------------------------------------
ZString constructor. Allocate memory and assign value
-----------------------------------------------------------------*/
ZString :: ZString (LPCSTR szStr)
{
	if (szStr == NULL)
	{	m_nLength = m_nSize = 0;
		m_pBuffer = NULL;
	}
	else
	{	m_nLength = strlen (szStr);
		m_nSize = m_nLength + 1;
		if ((m_pBuffer = (char *) malloc (m_nSize)) == NULL)
			throw;
		memcpy (m_pBuffer, szStr, m_nSize);
	}
}

/*-----------------------------------------------------------------
ZString constructor with string and length
-----------------------------------------------------------------*/
ZString :: ZString (LPCSTR szStr, int nLength)
{
	m_nLength = nLength;
	m_nSize = m_nLength + 1;
	if ((m_pBuffer = (char *) malloc (m_nSize)) == NULL)
		throw;
	memcpy (m_pBuffer, szStr, nLength);
	m_pBuffer [nLength] = 0;
}

/*-----------------------------------------------------------------
ZString copy constructor
-----------------------------------------------------------------*/
ZString :: ZString (const ZString &str)
{
	m_nLength = str.m_nLength;
	m_nSize = str.m_nSize;
	m_pBuffer = NULL;
	if (str.m_pBuffer)
	{
        	if ((m_pBuffer = (char *) malloc (m_nSize)) == NULL)
				throw;
		memcpy (m_pBuffer, str.m_pBuffer, m_nLength + 1);
	}
}

/*-----------------------------------------------------------------
Make sure that the string has enough space for the ith byte.
If bFill is TRUE, extend the string by padding with blanks and appending
a null byte.
-----------------------------------------------------------------*/
void ZString :: Stress (int i, BOOL bFill)
{
	int nNewSize;

	/*-----------------------------------------------------------------
	If necessary, extend the buffer by 25% or GROW_BY, whatever is larger,
	but not less than i + 2 (for the terminating null byte).
	-----------------------------------------------------------------*/
	if (i >= m_nSize)
	{	nNewSize = max (i + 2, m_nSize + max (GROW_BY, m_nSize / 4));
		m_pBuffer = m_pBuffer == NULL ?
			(char *) malloc (nNewSize) :
			(char *) realloc (m_pBuffer, nNewSize);
		if (m_pBuffer == NULL)
			throw;
		m_nSize = nNewSize;
		if (i == 0)
			*m_pBuffer = 0;		
	}

	if (!bFill)
		return;

	/*-----------------------------------------------------------------
	If necessary, fill with blanks and append a null byte
	-----------------------------------------------------------------*/
	if (i > m_nLength)
		memset (m_pBuffer + m_nLength, ' ', i - m_nLength);
	if (i >= m_nLength)
		m_pBuffer [m_nLength = i + 1] = 0;
}

/*-----------------------------------------------------------------
Return a reference to position i. Make sure string is large enough
-----------------------------------------------------------------*/
char & ZString :: operator [] (int i)
{
	Stress (i, TRUE);
	return m_pBuffer [i];
}

/*-----------------------------------------------------------------
ZString assignment
-----------------------------------------------------------------*/
ZString &ZString :: operator = (const ZString &str)
{
	if (str.m_pBuffer == NULL)
	{	if (m_pBuffer)
			free (m_pBuffer);
		m_pBuffer = NULL;
		m_nLength = m_nSize = 0;
	}
	else
	{	Stress (m_nLength = str.m_nLength);
		memcpy (m_pBuffer, str.m_pBuffer, m_nLength + 1);
	}
	return *this;
}

/*-----------------------------------------------------------------
ZString assignment
-----------------------------------------------------------------*/
ZString &ZString :: operator = (LPCSTR szStr)
{
	if (szStr == NULL)
	{	if (m_pBuffer)
			free (m_pBuffer);
		m_pBuffer = NULL;
		m_nLength = m_nSize = 0;
	}
	else
	{	Stress (m_nLength = strlen (szStr));
		strcpy (m_pBuffer, szStr);
	}
	return *this;
}

/*-----------------------------------------------------------------
Append one character to a string
-----------------------------------------------------------------*/
ZString &ZString :: operator += (char c)
{
	Stress (m_nLength + 1);
	m_pBuffer [m_nLength++] = c;
	m_pBuffer [m_nLength] = 0;
	return *this;
}

/*-----------------------------------------------------------------
Concatenate two strings
-----------------------------------------------------------------*/
ZString &ZString :: operator += (const ZString &str)
{
	Stress (m_nLength + str.m_nLength);
	memcpy (m_pBuffer + m_nLength, str.m_pBuffer, str.m_nLength);
	m_pBuffer [m_nLength += str.m_nLength] = 0;
	return *this;
}

/*-----------------------------------------------------------------
Concatenate two strings
-----------------------------------------------------------------*/
ZString ZString :: operator + (const ZString &str) const
{
	return ZString (*this) += str;
}

/*-----------------------------------------------------------------
Insert a character into position i
-----------------------------------------------------------------*/
char ZString :: InsertAt (int i, char c, int n)
{
	Stress (m_nLength += n);
	memmove (m_pBuffer + i + n, m_pBuffer + i, m_nLength - i);
	memset (m_pBuffer + i, c, n);
	return c;
}

/*-----------------------------------------------------------------
Delete a character from position i. Returns the deleted character
-----------------------------------------------------------------*/
char ZString :: DeleteFrom (int i)
{
	char cResult = m_pBuffer [i];

	memmove (m_pBuffer + i, m_pBuffer + i + 1, m_nLength - i);
	--m_nLength;
	return cResult;
}

/*-----------------------------------------------------------------
Compare two strings
-----------------------------------------------------------------*/
BOOL ZString :: operator == (const ZString &str) const
{
	return m_nLength == str.m_nLength && !memcmp (*this, str, m_nLength);
}

/*-----------------------------------------------------------------
Find a character c in a ZString, starting the search at offset iOffset.
Returns the distance of the index of the found character from nOffset,
or -1 if did not find
-----------------------------------------------------------------*/
int ZString :: Find (int c, int nOffset) const
{
	int i;

	for (i = nOffset; i < GetLength (); i++)
		if ((*this) [i] == c)
			return i - nOffset;
	return -1;
}

/*-----------------------------------------------------------------
Return a string made of a substring of this string.
If nLength is negative, substring is from nFrom till end of string
-----------------------------------------------------------------*/
ZString ZString :: Substr (int nFrom, int nLength) const
{
	if (nFrom >= m_nLength || nLength == 0)
		return ZString ();
	if (nLength < 0)
		nLength = m_nLength - nFrom;
	return ZString (
		m_pBuffer + nFrom,
		min (m_nLength - nFrom, nLength)
	);
}

/*-----------------------------------------------------------------
Cut nLength chars out of a string, starting at nFrom.
If nLength is negative, cut till end of string
-----------------------------------------------------------------*/
void ZString :: Cut (int nFrom, int nLength)
{
	if (nFrom >= m_nLength || nLength == 0)
		return;
	if (nLength < 0)
		nLength = m_nLength - nFrom;
	memmove (
		m_pBuffer + nFrom,
		m_pBuffer + nFrom + nLength,
		m_nLength - nFrom - nLength + 1
	);
	m_nLength -= nLength;
}

/*-----------------------------------------------------------------
Copy a string to the object. No more than nMax characters
will be copied
-----------------------------------------------------------------*/
ZString & ZString :: Copy (LPCSTR szStr, int nMax)
{
	m_nLength = min ((int) strlen (szStr), nMax);
	Stress (m_nLength);
	memcpy (m_pBuffer, szStr, m_nLength);
	m_pBuffer [m_nLength] = 0;
	return *this;
}

#ifdef WIN32

/*-------------------------------------------------------------------------
ZSemaphore constructor. Create unnamed non-inherited semaphore.
If lInitialCount is -1, it is set to lMaxCount.
-------------------------------------------------------------------------*/
ZSemaphore :: ZSemaphore (long lMaxCount, long lInitialCount)
{
	if (lInitialCount < 0)
		lInitialCount = lMaxCount;
	m_hSemaph = CreateSemaphore (NULL, lInitialCount, lMaxCount, NULL);
}

/*-------------------------------------------------------------------------
Try to reserve the semphore. Return immediately if semaphore already
reserved. Return TRUE if succeeded.
-------------------------------------------------------------------------*/
BOOL ZSemaphore :: Try ()
{
	return WaitForSingleObject (m_hSemaph, 0) != WAIT_TIMEOUT;
}	 

/*-------------------------------------------------------------------------
Try to reserve the mutex. Return immediately if mutex already
reserved. Return TRUE if succeeded.
-------------------------------------------------------------------------*/
BOOL ZMutex :: Try ()
{
	return WaitForSingleObject (m_hMutex, 0) != WAIT_TIMEOUT;
}	 

#ifdef _DEBUG
void ZMutex :: Trace (LPCSTR szOp)
{
	char szBuffer [80];

	sprintf (
		szBuffer,
		"Mutex handle: %x, operation: %s\n",
		m_hMutex,
		szOp
	);
	OutputDebugString (szBuffer);
}
#endif		// _DEBUG

#elif defined(PTHREADS)

#include <sys/ptimers.h>

/*-------------------------------------------------------------------------
ZSemaphore constructor. Create unnamed non-inherited semaphore.
If lInitialCount is -1, it is set to lMaxCount.
-------------------------------------------------------------------------*/
ZSemaphore :: ZSemaphore (long lMaxCount, long lInitialCount)
{
	if (lInitialCount < 0)
		lInitialCount = lMaxCount;
	m_nCount = lInitialCount;

	::pthread_cond_init (&m_Cond, NULL);
	::pthread_mutex_init (&m_Mutex, NULL);
}

/*-------------------------------------------------------------------------
ZSemaphore destructor.
-------------------------------------------------------------------------*/
ZSemaphore :: ~ZSemaphore ()
{
	::pthread_cond_destroy (&m_Cond);
	::pthread_mutex_destroy (&m_Mutex);
}

/*-------------------------------------------------------------------------
Reserve the semaphore. Return TRUE if succeeded.
-------------------------------------------------------------------------*/
DWORD ZSemaphore :: P (UINT uTimeout)
{
	::pthread_mutex_lock(&m_Mutex);

	if (uTimeout == INFINITE)
		while (m_nCount == 0)
		{	if (::pthread_cond_wait(&m_Cond, &m_Mutex))
			{
				::perror("pthread_cond_wait");
				::pthread_mutex_unlock(&m_Mutex);
				return FALSE;
			}
		}
	else
	{
		struct timeval currenttime;
		struct timespec timeout;

		::gettimeofday(&currenttime);

		timeout.tv_nsec =
			long (currenttime.tv_usec * 1000 + uTimeout * 1000000);
			timeout.tv_sec =
				currenttime.tv_sec + timeout.tv_nsec / 1000000000;
		timeout.tv_nsec %= 1000000;

		while (m_nCount == 0)
			switch (::pthread_cond_timedwait(
				&m_Cond,
				&m_Mutex,
				&timeout)
			){
			default:
				::perror("pthread_cond_timedwait");

			case ETIMEDOUT:
				::pthread_mutex_unlock(&m_Mutex);
				return FALSE;

			case 0:
				;
			}
	}

	m_nCount--;
	::pthread_mutex_unlock(&m_Mutex);
	return TRUE;
}

/*-------------------------------------------------------------------------
Release the semphore. Return TRUE if succeeded.
-------------------------------------------------------------------------*/
BOOL ZSemaphore :: V ()
{
	::pthread_mutex_lock(&m_Mutex);
	if (m_nCount == 0)
		::pthread_cond_signal(&m_Cond);
	m_nCount++;
	::pthread_mutex_unlock(&m_Mutex);
	return FALSE;
}

/*-------------------------------------------------------------------------
Try to reserve the semphore. Return immediately if semaphore already
reserved. Return TRUE if succeeded.
-------------------------------------------------------------------------*/
BOOL ZSemaphore :: Try ()
{
	BOOL ret;
	::pthread_mutex_lock(&m_Mutex);
	if (m_nCount == 0)
		ret = FALSE;
	else
	{
		m_nCount--;
		ret = TRUE;
	}
	::pthread_mutex_unlock(&m_Mutex);
	return ret;
}

/*-------------------------------------------------------------------------
Try to reserve the mutex. Return immediately if mutex already
reserved. Return TRUE if succeeded.
-------------------------------------------------------------------------*/
BOOL ZMutex :: Try ()
{
	switch (::pthread_mutex_trylock(&m_Mutex)) {
	case 0:
		return TRUE;
	default:
		perror("pthread_mutex_trylock");
	case EBUSY:
		return FALSE;
	}
}	 

#else

usptr_t* ZSGIArena::m_pArena = NULL;

/*-------------------------------------------------------------------------
ZSemaphore constructor. Create unnamed non-inherited semaphore.
If lInitialCount is -1, it is set to lMaxCount.
-------------------------------------------------------------------------*/
ZSemaphore :: ZSemaphore (long lMaxCount, long lInitialCount)
{
	if (lInitialCount < 0)
		lInitialCount = lMaxCount;
	
	m_pSema = usnewsema(ZSGIArena::arena(), lInitialCount);
	if (!m_pSema)
	  ::perror("usnewsema");
}

/*-------------------------------------------------------------------------
Try to reserve the mutex. Return immediately if mutex already
reserved. Return TRUE if succeeded.
-------------------------------------------------------------------------*/
BOOL ZMutex :: Try ()
{
	switch (uscsetlock(m_Mutex, 1)) {
	default:
		::perror("uscsetlock");
	case 0:
		return FALSE;
	case 1:
		return TRUE;
	}
}	 

/*-------------------------------------------------------------------------
Reserve the mutex.  Return TRUE if succeeded.
-------------------------------------------------------------------------*/
DWORD ZMutex :: P ()
{
	switch (ussetlock(m_Mutex)) {
	default:
		::perror("ussetlock");
	case 0:
		return FALSE;
	case 1:
		return TRUE;
	}
}	 

#endif

/*-----------------------------------------------------------------
A function that calculates how many bits are required to store a number
which is less than nMax
-----------------------------------------------------------------*/
int BitFieldWidth (int nMax)
{
	int nResult = 0;
	--nMax;
	while (nMax > 0)
	{	++nResult;
		nMax >>= 1;
	}
	return nResult;
}