// -*- C++ -*-

#if !defined(_ztl_h)	// Sentry, use file only if it's not already included.
#define _ztl_h

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  all
FILE:       ztl.h
AUTHOR:     Zvi Lifshitz (VDOnet Corp.)
OVERVIEW:   A template library for data containers.  Currently includes:

ZArray<type>
ZMap<type, type>
ZString
ZSemaphore
ZMutex
ZThread
Some general utility functions
-------------------------------------------------------------------------*/

#include "defs.h"
#include "limits.h"
#include "assert.h"

/*-----------------------------------------------------------------
Array class
-----------------------------------------------------------------*/
class ZArrayBase
{
	enum {GROW_BY = 16};

	int m_nElementSize;
	int m_nLength, m_nSize;

protected:
	void *m_pBuffer;

	ZArrayBase (int nElementSize) :
		m_nElementSize (nElementSize),
		m_nLength (0),
		m_nSize (0),
		m_pBuffer (NULL)
	{}
	~ZArrayBase () {if (m_pBuffer) free (m_pBuffer);}
	void Clear ();
	void Stress (int nSize);
	int Increase ()
	{
		int nResult = m_nLength;
		MakeRoomAt (m_nLength);
		return nResult;
	}

	void MakeRoomAt (int i);
	void DeleteFrom (int i);
	void *Search (class ZCriterion *pCriterion);
	void Pump (ZArrayBase &rArray);

public:

  /// Get number of elements

	int GetLength () const {return m_nLength;}
};


/*-----------------------------------------------------------------
Parameterised array class
-----------------------------------------------------------------*/
template <class T> class ZArray : public ZArrayBase
{
public:

	/*-----------------------------------------------------------------
	constructor
	-----------------------------------------------------------------*/
	ZArray () : ZArrayBase (sizeof (T)) {}

	/*-----------------------------------------------------------------
	destructor.
	-----------------------------------------------------------------*/
	~ZArray ()
	{
		int i;

		/*-----------------------------------------------------------------
		The following trick will make sure the destructors for all 
		elements are called.
		-----------------------------------------------------------------*/
		for (i = 0; i < GetLength (); i++)
		{	T defT;

			memcpy (&defT, (T*) m_pBuffer + i, sizeof (T));
		}
	}
		
	/*-----------------------------------------------------------------
	Clear all.
	-----------------------------------------------------------------*/
	void Clear  ()
	{
		int i;

		/*-----------------------------------------------------------------
		The following trick will make sure the destructors for all 
		elements are called.
		-----------------------------------------------------------------*/
		for (i = 0; i < GetLength (); i++)
		{	T defT;

			memcpy (&defT, (T*) m_pBuffer + i, sizeof (T));
		}
		ZArrayBase :: Clear ();
	}

	/*-----------------------------------------------------------------
	type conversion
	-----------------------------------------------------------------*/
	operator const T* () const {return (T*) m_pBuffer;}

	/*-----------------------------------------------------------------
	Expand the array, fill all new entries with default object
	-----------------------------------------------------------------*/
	void Expand (int i)
	{
		int j = GetLength ();

		if (i >= j)
		{	T defT;

			Stress (i);
			for (;j <= i; j++)
				memcpy ((T*) m_pBuffer + j, &defT, sizeof (T));
		}
	}
		
	/*-----------------------------------------------------------------
	set element
	-----------------------------------------------------------------*/
	T &SetAt (int i, const T t)
	{
		Expand (i);
		return ((T*) m_pBuffer) [i] = t;
	}

	/*-----------------------------------------------------------------
	Delete element at position i
	-----------------------------------------------------------------*/
	void DeleteFrom (int i)
	{
		T defT;

		/*-----------------------------------------------------------------
		The following trick will make sure the destructor of the element
		is called.
		-----------------------------------------------------------------*/
		memcpy (&defT, (T*) m_pBuffer + i, sizeof (T));
		ZArrayBase :: DeleteFrom (i);
	}

	/*-----------------------------------------------------------------
	get element
	-----------------------------------------------------------------*/
	T &ElementAt (int i) const
	{
		if (i >= GetLength ())
			assert (0);
		return ((T*) m_pBuffer) [i];
	}

	/*-----------------------------------------------------------------
	get element
	-----------------------------------------------------------------*/
	T &operator [] (int i)
	{
		if (i >= GetLength ())
			assert (0);
		return ((T*) m_pBuffer) [i];
	}

	const T &operator [] (int i) const
	{
		return ((T*) m_pBuffer) [i];
	}

	/*-----------------------------------------------------------------
	get element, expand if necessary
	-----------------------------------------------------------------*/
	T &ElementAt (int i)
	{
		Expand (i);
		return ((T*) m_pBuffer) [i];
	}

	/*-----------------------------------------------------------------
	Find an element and return its index
	-----------------------------------------------------------------*/
	int Find (const T &t)
	{
		for (int i = GetLength (); --i >= 0;)
			if ((*this) [i] == t)
				return i;
		return -1;
	}

	/*-----------------------------------------------------------------
	Add an element to the end of the array
	-----------------------------------------------------------------*/
	T &operator += (const T &t)
	{
		int n = Increase ();
		T defT;

		memcpy ((T*) m_pBuffer + n, &defT, sizeof (T));
		((T*) m_pBuffer) [n] = t;
		return ((T*) m_pBuffer) [n];
	}

	/*-----------------------------------------------------------------
	Find an element and delete it from the array
	-----------------------------------------------------------------*/
	BOOL operator -= (const T &t)
	{
		int i = Find (t);

		if (i >= 0)
		{
			DeleteFrom (i);
			return TRUE;
		}
		else
			return FALSE;
	}

	T &InsertAt (int i, const T &t)
	{
		if (i >= GetLength ())
			Expand (i);
		else
		{	T defT;

			MakeRoomAt (i);
			memcpy ((T*) m_pBuffer + i, &defT, sizeof (T));
		}
		return ((T*) m_pBuffer) [i] = t;
	}

	/*-----------------------------------------------------------------
	operator ++ appends one default T at the end of the array and returns
	a reference to it
	-----------------------------------------------------------------*/
	T &operator ++ ()
	{
		int n = Increase ();
		T defT;

		memcpy ((T*) m_pBuffer + n, &defT, sizeof (T));
		return ((T*) m_pBuffer) [n];
	}

	T *Search (class ZCriterion *pCriterion)
	{
		return (T*) ZArrayBase :: Search (pCriterion);
	}

	/*-----------------------------------------------------------------
	Copy the contents of another array. The other array is left empty.
	-----------------------------------------------------------------*/
	ZArray<T> &operator <<= (ZArray<T> &array)
	{
		Clear ();
		Pump (array);
		return *this;
	}
};

/*-----------------------------------------------------------------
ZCriterion is a base class for objects which have a virtual function,
IsOK(), which receives a pointer to NULL, casts it to pointer to
some object, and checks whether the object meets some criterion
-----------------------------------------------------------------*/
class ZCriterion
{
public:
	virtual BOOL IsOK (void *object) = 0;
};

/*-----------------------------------------------------------------
ZMap class
-----------------------------------------------------------------*/
template <class I, class T> class ZMap
{
	/*-----------------------------------------------------------------
	class is implemented as an array of structures which map indice to
	objects. No phisical deletion occurs when an element is deleted.
	Instead the position is marked as free
	-----------------------------------------------------------------*/
	struct Map
	{
		I m_i;
		T m_t;
		BOOL m_bFree;
		Map () : m_bFree (TRUE) {}
		Map (I i) : m_i (i), m_bFree (FALSE) {}
		BOOL operator == (const Map &map) const {return map.m_i == m_i;}
	};
	ZArray<Map> m_array;

public:
	/*-----------------------------------------------------------------
	Return a reference to the element at position index.
	Create a new element if does not exist
	-----------------------------------------------------------------*/
	T &operator [] (I index)
	{
		int nFree = -1;
		for (int i = m_array.GetLength (); --i >= 0;)
			if (m_array [i].m_bFree && nFree < 0)
				nFree = i;
			else if (m_array [i].m_i == index)
				return m_array [i].m_t;
		/*-----------------------------------------------------------------
		Not found. Either use a free cell or create a new one
		-----------------------------------------------------------------*/
		Map &rNew = i >= 0 ? m_array [i] : ++m_array;
		rNew.m_i = index;
		rNew.m_bFree = FALSE;
		return rNew.m_t;
	}

	/*-----------------------------------------------------------------
	The [] operator when this is const returns a const reference to the
	object at index.  The function asserts if the object does not exist
	-----------------------------------------------------------------*/
	const T& operator [] (I index) const
	{
		for (int i = m_array.GetLength (); --i >= 0;)
			if (!m_array [i].m_bFree && m_array [i].m_i == index)
				return m_array [i].m_t;
		assert (0);
		return m_array [0].m_t;
	}

	/*-----------------------------------------------------------------
	Finds first occurrence of an element and stores its index in rIndex
	and a reference to it in rT.  Returns FALSE if did not find any
	-----------------------------------------------------------------*/
	BOOL Find (T t, I &rIndex, T &rT)
	{
		for (int i = m_array.GetLength (); --i >= 0;)
			if (!m_array [i].m_bFree && m_array [i].m_t == t)
			{	rIndex = m_array [i].m_i;
				rT = m_array [i].m_t;
				return TRUE;
			}
		return FALSE;
	}

	/*-----------------------------------------------------------------
	Returns a reference to the object at index (reference is stored in 
	rT). Return FALSE if there is no element at index.
	-----------------------------------------------------------------*/
	BOOL Get (I index, T &rT)
	{
		for (int i = m_array.GetLength (); --i >= 0;)
			if (!m_array [i].m_bFree && m_array [i].m_i == index)
			{	rT = m_array [i].m_t;
				return TRUE;
			}
		return FALSE;
	}

	/*-----------------------------------------------------------------
	Finds first occurrence of an element and remove it from the map.  
	Returns FALSE if none found
	-----------------------------------------------------------------*/
	BOOL operator -= (T t)
	{
		for (int i = m_array.GetLength (); --i >= 0;)
			if (!m_array [i].m_bFree && m_array [i].m_t == t)
			{	Delete (m_array [i]);
				return TRUE;
			}
		return FALSE;
	}

	/*-----------------------------------------------------------------
	Deletes the element at index.  Returns FALSE if there is no element 
	at that position.
	-----------------------------------------------------------------*/
	BOOL DeleteFrom (I index)
	{
		for (int i = m_array.GetLength (); --i >= 0;)
			if (!m_array [i].m_bFree && m_array [i].m_i == index)
			{	Delete (m_array [i]);
				return TRUE;
			}
		return FALSE;
	}

	/*-----------------------------------------------------------------
	Deletes all elements.
	-----------------------------------------------------------------*/
	void Clear () {m_array.Clear ();}

private:
	/*-----------------------------------------------------------------
	Delete an element by marking it as free and storing a default T 
	object after deletion of the existing one
	-----------------------------------------------------------------*/
	void Delete (Map &rElement)
	{
		T oldT, newT;
		rElement.m_bFree = TRUE;
		memcpy (&oldT, &rElement.m_t, sizeof (T));
		memcpy (&rElement.m_t, &newT, sizeof (T));
	}
};

/*-----------------------------------------------------------------
String class
-----------------------------------------------------------------*/
class ZString
{
	enum {GROW_BY = 16};

	int m_nLength, m_nSize;
	LPSTR m_pBuffer;
	
private:
	void Stress (int i, BOOL bFill = FALSE);

public:

	/*-----------------------------------------------------------------
	constructor
	-----------------------------------------------------------------*/
	ZString (LPCSTR szStr, int nLength);

	/*-----------------------------------------------------------------
	constructors
	-----------------------------------------------------------------*/
	ZString (LPCSTR szStr = NULL);
	ZString (const ZString &str);

	~ZString ()
	{	if (m_pBuffer)
			free (m_pBuffer);
	}

	void Cut (int nFrom, int nLength = -1);
	ZString Substr (int nFrom, int nLength = -1) const;
	int Find (int c, int iOffset = 0) const;
	operator LPCSTR () const {return m_pBuffer ? m_pBuffer : "";}
	char &operator [] (int i);
	char operator [] (int i) const {return m_pBuffer [i];}
	ZString &operator += (char c);
	ZString &operator = (const ZString &str);
	ZString &operator = (LPCSTR szStr);
	BOOL operator == (const ZString &str) const;
	char InsertAt (int i, char c, int n = 1);
	char DeleteFrom (int i);
	int GetLength () const {return m_nLength;}
	ZString &operator += (const ZString &str);
	ZString operator + (const ZString &str) const;
	ZString &Copy (LPCSTR szSource, int nMax = INT_MAX);
};

#ifdef WIN32

class ZSemaphore
{
	HANDLE m_hSemaph;

public:
	ZSemaphore (long lMaxCount = 1, long lInitialCount = -1);
	~ZSemaphore () {CloseHandle (m_hSemaph);}
	DWORD P (UINT uTimeout = INFINITE)
		{return WaitForSingleObject (m_hSemaph, uTimeout);}
	BOOL V () {return ReleaseSemaphore (m_hSemaph, 1, NULL);}
	BOOL Try ();
};

class ZMutex
{
	HANDLE m_hMutex;

public:
	ZMutex () {m_hMutex = CreateMutex (NULL, FALSE, NULL);}
	~ZMutex () {CloseHandle (m_hMutex);}
	DWORD P (UINT uTimeout = INFINITE)
		{return WaitForSingleObject (m_hMutex, uTimeout);}
	BOOL V () {return ReleaseMutex (m_hMutex);}
	BOOL Try ();

#ifdef _DEBUG
	void Trace (LPCSTR szOp);
#endif
};

#elif defined(PTHREADS)		// WIN32

#include <pthread.h>

class ZSemaphore
{
    pthread_cond_t m_Cond;
    pthread_mutex_t m_Mutex;
    int m_nCount;

public:
	ZSemaphore (long lMaxCount = 1, long lInitialCount = -1);
	~ZSemaphore ();
	DWORD P (UINT uTimeout = INFINITE);
	BOOL V ();
	BOOL Try ();
};

class ZMutex
{
	pthread_mutex_t m_Mutex;

public:
	ZMutex ()
	{	if (::pthread_mutex_init (&m_Mutex, NULL))
			perror ("pthread_mutex_init");
	}
	~ZMutex ()
	{	if (::pthread_mutex_destroy (&m_Mutex))
			perror ("pthread_mutex_destroy");
	}
	DWORD P ()
	{	if (::pthread_mutex_lock (&m_Mutex))
		{	perror ("pthread_mutex_lock");
			return FALSE;
		}
		else
			return TRUE;
	}
	BOOL V ()
	{	if (::pthread_mutex_unlock (&m_Mutex))
		{	perror ("pthread_mutex_unlock");
			return FALSE;
		}
		else
			return TRUE;
	}
	BOOL Try ();
};

#else		// PTHREADS

#include <ulocks.h>

class ZSGIArena
{
public:
	static usptr_t *arena()
	{
		if (!m_pArena)
		{
			::usconfig(CONF_INITSIZE, 1024 * 1024);
			::usconfig(CONF_INITUSERS, 32);
			::usconfig(CONF_ARENATYPE, US_SHAREDONLY);
			if (!(m_pArena = ::usinit("/dev/zero")))
				::perror("usinit");
		}
		return m_pArena;
	}

private:
	static usptr_t *m_pArena;
};

class ZSemaphore
{
	usema_t *m_pSema;

public:
	ZSemaphore (long lMaxCount = 1, long lInitialCount = -1);
	~ZSemaphore () {::usfreesema(m_pSema, ZSGIArena::arena());}
	DWORD P ()
	{	if (::uspsema(m_pSema) < 0)
		{	::perror("uspsema");
			return FALSE;
		}
		else
			return TRUE;
	}
	BOOL V ()
	{	if (::usvsema(m_pSema) < 0)
		{	::perror("usvsema");
			return FALSE;
		}
		else
			return TRUE;
	}
	BOOL Try () {return ::uscpsema(m_pSema);}
};

class ZMutex
{
	ulock_t m_Mutex;

public:
	ZMutex ()
	{	if (!(m_Mutex = usnewlock(ZSGIArena::arena())))
			::perror("usnewlock");
	}
	~ZMutex () {usfreelock (m_Mutex, ZSGIArena::arena());}
	DWORD P ();
    BOOL V ()
	{	if (usunsetlock(m_Mutex))
		{	::perror("usunsetlock");
			return FALSE;
		}
		else 
			return TRUE;
	}
	BOOL Try ();
};

#endif		// WIN32

/*-----------------------------------------------------------------
A template class which encapsultes thread control
-----------------------------------------------------------------*/
#ifdef WIN32
#include <process.h>
#else
typedef enum
{
	THREAD_PRIORITY_ABOVE_NORMAL,
	THREAD_PRIORITY_BELOW_NORMAL,
	THREAD_PRIORITY_NORMAL
} ZThreadPriority;
#endif

template <class T> class ZThread
{
public:
	typedef void (T :: *ThreadFunction) ();

	ZThread (
		ThreadFunction function,
		T *object,
		int nPriority = THREAD_PRIORITY_NORMAL
	) :
		m_pObject (object),
		m_function (function),
		m_nPriority (nPriority),
		m_smpThreadRunning (1, 0)
	{}
	int Begin ()
	{
#ifdef WIN32
		int hResult = 
			_beginthread ((void (*) (void *)) ThreadProc, 0, this);
#elif defined (PTHREADS)
		pthread_t thread;
		int hResult =
			::pthread_create (
				&thread,
				NULL,
				(void *(*)(void*)) ThreadProc,
				this
			);
#else
		int hResult =
			::sproc (ThreadProc, PR_SADDR | PR_SFDS, this);
#endif

		/*-----------------------------------------------------------------
		Wait till thread retrieves values and run (if we don't wait the
		object might be deleted before the thread has the chance to
		retrieve the values
		-----------------------------------------------------------------*/
		if (hResult != -1)
			m_smpThreadRunning.P ();
		return hResult;
	}

private:
	T *m_pObject;
	ThreadFunction m_function;
	ZSemaphore m_smpThreadRunning;
	int m_nPriority;

	static void ThreadProc (ZThread<T> *pInstance)
	{
		T *pObject = pInstance -> m_pObject;
		ThreadFunction function = pInstance -> m_function;
#ifdef WIN32
		SetThreadPriority (GetCurrentThread(), pInstance -> m_nPriority);
#endif
		pInstance -> m_smpThreadRunning.V ();
		(*pObject.*function) ();
	}
};

/*-----------------------------------------------------------------
Utilities
-----------------------------------------------------------------*/
int BitFieldWidth (int nMax);

#endif                                      // _ztl_h sentry.
