/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Multiplexer
FILE:       JPEG.cpp
AUTHOR:     Matthew Leditschke (CSELT)
OVERVIEW:   Implementation of JPEGStream, the class that handles JPEG
input streams.
-------------------------------------------------------------------------*/
#include "defs.h"
#include "JPEG.h"

/*-----------------------------------------------------------------
JPEGStream constructor determines how the time between JPEG images
will be calculated, depending on the streamData.
  - if not given, defaults to 1 second.
  - if an floating point number is given, this gives interval
  - otherwise the streamData is the name of a file to read
    presentation times from.  Once the end of this file is reached,
	the last difference is used.
-----------------------------------------------------------------*/
JPEGStream :: JPEGStream (FILE *fStream, const char* streamData) : Stream (fStream)
{
	m_timeFile = NULL;
	m_prevChar[0] = 0;
	m_prevChar[1] = 0;
	m_readAhead = FALSE;

	// Handle the default when the streamData field is not given
	if ((streamData == NULL) || (streamData[0] == 0))
	{
		m_constant = TRUE;
		m_nextTime = 0;
		m_intervalTime = 1;
	}

	// Check if the stream data is numeric
	else if (isdigit(streamData[0]))
	{
		m_constant = TRUE;
		m_nextTime = 0;
		m_intervalTime = atof(streamData);
		if (m_intervalTime < 0.0001)
		{
			fprintf(stderr, "visible/JPEG: The display interval time must be greater than 0.\n");
			exit(1);
		}
	}

	// Otherwise assume that the streamData is the name of a file
	else
	{
		m_timeFile = fopen(streamData, "r");
		if (m_timeFile == NULL)
		{
			fprintf(stderr, "visible/JPEG: Error opening '%s'.\n", streamData);
			exit(1);
		}
		m_constant = FALSE;
		m_prevTime[0] = -1;
		m_prevTime[1] = -1;
	}
}

/*-----------------------------------------------------------------
JPEGStream destructor closes the presentation time file if it is
still open.
-----------------------------------------------------------------*/
JPEGStream :: ~JPEGStream ()
{
	if (m_timeFile)
		fclose (m_timeFile);
}

/*-----------------------------------------------------------------
Time stamps are either calculated using a fixed increment, or they
are read from a file.
-----------------------------------------------------------------*/
long JPEGStream :: GetTimeStamp ()
{
	float fTime;

	// Return -1 if the end of the file has been reached
	if (m_prevChar[0] == EOF)
		return -1;

	if (m_constant)
	{
		fTime = m_nextTime;
		m_nextTime += m_intervalTime;
	}

	else
	{
		// Read the next presentation time (in seconds) from the file
		int res = fscanf(m_timeFile, "%f", &fTime);

		// If the end is reached, switch to fixed increments
		if (res == EOF)
		{
			// The increment is the difference in the last two times
			m_constant = TRUE;
			if (m_prevTime[0] == m_prevTime[1])
				m_intervalTime = 1;
			else
				m_intervalTime = m_prevTime[0] - m_prevTime[1];
			fTime = m_prevTime[0] + m_intervalTime;
			m_nextTime = m_prevTime[0] + 2 * m_intervalTime;
		}
		else
		{
			// Keep track of the previous two times
			m_prevTime[1] = m_prevTime[0];
			m_prevTime[0] = fTime;

			//Abort if the times aren't always increasing
			if (m_prevTime[1] >= m_prevTime[0])
			{
				fprintf(stderr, "visible/JPEG: The presentation times aren't always increasing.\n");
				exit(1);
			}
		}
	}

	// Convert the time to milliseconds
	return (long)(fTime * 1000.0);
}

/*-----------------------------------------------------------------
GetChar returns a character and reads one next char. We are always
3 chars ahead, which are used to check whether this is end of AU.
-----------------------------------------------------------------*/
int JPEGStream :: GetChar ()
{
	// If the previous two characters read were the end code of a
	// picture, then return -1.
	if ((m_prevChar[1] == 0xFF) && (m_prevChar[0] == 0xD9))
	{
		m_prevChar[0] = 0; 
		m_prevChar[1] = 0;

		// Before returning -1, we need to check if reading the
		// next byte while reach the end of the file
		m_nextChar = getc();
		m_readAhead = TRUE;
		if (m_nextChar == EOF)
			m_prevChar[0] = m_nextChar;

		return -1;
	}

	// Otherwise read in a new character (allowing for possible
	// read ahead), and update the list of the past two characters
	// read.
	int c = (m_readAhead) ? m_nextChar : getc();
	m_readAhead= FALSE;
	m_prevChar[1] = m_prevChar[0];
	m_prevChar[0] = c;

	return c;
}

static StreamFactory<JPEGStream> bifsFactory ("visual/JPEG");